import logging
from azure.identity import DefaultAzureCredential
from azure.identity import ManagedIdentityCredential 
import xml.etree.ElementTree as ET
from azure.mgmt.storage import StorageManagementClient
from azure.storage.blob import  BlobServiceClient, RetentionPolicy
from azure.core.exceptions import ResourceExistsError, ServiceRequestError, HttpResponseError

from typing import List

from ec2init import Procedure
from ec2init.utils.bucket_configuration import BucketConfiguration
from pathlib import Path
from ec2init.utils.exceptions import ConfigException

logger = logging.getLogger(__name__)

class CheckAzureContainerPermissions(Procedure):
    def __init__(self, name: str, check_file: Path, container_configuration: BucketConfiguration, region: str, test_obj: Path, 
                 storage_account: str, client_id:str, 
                 deleted_blocks_retention_days: int, is_backup_or_upgrade: bool, azure_subscription_id: str, azure_resource_group: str) -> None:
        self.name = name
        self.check_file = check_file
        self.container_configuration = container_configuration
        self.test_obj = test_obj
        self.region = region
        self.storage_account = storage_account
        self.client_id = client_id
        self.deleted_blocks_retention_days=deleted_blocks_retention_days
        self.is_backup_or_upgrade = is_backup_or_upgrade
        self.azure_subscription_id = azure_subscription_id
        self.azure_resource_group = azure_resource_group

    def pre(self) -> None:
        pass

    PERMISSION_ERROR_MSG = "We were not able to {action} the storage account you provided. Make sure that your managed identity has the following permission and retry: {permission}. " + \
                    "See our documentation for more details: https://docs.hopsworks.ai/latest/setup_installation/azure/getting_started/#step-4-create-a-managed-identity"
    def execute(self) -> List[Procedure]:
        logger.info("Checking if this instance has proper permissions to access Azure Container: https://%s.blob.core.windows.net/%s",
                    self.storage_account, self.container_configuration.name)
        try :
            # Connect using managed identities
            self.friendly_name = "Unable to get Credentials for Azure"
            if self.client_id:
                logger.info("Using managed identity credentials with client ID: %s", self.client_id)
                credential = ManagedIdentityCredential(client_id=self.client_id)
            else:
                credential = DefaultAzureCredential()

            # Test connection to blob store
            self.friendly_name = "Unable to connect to Azure Blob Storage"
            container_uri = f"https://{self.storage_account}.blob.core.windows.net"
            blob_service_client = BlobServiceClient(account_url=container_uri, credential=credential)


            # Test container exists or create one
            self.friendly_name = "Checking Azure Blob Storage permissions"

            try:
                containers = blob_service_client.list_containers()
                for c in containers: 
                    # we need to try to read the containers to get the error
                    break
            except HttpResponseError as ex:
                logger.error("Failed list the containers", exc_info=ex)
                if ex.status_code == 401:
                    raise ConfigException(f"We were not able to list the container in the storage account you provided ({self.storage_account}). Make sure that the storage account exist and has the correct name.")
                if ex.status_code == 403:
                    raise ConfigException(self.PERMISSION_ERROR_MSG.format(action="list the container in", permission="Microsoft.Storage/storageAccounts/blobServices/containers/read"))
                raise

            try:
                blob_service_client.create_container(self.container_configuration.name)
            except ResourceExistsError as ex:
                logger.info("Container already exists.")
            except HttpResponseError as ex:
                logger.error("Failed to create the container", exc_info=ex)
                if ex.status_code == 403:
                    raise ConfigException(self.PERMISSION_ERROR_MSG.format(action="create a container in", permission="Microsoft.Storage/storageAccounts/blobServices/containers/write"))
                raise

            # Check Soft Delete support
            try:
                self.checkSoftDeletes(blob_service_client)
            except HttpResponseError as ex:
                logger.error("Failed to check soft delete", exc_info=ex)
                if ex.status_code == 403:
                    raise ConfigException(self.PERMISSION_ERROR_MSG.format(action="check if soft delete is enabled on", permission="Microsoft.Storage/storageAccounts/blobServices/read"))
                raise

            # Enable soft deletes 
            if self.deleted_blocks_retention_days > 0:
                try:
                    self.enableSoftDeletes(blob_service_client)
                except HttpResponseError as ex:
                    logger.error("Failed to enable soft delete", exc_info=ex)
                    if ex.status_code == 403:
                        raise ConfigException(self.PERMISSION_ERROR_MSG.format(action="enable soft delete on", permission="Microsoft.Storage/storageAccounts/blobServices/write"))
                    raise

                try:
                    storage_client  = StorageManagementClient(credential=credential, subscription_id=self.azure_subscription_id)
                    storage_client.storage_accounts.list_keys(resource_group_name=self.azure_resource_group, account_name=self.storage_account)
                except HttpResponseError as ex:
                    logger.error("Failed to list storage keys", exc_info=ex)
                    if ex.status_code == 403:
                        raise ConfigException(self.PERMISSION_ERROR_MSG.format(action="list storage access keys on", permission="Microsoft.Storage/storageAccounts/listKeys/action"))
                    raise

            # Test listing. Make sure that container is empty
            try:
                container_client = blob_service_client.get_container_client(self.container_configuration.name)
                blobs = container_client.list_blobs()
                count = 0
                for _ in blobs:
                    count = count + 1
                    break
            except HttpResponseError as ex:
                logger.error("Failed to list the blobs", exc_info=ex)
                if ex.status_code == 403:
                    raise ConfigException(self.PERMISSION_ERROR_MSG.format(action="list blobs in", permission="Microsoft.Storage/storageAccounts/blobServices/containers/blobs/read"))
                raise
            
            if not self.is_backup_or_upgrade:
                if count > 0 :
                    raise ConfigException("The storage container you provided is not empty. Please empty the container and retry or create a new cluster with a different container.") 

            # Test uploading to the blob store
            try:
                blob_client = blob_service_client.get_blob_client(container=self.container_configuration.name, blob="test-obj")
                logger.info("Uploading test obj to Azure Storage ")
                with open(self.test_obj, "rb") as data:
                    blob_client.upload_blob(data)
            except HttpResponseError as ex:
                logger.error("Failed to upload a blob", exc_info=ex)
                if ex.status_code == 403:
                    raise ConfigException(self.PERMISSION_ERROR_MSG.format(action="write to a container in", permission="Microsoft.Storage/storageAccounts/blobServices/containers/blobs/write"))
                raise

            # Test downloading from the blob store
            try:
                download_file_path = str(self.test_obj) + ".new" 
                logger.info("Downloading blob to %s" + download_file_path)
                with open(download_file_path, "wb") as download_file:
                    download_file.write(blob_client.download_blob().readall())
            except HttpResponseError as ex:
                logger.error("Failed to read a blob", exc_info=ex)
                if ex.status_code == 403:
                    raise ConfigException(self.PERMISSION_ERROR_MSG.format(action="read from a container in", permission="Microsoft.Storage/storageAccounts/blobServices/containers/blobs/read"))
                raise

            # Test deleting an object
            try:
                blob_client = blob_service_client.get_blob_client(container=self.container_configuration.name, blob="test-obj")
                blob_client.delete_blob()
            except HttpResponseError as ex:
                logger.error("Failed to delete a blob", exc_info=ex)
                if ex.status_code == 403:
                    raise ConfigException(self.PERMISSION_ERROR_MSG.format(action="delete from a container in", permission="Microsoft.Storage/storageAccounts/blobServices/containers/blobs/delete"))
                raise

            #Every thing is OK. 
            if self.check_file is not None:
                self.check_file.touch()

        except ServiceRequestError as ex:
            logger.error("Failed to check container access (ServiceRequestError)", exc_info=ex)
            if "Name or service not known" in ex.message:
                self.raise_container_exception(f"The storage account {self.storage_account} could not be found, verify {self.storage_account} is identical to the storage account you created.\n"+
                    "For information how to setup the storage account, see https://docs.hopsworks.ai/latest/hopsworksai/azure/getting_started/#step-2-create-a-storage", ex)
            self.raise_container_exception(str(ex), ex)
        except ConfigException:
            raise
        except Exception as ex:
            logger.error("Failed to check container access", exc_info=ex)
            self.raise_container_exception(str(ex), ex)

        return []

    def raise_container_exception(self, reason: str, ex: Exception) -> None:
        logger.warn("%s - Could not execute Azure container checks. Container: %s %s", self.name, self.container_configuration.name, ex)
        self.friendly_name = str(self.friendly_name) + "  Reason: " + reason
        raise Exception(self.friendly_name)

    def rollback(self) -> None:
        pass

    def checkSoftDeletes(self, blob_service_client: BlobServiceClient) -> None:
        logger.info("Reading soft delete retention days")
        blob_service_client.get_service_properties()["delete_retention_policy"]

    def enableSoftDeletes(self, blob_service_client: BlobServiceClient) -> None:
        logger.info("Getting soft delete retention days")
        retention_policy = blob_service_client.get_service_properties()["delete_retention_policy"]
        retention_policy_days = 0 if retention_policy.days is None else int(retention_policy.days)
        if retention_policy.enabled and retention_policy_days > self.deleted_blocks_retention_days : 
            logger.warn(f"letting the retention day to its current value {retention_policy_days} which is higher than the new one {self.deleted_blocks_retention_days}")
            return

        logger.info("Setting soft delete retention days to %s", str(self.deleted_blocks_retention_days))
        delete_retention_policy = RetentionPolicy(enabled=True, days=self.deleted_blocks_retention_days)
        blob_service_client.set_service_properties(delete_retention_policy=delete_retention_policy)


