import logging

from ec2init import Procedure
from pathlib import Path
from google.cloud import storage # type: ignore
from ec2init.utils.bucket_configuration import BucketConfiguration
from google.api_core import exceptions

from typing import List

from ec2init.utils.exceptions import ConfigException
from ec2init.utils.cloud import GCP
from configparser import ConfigParser
from ec2init.utils.http import HttpUtils
from requests.exceptions import HTTPError
import xml.etree.ElementTree as ET

logger = logging.getLogger(__name__)

class CheckGCPBucketPermissions(Procedure):
    def __init__(self, name: str, check_file: Path, bucket_configuration: BucketConfiguration, zone: str, deleted_blocks_retention_days: int, is_backup_or_upgrade: bool, test_obj: Path, config: ConfigParser,
    service_account: str) -> None:
        self.name = name
        self.check_file = check_file
        self.bucket_configuration = bucket_configuration
        self.zone = zone
        self.deleted_blocks_retention_days = deleted_blocks_retention_days
        self.is_backup_or_upgrade = is_backup_or_upgrade
        self.test_obj = test_obj
        self.config = config
        self.service_account = service_account

    def pre(self) -> None:
        if self.report_callback is not None:
            try:
                self.report_callback.execute()
            except Exception as ex:
                logger.warning("%s - Could not execute pre-execute: %s", self.name, ex)
    
    def rollback(self) -> None:
        pass

    BUCKET_PATH = 'http://{bucket}.storage.googleapis.com/{object_name}'
    def start_multipart_upload(self, bucket: str, object_name: str, access_token: str) -> str:
        url = self.BUCKET_PATH.format(bucket = bucket, object_name = object_name)
        headers = {
            'Authorization': 'Bearer {}'.format(access_token)
        }

        params = {
            'uploads':'true'
        }

        body = {
            "test":"test"
        }
        response = HttpUtils.do_api_call(
            self.config, url, "POST", params=params, data=body, headers=headers, max_tries=0)
        tree = ET.fromstring(response['response'])

        upload_id = tree.find('.//{http://s3.amazonaws.com/doc/2006-03-01/}UploadId')
        if upload_id is not None and upload_id.text is not None:
            return upload_id.text
        raise Exception(f"Failed to start a multipart upload. The response did not contain the expected uploade Id. Contact support. response: {str(response)}")

    def list_multipart_upload(self, bucket: str, object_name : str, access_token: str):
        url = self.BUCKET_PATH.format(bucket = bucket, object_name = "")
        headers = {
            'Authorization': 'Bearer {}'.format(access_token)
        }

        params = {
            'uploads':'true',
            'prefix': object_name,
        }

        HttpUtils.do_api_call(
            self.config, url, "GET", params=params, headers=headers, max_tries=0)

    def list_multipart_upload_parts(self, bucket: str, object_name : str, access_token: str, upload_id:str):
        url = self.BUCKET_PATH.format(bucket = bucket, object_name = object_name)
        headers = {
            'Authorization': 'Bearer {}'.format(access_token)
        }

        params = {
            'uploadId':upload_id
        }

        HttpUtils.do_api_call(
            self.config, url, "GET", params=params, headers=headers, max_tries=0)

    def abort_multipart_upload(self, bucket: str, object_name : str, access_token: str, upload_id:str):
        url = self.BUCKET_PATH.format(bucket = bucket, object_name = object_name)
        headers = {
            'Authorization': 'Bearer {}'.format(access_token)
        }

        params = {
            'uploadId':upload_id
        }

        HttpUtils.do_api_call(
            self.config, url, "DELETE", params=params, headers=headers, max_tries=0)

    PERMISSION_ERROR_MSG = "We were not able to {action} the storage bucket you provided. Make sure that your Instance service account has the {access} access and retry. " + \
                    "See our documentation for more details: https://docs.hopsworks.ai/latest/setup_installation/gcp/getting_started/#step-3-creating-a-service-account-for-your-cluster-instances"
    def execute(self) -> List[Procedure]:
        logger.info("Checking if this instance has proper permissions to access the Bucket . Retention days "+ str(self.deleted_blocks_retention_days))
        try :
            try:
                storage_client = storage.Client()
                bucket = storage_client.get_bucket(self.bucket_configuration.name)
            except exceptions.Forbidden as ex:
                    logger.error("Failed to get the bucket", exc_info=ex)
                    raise ConfigException(f"We were not able to get the storage bucket you provided ({self.bucket_configuration.name}). Make sure that you provided the proper bucket name and that " +
                    "your Instance service account has the storage.buckets.get access and retry. " + 
                    "See our documentation for more details: https://docs.hopsworks.ai/latest/setup_installation/gcp/getting_started/#step-3-creating-a-service-account-for-your-cluster-instances") 

            # Enable versioning
            if self.deleted_blocks_retention_days > 0:
                try:
                    bucket.versioning_enabled = True
                    bucket.add_lifecycle_delete_rule(is_live=False, days_since_noncurrent_time=self.deleted_blocks_retention_days)
                    bucket.patch()
                except exceptions.Forbidden as ex:
                    logger.error("Failed to add lifcycle delete rules", exc_info=ex)
                    raise ConfigException(self.PERMISSION_ERROR_MSG.format(action="set the retention period on", access="storage.buckets.update")) 
            
            # Test listing. Make sure that container is empty
            try:
                blobs = bucket.list_blobs()
                count = 0
                for _ in blobs:
                    count = count + 1
                    break
            except exceptions.Forbidden as ex:
                    logger.error("Failed to list bucket content", exc_info=ex)
                    raise ConfigException(self.PERMISSION_ERROR_MSG.format(action="list the content of", access="storage.objects.list")) 
            
            if not self.is_backup_or_upgrade:
                if count > 0 :
                    raise ConfigException("The bucket you provided is not empty. Please empty the bucket and retry or create a new cluster with a different bucket.") 
            
            # Test uploading to the blob store
            try:
                logger.info("Uploading test obj to the bucket ")
                blob = bucket.blob("test-obj")
                blob.upload_from_filename(str(self.test_obj))
            except exceptions.Forbidden as ex:
                logger.error("Failed to write to the bucket", exc_info=ex)
                raise ConfigException(self.PERMISSION_ERROR_MSG.format(action="write to", access="storage.objects.create")) 

            access_token =GCP.get_access_token(self.config, self.service_account)
            try:
                upload_id = self.start_multipart_upload(bucket.name, "test_multipart", access_token)
            except HTTPError as ex:
                logger.error("Error starting the multipart upload ", exc_info=ex)
                if ex.response.status_code == 403:
                    raise ConfigException(self.PERMISSION_ERROR_MSG.format(action="start a multipart upload on", access="storage.multipartUploads.create")) 
                raise

            try:
                self.list_multipart_upload(bucket.name, "test_multipart", access_token)
            except HTTPError as ex:
                logger.error("Error listing the multipart upload", exc_info=ex)
                if ex.response.status_code == 403:
                    raise ConfigException(self.PERMISSION_ERROR_MSG.format(action="list the multipart uploads on", access="storage.multipartUploads.list")) 
                raise
            
            try:
                self.list_multipart_upload_parts(bucket.name, "test_multipart", access_token, upload_id)
            except HTTPError as ex:
                logger.error("Error listing the multipart upload parts ", exc_info=ex)
                if ex.response.status_code == 403:
                    raise ConfigException(self.PERMISSION_ERROR_MSG.format(action="list the multipart uploads parts on", access="storage.multipartUploads.listParts")) 
                raise

            try:
                self.abort_multipart_upload(bucket.name, "test_multipart", access_token, upload_id)
            except HTTPError as ex:
                logger.error("Error aborting the multipart upload ", exc_info=ex)
                if ex.response.status_code == 403:
                    raise ConfigException(self.PERMISSION_ERROR_MSG.format(action="abort a multipart uploads on", access="storage.multipartUploads.abort")) 
                raise

            # Test downloading from the blob store
            try:
                download_file_path = str(self.test_obj) + ".new" 
                logger.info("Downloading blob to %s" + download_file_path)
                storage_client.close()
                storage_client = storage.Client()
                bucket = storage_client.get_bucket(self.bucket_configuration.name)
                blob = bucket.get_blob(blob.name)
                blob.download_to_filename(download_file_path)
            except exceptions.Forbidden as ex:
                logger.error("Failed to get from the bucket", exc_info=ex)
                raise ConfigException(self.PERMISSION_ERROR_MSG.format(action="get an object from", access="storage.objects.get")) 
            

            # Test deleting an object
            try:
                bucket.delete_blob("test-obj")
            except exceptions.Forbidden as ex:
                logger.error("Failed to delete from the bucket", exc_info=ex)
                raise ConfigException(self.PERMISSION_ERROR_MSG.format(action="delete an object from", access="storage.objects.delete access")) 

            #Every thing is OK. 
            if self.check_file is not None:
                self.check_file.touch()

            
        except Exception as ex:
            logger.warning("%s - Could not execute bucket checks. Bucket: %s %s", self.name, self.bucket_configuration.name, ex)
            raise

        return []
