import logging
import pathlib
from configparser import ConfigParser

from typing import List

from ec2init import Procedure
from ec2init.utils.system import SystemUtils
import pathlib
import os 
import docker 
import yaml 
from ec2init.plans.init.setup_ecr_docker_images import SetupGCRDockerImages

logger = logging.getLogger(__name__)

# Workaround to pull smarter device manager plugin image on private GKE clusters since we won't be able to pull from public registries 
# It might be better to do that outside of ec2-init but i'm gonna leave it here for now untill we have terraform integration in place and then we can decide on this 
class ConfigureSmartDeviceManagerOnGCR(Procedure):
    def __init__(self, name: str, config: ConfigParser, check_file: pathlib.Path, gcr_user_registry_url: str, instance_id: str) -> None:
        self.name = name
        self.config = config
        self.check_file = check_file
        self.gcr_user_registry_url = gcr_user_registry_url
        self.instance_id = instance_id

    def execute(self) -> List[Procedure]:
        installation_dir = self.config.get('general', 'installation_dir', fallback='/root/ec2init')
        config_yml = f"{installation_dir}/smarter-device-manager-plugin.yml"
        if os.path.exists(config_yml):
            self.configure(config_yml)
        else:
            logger.info("%s - skip configuring smarter-device-manager-plugin", self.name)

        if self.check_file is not None:
            self.check_file.touch()
            
        return []


    def configure(self, config_yml: str):                
        try:
            with open(config_yml, 'r') as f:
                config_data = [data for data in yaml.load_all(f, Loader=yaml.loader.SafeLoader)]
        
            docker_image = config_data[1]['spec']['template']['spec']['containers'][0]['image']
            logger.info("%s - current docker_image for smarter-device-manager-plugin is %s", self.name, docker_image)
            docker_client = docker.from_env(timeout=360)
            image = docker_client.images.pull(docker_image)
            new_image_tag = f"{SetupGCRDockerImages.get_user_artifact_registry(self.gcr_user_registry_url, self.instance_id)}/smarter/smarter-device-manager:{docker_image.split(':')[1]}"
            image.tag(new_image_tag)
            docker_client.images.push(new_image_tag)
            docker_client.images.remove(docker_image)
            docker_client.images.remove(new_image_tag)

            config_data[1]['spec']['template']['spec']['containers'][0]['image'] = new_image_tag
            
            with open(config_yml, 'w') as f:
                yaml.dump_all(config_data, f, sort_keys=False, default_flow_style=False)

            SystemUtils.execute_command_capture_with_print_error(["kubectl", "apply", "-f", config_yml], timeout=120, env={"KUBECONFIG" : "/root/.kube/config", "PATH": "/usr/local/bin:/usr/bin"})
        except Exception as ex:
            logger.warning("%s - failed to configure smarter-device-manager-plugin", exc_info=ex)

    def rollback(self) -> None:
        pass
