import logging
import time

import dns.resolver
from typing import List
from configparser import ConfigParser

from ec2init import Procedure
from ec2init.utils.system import SystemUtils
from ec2init.utils.dns import DnsUtils
from ec2init.utils.cloud import get_cloud_provider

logger = logging.getLogger(__name__)


class IssueCertificate(Procedure):
    def __init__(self, name: str, config: ConfigParser, instance_id: str, cloud_provider: str, cluster_domain_prefix:str) -> None:
        self.name = name
        self.config = config
        self.instance_id = instance_id
        self.cloud_provider = cloud_provider
        self.cluster_domain_prefix = cluster_domain_prefix

    def pre(self) -> None:
        if self.report_callback is not None:
            try:
                self.report_callback.execute()
            except Exception as ex:
                logger.warning(
                    "%s - Could not execute pre-execute: %s", self.name, ex)

    def execute(self) -> List[Procedure]:
        # First wait long enough for the domain name to be visible
        # Let's Encrypt asks the Authoritative Server for the domain
        # so if we can resolve it from the instance, Let's Encrypt can
        # resolve it too.
        domain_name = get_cloud_provider(self.cloud_provider)\
            .get_public_domain_name(self.config, self.instance_id, self.cluster_domain_prefix)
        DnsUtils.get_ip_from_name(domain_name)

        # Now it's time to call certbot
        command = []
        command.append("certbot")
        command.append("certonly")
        command.append("--standalone")
        command.append("--register-unsafely-without-email")
        command.append("--agree-tos")
        command.append("--non-interactive")
        command.append("--key-type")
        command.append("rsa")
        # Certificates issued in staging are not trusted by browsers
        if SystemUtils.is_staging():
            command.append("--staging")
        command.append("-d")
        command.append(domain_name)
        logger.info("%s - Calling certbot to issue certificate", self.name)
        SystemUtils.execute_command(command, timeout=180)
        return []

    def _get_ip(self, dns_response: dns.resolver.Answer) -> str:
        ips = [a.address for a in dns_response]
        return ips[0]

    def rollback(self) -> None:
        pass
