import configparser
import logging
import pathlib

from urllib.parse import urlparse

from typing import List, Set, Dict, cast

from ec2init import Procedure
from ec2init.utils.common import CommonUtils
from ec2init.utils.constants import Url
from ec2init.utils.system import SystemUtils
from ec2init.utils.http import HttpUtils

logger = logging.getLogger(__name__)

class RevokeCertificates(Procedure):
    def __init__(self, name: str, config: configparser.ConfigParser, check_file: pathlib.Path, subjects: List[str]) -> None:
        self.name = name
        self.config = config
        self.check_file = check_file
        self.subjects = subjects

    def execute(self) -> List[Procedure]:
        logger.info(f"{self.name} - Starting execution")
        subjects_to_revoke: Set[str] = set()
        for s in self.subjects:
            self._find_valid_certificate_subjects(subjects_to_revoke, s)
        
        if not subjects_to_revoke:
            logger.info(f"{self.name} - Did not find any certificates to revoke, skipping")
            self._done()
            return []
            
        agent_username, agent_password = CommonUtils._get_agent_credentials(self.config)
        if not agent_username or not agent_password:
            raise Exception("Agent username or password is not set")

        method, uri = Url.HOPSWORKS_LOGIN
        uri = urlparse(uri)._replace(netloc="127.0.0.1").geturl()
        data = {
            'email': agent_username,
            'password': agent_password
        }
        headers = {
            'Content-Type': 'application/x-www-form-urlencoded',
        }
        logger.debug("%s - Logging in to Hopsworks", self.name)

        response: Dict[str, str] = HttpUtils.do_plain_call(self.config, uri, method, data=data, headers=headers, verify=False)
        response_headers: Dict[str, str] = cast(Dict[str, str], response['headers'])
        token = response_headers['Authorization']

        headers['Content-Type'] = "application/json"
        headers['Authorization'] = token
        method, uri = Url.HOST_CA_REVOKE
        
        for subject in subjects_to_revoke:
            logger.info(f"{self.name} - Revoking certificate with subject: {subject}")
            self._revoke_certificate(uri, method, headers, subject)
            logger.info(f"{self.name} - Revoked certificate with subject: {subject}")

        self._done()

        return []

    def _find_valid_certificate_subjects(self, subjects_to_revoke: Set[str], partial_subject: str) -> None:
        rows = SystemUtils.execute_mysql_statement(f"SELECT subject FROM pki_certificate WHERE subject like '%{partial_subject}%' AND status=0")
        if not rows:
            return

        for r in rows:
            sub = r[0]
            subjects_to_revoke.add(sub.strip())

    def _revoke_certificate(self, url: str, method: str, auth_headers: Dict[str, str], subject: str) -> None:
        params = {
            "certId": f"{subject}",
            "exact": "true"
        }
        HttpUtils.do_api_call(self.config, url, method, params=params, headers=auth_headers, verify=False, timeout=120)
        
    def _done(self) -> None:
        if self.check_file:
            self.check_file.touch()

    def rollback(self) -> None:
        pass