import logging
import boto3
import docker
import time
from ec2init.utils.systemd import SystemdUtils 
import kubernetes
import os 
import shutil
import json 

from crontab import CronTab
from pathlib import Path
from typing import Dict, cast
from mypy_boto3_ecr import ECRClient
from mypy_boto3_ecr.type_defs import TagTypeDef

from ec2init import MultiThreadProcedure
from ec2init.utils.system import SystemUtils
from ec2init.utils.offline_backup_information import OffLineBackupInformation
from ec2init.plans.init.load_docker import LoadDocker
from ec2init.plans.init.setup_eks import HOPSWORKS_NAMESPACE
from ec2init.utils.docker import DockerImageName;
from ec2init.utils.services import Service
from ec2init.utils.exceptions import ConfigException
from abc import abstractmethod
import subprocess
from ec2init.utils.constants import Scripts
from google.cloud import artifactregistry_v1
import grpc 

logger = logging.getLogger(__name__)

class SetupDockerImages(MultiThreadProcedure):
    def __init__(self, name: str, check_file: Path, instance_id: str, region: str, ecr_user_registry_url: str, ecr_main_registry_url: str,
     ecr_repos: Dict[DockerImageName, str], hopsworks_version: str, node_type: str, eks_cluster_enabled: bool, backup_info: OffLineBackupInformation,
     default_registry_url: str, python_image_name: str, is_upgrade :bool) -> None:
        self.name = name
        self.check_file = check_file
        self.instance_id = instance_id
        self.region = region
        self.hopsworks_version = hopsworks_version
        self.python_base_image_tag = f"{python_image_name}_{ecr_repos[DockerImageName.Base]}"
        self.tasks = [(repo.value, version) for repo, version in ecr_repos.items()]
        self.tasks.append((DockerImageName.Base.value, self.python_base_image_tag))
        self.max_workers = min(6, len(self.tasks))
        self._start_time = time.time()
        self.ecr_user_registry_url =  ecr_user_registry_url
        self.ecr_main_registry_url = ecr_main_registry_url
        self.node_type = node_type
        self.eks_cluster_enabled = eks_cluster_enabled
        self.backup_info = backup_info
        # If we're executing this code, the user_registry_url is there so it's safe to cast it from Optional
        self.backup_ecr_user_registry_url = str(backup_info.user_registry_url)
        self.default_registry_url = default_registry_url
        self.is_upgrade = is_upgrade

    def pre(self) -> None:
        if os.path.isfile("/srv/hops/hadoop/etc/hadoop/container-executor.cfg"):
            SystemUtils.execute_command(["sh", "-c", "sed -i 's/docker.trusted.registries=" + self.default_registry_url + "/docker.trusted.registries=" + self.ecr_user_registry_url.replace('/', '\/') + "/g' /srv/hops/hadoop/etc/hadoop/container-executor.cfg"])

        if self.node_type == "master":
            # update DB
            SystemUtils.execute_command([Scripts.MYSQL_CLIENT, "hopsworks", "-e", "REPLACE INTO variables (id, value) VALUES ('docker_namespace', '"+ self.instance_id +"')"])

            # registr ECR registry to consul
            registry_consul_file = "/srv/hops/consul/consul.d/registry-consul.hcl"
            registry_consul_conf = ['services = [',
                                    '{',
                                    'id = "registry"',
                                    'name = "registry"',
                                    'port = 443',
                                    'address = "' + self.ecr_user_registry_url + '"',
                                    '}',
                                    ']']
            
            if os.path.exists(registry_consul_file):
                os.remove(registry_consul_file)

            with open(registry_consul_file, 'w') as out_file:
                out_file.writelines(registry_consul_conf)

            os.chmod(registry_consul_file, 0o500)
            shutil.chown(registry_consul_file, group="consul", user="consul")

    def task_function(self, task):
        repo_name, image_tag = task

        user_repo_name = self.get_repo_name(self.instance_id, repo_name)
        t1 = time.time()
        docker_client = docker.from_env(timeout=360)
        
        if not self.is_upgrade and OffLineBackupInformation.is_backup(self.backup_info) and self.backup_ecr_user_registry_url and self.node_type == "master":
            self.restore_from_backup(docker_client, repo_name, user_repo_name, image_tag)
        else:
            self.pull_and_push_image(docker_client, repo_name, user_repo_name, image_tag)
        
        self.clean_up_service_images(docker_client, repo_name, image_tag)

        self.clean_up_old_docker_registry(docker_client)

        elapsed_time = time.time() - t1
        logger.info("%s took %d seconds to setup image (%s:%s)", self.name, elapsed_time, repo_name, image_tag)

    def clean_up_service_images(self, docker_client, repo_name :str, image_tag :str):
        docker_repo_name = DockerImageName(repo_name)
        ecr_user_registry_url = self.ecr_user_registry_url.replace('/', '\/')
        if docker_repo_name in [DockerImageName.OnlineFS, DockerImageName.Airflow, DockerImageName.FlyingDuck]:
            if docker_repo_name == DockerImageName.OnlineFS and SystemdUtils.is_service_installed(Service.Onlinefs.value):
                SystemUtils.execute_command_with_retry(["systemctl", "stop", Service.Onlinefs.value], timeout=60)
                SystemUtils.execute_command(["sh", "-c", f"sed -i 's/{self.default_registry_url}/{ecr_user_registry_url}\/{self.instance_id}/g' /lib/systemd/system/{Service.Onlinefs.value}.service"])
            elif docker_repo_name == DockerImageName.Airflow:
                for service in [Service.AirflowScheduler, Service.AirflowWebserver]:
                    if SystemdUtils.is_service_installed(service.value):
                        SystemUtils.execute_command_with_retry(["systemctl", "stop", service.value], timeout=60)
                        SystemUtils.execute_command(["sh", "-c", f"sed -i 's/{self.default_registry_url}/{ecr_user_registry_url}\/{self.instance_id}/g' /lib/systemd/system/{service.value}.service"])
            elif docker_repo_name == DockerImageName.FlyingDuck and SystemdUtils.is_service_installed(Service.FlyingDuck.value):
                SystemUtils.execute_command_with_retry(["systemctl", "stop", Service.FlyingDuck.value], timeout=60)
                SystemUtils.execute_command(["sh", "-c", f"sed -i 's/{docker_repo_name.value}:{image_tag}/{ecr_user_registry_url}\/{self.instance_id}\/{docker_repo_name.value}:{image_tag}/g' /lib/systemd/system/{Service.FlyingDuck.value}.service"])
            
            SystemUtils.execute_command(["systemctl", "daemon-reload"])

            self._remove_image(docker_client, self.get_repo_image(self.default_registry_url,docker_repo_name.value,image_tag))
            self._remove_image(docker_client, self.get_repo_image("docker.hops.works", docker_repo_name.value, image_tag))
            self._remove_image(docker_client, f"{docker_repo_name.value}:{image_tag}")
    
    def _remove_image(self, docker_client, image: str):
        SetupDockerImages.remove_image_int(docker_client, image)
    
    @staticmethod
    def remove_image(image: str) -> None:
        SetupDockerImages.remove_image_int(docker.from_env(timeout=360), image)

    @staticmethod
    def remove_image_int(docker_client, image: str) -> None:
        try:
            docker_client.images.remove(image=image, force=True)
        except docker.errors.ImageNotFound as ex:
            logger.info("%s - old %s image is not found ex: %s", SetupDockerImages.__name__, image, str(ex))

    def clean_up_old_docker_registry(self, docker_client):
        try:
            container = docker_client.containers.get("registry")
            container.stop()
            self._remove_image(docker_client, "registry:latest")
        except docker.errors.NotFound as ex:
            logger.info("%s local docker registry do not exists", self.name)

    def pull_and_push_image(self, docker_client, main_repo_name: str, user_repo_name: str, image_tag: str):
        if self.node_type == "rondb_all-in-one" or self.node_type=="ndb_mgm":
            self.pull_and_push_image_test(docker_client, main_repo_name, user_repo_name, image_tag)
        elif self.node_type == "master":
            self.pull_and_push_image_on_master(docker_client, main_repo_name, user_repo_name, image_tag)
        elif self.node_type == "worker":
            self.pull_images_on_worker(docker_client, main_repo_name, user_repo_name, image_tag)
        elif self.node_type == "mysqld":
            self.pull_and_push_images_on_mysqld(docker_client, main_repo_name, user_repo_name, image_tag)

    @abstractmethod
    def get_pull_token_solution(self):
        pass

    @abstractmethod
    def get_pull_auth_solution(self, main_repo: bool):
        pass

    @abstractmethod
    def handle_docker_push_error(self, line: dict):
        pass
    
    @abstractmethod
    def test_list_delete(self, image_tag: str):
        pass

    def get_repo_name(self, registry: str, repo_name: str) -> str :
        return f"{registry}/{repo_name}"

    def get_repo_image(self, registry: str, repo_name: str, image_tag: str) -> str :
        return f"{self.get_repo_name(registry, repo_name)}:{image_tag}"

    def pull_and_push_image_test(self, docker_client, main_repo_name: str, user_repo_name: str, image_tag: str):
        # pulling and pushing only the filebeat image because it is "small" so it should not take too long
        if DockerImageName.Filebeat == DockerImageName(main_repo_name):
            logger.info("%s - Pulling image from hopsworks.ai main repo - %s:%s", self.name, main_repo_name, image_tag)

            try:
                main_repo_image = self.get_repo_image(self.ecr_main_registry_url , main_repo_name, image_tag)
                image = docker_client.images.pull(main_repo_image)
            except docker.errors.APIError as ex:
                logger.error("Error pulling the image from the main repo", exc_info=ex)
                if "no basic auth credentials" in ex.explanation: #sadly the error number is a generic 500 so we need to parse
                    raise ConfigException("Unable to get the auth credentials for the docker repository. " +
                    self.get_pull_token_solution())
                if self.docker_pull_failure_condition(ex):
                    raise ConfigException("Unable to pull the docker image from the source docker repository. " +
                    self.get_pull_auth_solution(True))
                raise
            

            user_repo_image = self.get_repo_image(self.ecr_user_registry_url, user_repo_name, image_tag)
            image.tag(user_repo_image)
            docker_client.images.remove(main_repo_image)

            logger.info("%s - pushing image to user repo - %s:%s", self.name, main_repo_name, image_tag)
            
            self.push_docker_image(docker_client, user_repo_image)

            #delete local image and pull again to make sure that we have the permissions to pull
            docker_client.images.remove(user_repo_image)
            try:
                image = docker_client.images.pull(user_repo_image)
            except docker.errors.APIError as ex:
                logger.error("Error pulling the image from the user's repo", exc_info=ex)
                if self.docker_pull_failure_condition(ex):
                    raise ConfigException("Unable to pull the docker image from your docker repository. " +
                    self.get_pull_auth_solution(False))
                raise
            
            docker_client.images.remove(user_repo_image)
            self.test_list_delete(image_tag)

    def docker_pull_failure_condition(self, ex : docker.errors.APIError) -> bool:
        #we are sure that the registry exist, so if we do not find it it is because we don't have the permissions for it
        return ex.status_code == 404
     
    def pull_and_push_image_on_master(self, docker_client, main_repo_name: str, user_repo_name: str, image_tag: str):
        if DockerImageName(main_repo_name) not in [DockerImageName.Base, DockerImageName.Filebeat, DockerImageName.Airflow, DockerImageName.Git, DockerImageName.OnlineFS, DockerImageName.TestConnector]:
            logger.info("%s - skip pulling image %s on head node", self.name, main_repo_name)
            return
        
        logger.info("%s - Pulling image from hopsworks.ai main repo - %s:%s", self.name, main_repo_name, image_tag)

        main_repo_image = self.get_repo_image(self.ecr_main_registry_url, main_repo_name, image_tag)
        image = docker_client.images.pull(main_repo_image)
        user_repo_image = self.ecr_user_registry_url + "/" + user_repo_name + ":" + image_tag

        image.tag(user_repo_image)
        docker_client.images.remove(main_repo_image)

        self.push_docker_image(docker_client, user_repo_image)
        
        if self.eks_cluster_enabled and image_tag == self.python_base_image_tag:
            self.eks_prepull_images(user_repo_image)

    def pull_images_on_worker(self, docker_client, main_repo_name: str, user_repo_name: str, image_tag: str):
        if DockerImageName(main_repo_name) not in [DockerImageName.Base, DockerImageName.Filebeat]:
            logger.info("%s - skip pulling image %s on head node", self.name, main_repo_name)
            return

        logger.info("%s - Pulling image from cluster's repo - %s:%s", self.name, user_repo_name, image_tag)
        user_repo_image = self.get_repo_image(self.ecr_user_registry_url, user_repo_name , image_tag)
        SystemUtils.retry_operation_on_exception(lambda : docker_client.images.pull(user_repo_image), retries=15, waitTime=120)
    
    def pull_and_push_images_on_mysqld(self, docker_client, main_repo_name: str, user_repo_name: str, image_tag: str):
        if main_repo_name != DockerImageName.FlyingDuck.value:
            logger.info("%s - skip pulling image %s on mysqld node", self.name, main_repo_name)
            return
        
        user_repo_image = self.get_repo_image(self.ecr_user_registry_url, user_repo_name , image_tag)

        try:
            logger.info("%s - Pulling image from user repo - %s", self.name, user_repo_image)
            docker_client.images.pull(user_repo_image)
            return 
        except docker.errors.NotFound as ex:
            logger.info("%s - %s is not found yet so we should pull it from main and set it up first", self.name, main_repo_name)


        logger.info("%s - Pulling image from hopsworks.ai main repo - %s:%s", self.name, main_repo_name, image_tag)

        main_repo_image = self.get_repo_image(self.ecr_main_registry_url, main_repo_name, image_tag)
        image = docker_client.images.pull(main_repo_image)

        image.tag(user_repo_image)
        docker_client.images.remove(main_repo_image)

        self.push_docker_image(docker_client, user_repo_image)

    def push_docker_image(self, docker_client, image: str):
        for line in docker_client.images.push(image, stream=True, decode=True):
            if 'error' in line:
                self.handle_docker_push_error(line)

    def restore_from_backup(self, docker_client, main_repo_name: str, user_repo_name: str, image_tag: str):
        if DockerImageName(main_repo_name) not in [DockerImageName.Base, DockerImageName.Filebeat, DockerImageName.Airflow, DockerImageName.Git, DockerImageName.OnlineFS, DockerImageName.TestConnector]:
            logger.info("%s - skip restoring image %s on head node", self.name, main_repo_name)
            return
        t1 = time.time()
        user_repo_image_no_tag = self.get_repo_name(self.ecr_user_registry_url, user_repo_name)
        user_repo_image = user_repo_image_no_tag + ":" + image_tag

        # restoring backup to a different ECR registry
        if self.backup_ecr_user_registry_url != self.ecr_user_registry_url:
            logger.info("%s restore from backup to the another user registry (%s:%s)", self.name, user_repo_name, image_tag)
            old_user_repo_image_no_tag = self.get_repo_name(self.backup_ecr_user_registry_url, user_repo_name)
            old_user_repo_image = old_user_repo_image_no_tag + ":" + image_tag

            old_image = docker_client.images.get(old_user_repo_image)
            old_image.tag(user_repo_image)
            docker_client.images.remove(old_user_repo_image)
            logger.info("%s change tag from %s to %s", self.name, old_user_repo_image, user_repo_image)

            if image_tag == self.python_base_image_tag:
                all_old_user_images = docker_client.images.list(name=old_user_repo_image_no_tag)
                for old_image in all_old_user_images:
                    for old_tag in old_image.tags:
                        new_tag = old_tag.replace(old_user_repo_image_no_tag, user_repo_image_no_tag)
                        if old_tag != new_tag:
                            old_image.tag(new_tag)
                            docker_client.images.remove(old_tag)
                            logger.info("%s change tag from %s to %s", self.name, old_tag, new_tag)
 
        if self.node_type == "master":
            if main_repo_name == "base":
                LoadDocker.load_docker_image(docker_client, user_repo_image, detach=False)

            self.push_docker_image(docker_client, user_repo_image)

            if image_tag == self.python_base_image_tag:
                if self.eks_cluster_enabled: 
                    self.eks_prepull_images(user_repo_image)

                other_user_images = docker_client.images.list(name=user_repo_image_no_tag)
                for image in other_user_images:
                    for tag in image.tags:
                        self.push_docker_image(docker_client, tag)
                        logger.info("%s push image %s", self.name, tag)
                        
        logger.info("%s took %d seconds to restore backup for %s:%s", self.name, (time.time() - t1), user_repo_name, image_tag)

    def eks_prepull_images(self, image):
        t1 = time.time()        
        kubernetes.config.load_kube_config()
        core_v1 = kubernetes.client.CoreV1Api()
        apps_api = kubernetes.client.AppsV1Api()
        name = 'hopsworks-prepull-' + self.instance_id

        pod_manifest = {
            'apiVersion': 'apps/v1',
            'kind': 'DaemonSet',
            'metadata': {
                'name': name
            },
            'spec': {
                'selector': {
                    'matchLabels': {
                        'name': name
                    }
                },
                'template': {
                    'metadata': {
                        'labels': {
                            'name': name
                        }
                    },
                    'spec': {
                        'initContainers': [{
                            'name': 'prepull',
                            'image': image,
                            'command': ["sh", "-c", "'true'"]
                        }],
                        'containers': [{
                            'name': 'pause',
                            'image': 'gcr.io/google_containers/pause'
                        }]
                    }
                }
            }
        }

        try: 
            core_v1.read_namespace(HOPSWORKS_NAMESPACE)
        except kubernetes.client.rest.ApiException as ex:
            if ex.status == 404:
                hopsworks_namespace_manifest = {
                    'apiVersion': 'v1',
                    'kind': 'Namespace',
                    'metadata': {
                        'name': HOPSWORKS_NAMESPACE
                    }
                }
                core_v1.create_namespace(hopsworks_namespace_manifest)
            else:
                raise ex
        try:
            apps_api.create_namespaced_daemon_set(HOPSWORKS_NAMESPACE, pod_manifest)
        except kubernetes.client.rest.ApiException as ex:
            raise_exp = True
            if ex.status == 409:
                if ex.body is not None:
                    resp = json.loads(ex.body)
                    if resp['reason'] == 'AlreadyExists':
                        logger.info("%s - Skipping since %s",
                                    self.name, resp['message'])
                        raise_exp = False
            if raise_exp:
                raise ex
        elapsed_time = time.time() - t1
        logger.info("%s took %d seconds to prepull images on EKS", self.name, elapsed_time)

    def task_rollback_function(self, task):
        pass

    def rollback(self) -> None:
        pass

    def completed_callback(self):
        if self.check_file is not None:
            self.check_file.touch()

        elapsed_time = time.time() - self._start_time
        logger.info("%s took %d seconds to finish setup for %d images", self.name, elapsed_time, len(self.tasks))

class SetupECRDockerImages(SetupDockerImages):
    def __init__(self, name: str, check_file: Path, instance_id: str, region: str, ecr_user_registry_url: str, ecr_main_registry_url: str,
     ecr_repos: Dict[DockerImageName, str], hopsworks_version: str, node_type: str, eks_cluster_enabled: bool, offline_backup_info: OffLineBackupInformation,
     default_registry_url: str, python_image_name: str, is_upgrade :bool, tags: Dict[str, str], flyingduck_enabled: bool = False) -> None:

        super().__init__(name, check_file, instance_id, region, ecr_user_registry_url, ecr_main_registry_url,
            ecr_repos, hopsworks_version, node_type, eks_cluster_enabled, offline_backup_info,
            default_registry_url, python_image_name, is_upgrade)
        self.tags = tags
        self.flyingduck_enabled = flyingduck_enabled

    def _can_create_repository(self, repo_name: str) -> bool:
        if repo_name == DockerImageName.FlyingDuck.value:
            return self.node_type == "mysqld" and self.flyingduck_enabled
        else:
            return self.node_type == "rondb_all-in-one" or self.node_type == "ndb_mgm"

    def task_function(self, task):
        repo_name, image_tag = task
        ecr_client: ECRClient = boto3.client("ecr", region_name=self.region)
        
        user_repo_name = self.get_repo_name(self.instance_id, repo_name)

        if self._can_create_repository(repo_name):
            self.create_repository(ecr_client, user_repo_name)
        
        super().task_function(task)

    def create_repository(self, ecr_client: ECRClient, user_repo_name: str):
        try:
            repo_tags = [cast(TagTypeDef, {'Key': k, 'Value': v}) for k, v in self.tags.items()]
            ecr_client.create_repository(repositoryName=user_repo_name, tags=repo_tags)
        except ecr_client.exceptions.RepositoryAlreadyExistsException:
            logger.info("%s - Skipping repository creation (%s) since it already exists", self.name, user_repo_name)
        except ecr_client.exceptions.ClientError as ex:
            logger.error("Error creating the ecr repository " + str(ex), exc_info=ex)
            if "AccessDeniedException" in str(ex): #sadly they don't give us an object on which we can read the error code
                raise ConfigException("Unable to get or create the ECR repository. " +
                "Make sure that you have added the ecr:CreateRepository permission to the instance profile. " +
                "See our documentation for more details: https://docs.hopsworks.ai/latest/setup_installation/aws/getting_started/#step-3-creating-instance-profile")
            raise
        
        lifecycle_policy = {
            "rules": [
                {
                    "rulePriority": 1,
                    "description": "Expire images older than 7 days",
                    "selection": {
                        "tagStatus": "untagged",
                        "countType": "sinceImagePushed",
                        "countUnit": "days",
                        "countNumber": 7
                    },
                    "action": {
                        "type": "expire"
                    }
                }
            ]
        }

        lifecycle_policy_text = json.dumps(lifecycle_policy)

        add_policy = True
        try:
            lifecycle_policy_response = ecr_client.get_lifecycle_policy(repositoryName=user_repo_name)
            add_policy = lifecycle_policy_response['lifecyclePolicyText'] != lifecycle_policy_text
            logger.info("%s - Lifecycle policy found for repo %s \n %s", self.name, user_repo_name, str(lifecycle_policy_response))
        except ecr_client.exceptions.LifecyclePolicyNotFoundException:
            logger.info("%s - Lifecycle policy not found for repo %s", self.name, user_repo_name)
            add_policy = True
        except ecr_client.exceptions.ClientError as ex:
            logger.error("Error getting the ecr life cycle policy " + str(ex), exc_info=ex)
            if "AccessDeniedException" in str(ex): #sadly they don't give us an object on which we can read the error code
                raise ConfigException("Unable to get the ECR life cycle policy. " +
                "Make sure that you have added the ecr:GetLifecyclePolicy permission to the instance profile. " +
                "See our documentation for more details: https://docs.hopsworks.ai/latest/setup_installation/aws/getting_started/#step-3-creating-instance-profile")
            raise
        
        if add_policy:
            try:
                logger.info("%s - Attach a new lifecycle policy to repo %s", self.name, user_repo_name)
                ecr_client.put_lifecycle_policy(repositoryName=user_repo_name, lifecyclePolicyText=lifecycle_policy_text)
            except ecr_client.exceptions.ClientError as ex:
                logger.error("Error getting the ecr life cycle policy " + str(ex), exc_info=ex)
                if "AccessDeniedException" in str(ex): #sadly they don't give us an object on which we can read the error code
                    raise ConfigException("Unable to set the ECR life cycle policy. " +
                    "Make sure that you have added the ecr:PutLifecyclePolicy permission to the instance profile. " +
                    "See our documentation for more details: https://docs.hopsworks.ai/latest/setup_installation/aws/getting_started/#step-3-creating-instance-profile")

    def get_pull_token_solution(self):
        return "Make sure that you have added the ecr:GetAuthorizationToken permission for the main repository to the instance profile. " + \
        "See our documentation for more details: https://docs.hopsworks.ai/latest/setup_installation/aws/getting_started/#step-3-creating-instance-profile"
    
    solution_msg="Make sure that you have added the {perm} permission{s} for the {repo} repository to the instance profile. " + \
        "See our documentation for more details: https://docs.hopsworks.ai/latest/setup_installation/aws/getting_started/#step-3-creating-instance-profile"
    
    def get_pull_auth_solution(self, main_repo: bool):
        repo= "user"
        if main_repo:
            repo = "main"
        return self.solution_msg.format(perm="ecr:GetDownloadUrlForLayer and ecr:BatchGetImage", repo=repo, s="s")

    def handle_docker_push_error(self, line: dict):
        error_msg_base="Unable to push the docker image to your ECR repository. " + self.solution_msg

        if line['error'] == "EOF":
            raise ConfigException(error_msg_base.format(perm="ecr:InitiateLayerUpload", repo="user", s=""))
        if "ecr:UploadLayerPart" in line['error']:
            raise ConfigException(error_msg_base.format(perm="ecr:UploadLayerPart", repo="user", s=""))
        if "ecr:CompleteLayerUpload" in line['error']:
            raise ConfigException(error_msg_base.format(perm="ecr:CompleteLayerUpload", repo="user", s=""))
        if "403" in line['error']:
            raise ConfigException(error_msg_base.format(perm="ecr:BatchCheckLayerAvailability", repo="user", s=""))
        if "ecr:PutImage" in line['error']:
            raise ConfigException(error_msg_base.format(perm="ecr:PutImage", repo="user", s=""))
        raise Exception(line['error'])

    def test_list_delete(self, image_tag: str):
        ecr_client: ECRClient = boto3.client("ecr", region_name=self.region)
        user_repo_name = self.get_repo_name(self.instance_id , "filebeat")
        try:
            ecr_client.list_images(repositoryName=user_repo_name)
        except ecr_client.exceptions.ClientError as ex:
            logger.error("Error when listing the images", exc_info=ex)
            if "AccessDeniedException" in str(ex): #sadly they don't give us an object on which we can read the error code
                raise ConfigException("Unable to list the images in the ECR repository. " +
                self.solution_msg.format(perm="ecr:ListImages", repo="user", s=""))
            raise
        try:
            ecr_client.batch_delete_image(repositoryName=user_repo_name, imageIds=[{
                'imageTag': image_tag
            }])
        except ecr_client.exceptions.ClientError as ex:
            logger.error("Error when deleting the image", exc_info=ex)
            if "AccessDeniedException" in str(ex): #sadly they don't give us an object on which we can read the error code
                raise ConfigException("Unable to delete the image in the ECR repository. " +
                self.solution_msg.format(perm="ecr:BatchDeleteImage", repo="user", s=""))
            raise

class SetupAzureDockerImages(SetupDockerImages):
    def __init__(self, name: str, check_file: Path, instance_id: str, region: str, ecr_user_registry_url: str, ecr_main_registry_url: str,
     ecr_repos: Dict[DockerImageName, str], hopsworks_version: str, node_type: str, eks_cluster_enabled: bool, offline_backup_info: OffLineBackupInformation, 
     default_registry_url: str, python_image_name: str, azure_acr_username: str, azure_acr_password: str, azure_managed_identity_id: str, is_upgrade :bool) -> None:
        self.azure_acr_password = azure_acr_password
        self.azure_acr_username = azure_acr_username
        self.azure_managed_identity_id = azure_managed_identity_id
        # Hardcoded because the AKS images start from version 2.1.0 while hopsworks.ai version was 2.0.0
        self.hopsworks_version = hopsworks_version
        super().__init__(name, check_file, instance_id, region, ecr_user_registry_url, ecr_main_registry_url,
            ecr_repos, self.hopsworks_version, node_type, eks_cluster_enabled, offline_backup_info,
            default_registry_url, python_image_name, is_upgrade)

    PERMISSION_ERROR_MSG = "We were not able to {action} the docker repository. Make sure that your managed identity has the following permission and retry: {permission}. " + \
                    "See our documentation for more details: https://docs.hopsworks.ai/latest/setup_installation/azure/getting_started/#step-4-create-a-managed-identity"
    def pre(self): 
        super().pre()
        logger.info("az login --identity --username %s", self.azure_managed_identity_id)
        SystemUtils.execute_command(["sh", "-c", f"az login --identity --username {self.azure_managed_identity_id}"], timeout=40)
        logger.info("az acr login --name %s", self.ecr_user_registry_url)
        try:
            SystemUtils.execute_command(["sh", "-c", f"az acr login --name {self.ecr_user_registry_url}"], timeout=40)
        except subprocess.TimeoutExpired as ex:
            logger.error(f"error while login in to ecr {self.ecr_user_registry_url}", exc_info=ex)
            # If the az command fail because of permissions it ask for the user name and password to be entered in an interactive way
            # As a result it time out. So we assume that if timeout it is because the permissions are missing. 
            # This is not ideal but the CLI does not offer a better solution.
            raise ConfigException(f"We were not able to log into the docker repository ({self.ecr_user_registry_url}). Make sure that the registry url is correct and " + \
                                  ", that your managed identity has the following permission then retry: Microsoft.ContainerRegistry/registries/pull/read. " + \
                                  "See our documentation for more details: https://docs.hopsworks.ai/latest/setup_installation/azure/getting_started/#step-4-create-a-managed-identity")
        with CronTab(user='root') as cron:
            job = cron.new(command=f"az acr login --name {self.ecr_user_registry_url}")
            job.hour.every(2)

        if self.node_type == "master":
            SystemUtils.execute_command(["sudo", "-H", "-u", "glassfish", "sh", "-c", f"az login --identity --username {self.azure_managed_identity_id}"], timeout=40)
            SystemUtils.execute_command(["sudo", "-H", "-u", "glassfish", "sh", "-c", f"sudo /srv/hops/sbin/dockerImage.sh loginaz {self.ecr_user_registry_url}"])
            with CronTab(user='glassfish') as cron:
                job = cron.new(command=f"sudo /srv/hops/sbin/dockerImage.sh loginaz {self.ecr_user_registry_url}")
                job.hour.every(2)

        # Docker login using the python api does not persist the credentials, hence use the cli
        # to reduce code dup, see https://github.com/docker/docker-py/issues/1061
        logger.info("docker login -u %s %s", self.azure_acr_username, self.ecr_main_registry_url)
        SystemUtils.execute_command(["sh", "-c", f"docker login -u {self.azure_acr_username} -p {self.azure_acr_password} {self.ecr_main_registry_url}"])

    def get_pull_token_solution(self):
        return "This should not happen, please retry"

    def get_pull_auth_solution(self, main_repo: bool):
        return "This should not happen, please retry"

    def handle_docker_push_error(self, line: dict):
        if "denied" in line['error']:
            raise ConfigException(self.PERMISSION_ERROR_MSG.format(action="push an image to", permission="Microsoft.ContainerRegistry/registries/push/write"))
        raise Exception(line['error'])

    def test_list_delete(self, image_tag: str):
        user_repo_name = self.get_repo_name(self.instance_id , "filebeat")
        
        SystemUtils.execute_command(["sh", "-c", f"az acr repository show-tags --name {self.ecr_user_registry_url} --repository {user_repo_name} -o tsv --only-show-errors"])

        result = SystemUtils.execute_command(["sh", "-c", f"az acr repository untag --name {self.ecr_user_registry_url} --image {user_repo_name}:{image_tag} 2>&1"], check=False)
        if result.returncode != 0:
            logger.error("Error while deleting image: " + result.stdout.decode("utf-8"))
            if "authentication required" in result.stdout.decode("utf-8"):
                raise ConfigException(self.PERMISSION_ERROR_MSG.format(action="delete an image in", permission="Microsoft.ContainerRegistry/registries/artifacts/delete"))
            raise Exception(result.stdout.decode("utf-8")) 
        
class SetupGCRDockerImages(SetupDockerImages):
    def __init__(self, name: str, check_file: Path, instance_id: str, region: str, gcr_user_registry_url: str, gcr_main_registry_url: str,
     repos: Dict[DockerImageName, str], hopsworks_version: str, node_type: str, gke_cluster_enabled: bool, offline_backup_info: OffLineBackupInformation,
     default_registry_url: str, python_image_name: str, is_upgrade :bool, gcp_project_id: str) -> None:
        
        docker_namespace = SetupGCRDockerImages._get_docker_namespace_name(instance_id)
        super().__init__(name, check_file,docker_namespace , region, gcr_user_registry_url, gcr_main_registry_url,
            repos, hopsworks_version, node_type, gke_cluster_enabled, offline_backup_info,
            default_registry_url, python_image_name, is_upgrade)
        self.gcp_project_id = gcp_project_id

    PERMISSION_ERROR_MSG = "We were not able to {action} the artifact registry repository. Make sure that your service account has the following permission and retry: {permission}. " + \
                    "See our documentation for more details: https://docs.hopsworks.ai/latest/setup_installation/gcp/getting_started/#step-3-creating-a-service-account-for-your-cluster-instances"
    
    @staticmethod
    def _get_docker_namespace_name(instance_id: str) -> str:
        return f"c-{instance_id}"
    
    @staticmethod
    def get_user_artifact_registry(gcr_user_registry_url: str, instance_id: str) -> str:
        return f"{gcr_user_registry_url}/{SetupGCRDockerImages._get_docker_namespace_name(instance_id)}"
    
    def pre(self) -> None:
        super().pre()
        if self.node_type == "rondb_all-in-one" or self.node_type == "ndb_mgm":
            self.create_repository()

        gcr_main_registry_url = self.ecr_main_registry_url.split('/')[0]
        gcr_user_registry_url = self.ecr_user_registry_url.split('/')[0]
        SystemUtils.execute_command(["sh", "-c", f"docker-credential-gcr configure-docker --registries {gcr_main_registry_url}"], timeout=120)
        SystemUtils.execute_command(["sh", "-c", f"docker-credential-gcr configure-docker --registries {gcr_user_registry_url}"], timeout=120)

    def create_repository(self):
        client = artifactregistry_v1.ArtifactRegistryClient()
        request = artifactregistry_v1.CreateRepositoryRequest(
            parent=f"projects/{self.gcp_project_id}/locations/{self.region}", 
            repository_id=self.instance_id,
            repository={
                "name": f"projects/{self.gcp_project_id}/locations/{self.region}/repositories/{self.instance_id}",
                "format": "DOCKER",
                "mode": "STANDARD_REPOSITORY",
                "cleanup_policies": {
                    "deleteUntaggedImages": {
                        "id": "deleteUntaggedImages",
                        "action": "DELETE",
                        "condition": {
                            "tag_state": "UNTAGGED",
                            "older_than": "604800s"
                        }
                    }
                }
            }
        )

        try:
            operation = client.create_repository(request=request)
            response = operation.result()
            logger.info("%s created Repositroy %s", self.name, response)
        except Exception as ex:
            logger.error("Error creating an artifact registry repository", exc_info=ex)
            if ex.__cause__ is None and ex.__context__ is not None and hasattr(ex.__context__, 'code'):
                if ex.__context__.code() == grpc.StatusCode.ALREADY_EXISTS: # type: ignore
                    logger.info("%s respository %s exists", self.name, self.instance_id)
                    return 
                elif ex.__context__.code() == grpc.StatusCode.PERMISSION_DENIED: # type: ignore
                    raise ConfigException(self.PERMISSION_ERROR_MSG.format(action="create", permission="artifactregistry.repositories.create and artifactregistry.repositories.get")) 
            raise ex

    def get_pull_token_solution(self):
        return "This should not happen, please retry"

    def docker_pull_failure_condition(self, ex : docker.errors.APIError) -> bool:
        # GCP throughs internal server error with Permission denied
        return ex.status_code == 404 or ex.status_code == 500
    
    def get_pull_auth_solution(self, main_repo: bool):
        return self.PERMISSION_ERROR_MSG.format(action="pull an image from", permission="artifactregistry.repositories.downloadArtifacts")

    def handle_docker_push_error(self, line: dict):
        if "denied" in line['error']:
            raise ConfigException(self.PERMISSION_ERROR_MSG.format(action="push an image to", permission="artifactregistry.repositories.uploadArtifacts"))
        raise Exception(line['error'])

    def test_list_delete(self, image_tag: str):
        client = artifactregistry_v1.ArtifactRegistryClient()
        filebeat_repo = f"projects/{self.gcp_project_id}/locations/{self.region}/repositories/{self.instance_id}/packages/{DockerImageName.Filebeat.value}"
        filebeat_image_tag = f"{filebeat_repo}/tags/{image_tag}"
        try:
            client.list_tags(parent=filebeat_repo)
        except Exception as ex:
            logger.error("Error listing image tags for %s", filebeat_repo, exc_info=ex)
            if ex.__cause__ is None and ex.__context__ is not None and hasattr(ex.__context__, 'code'):
                if ex.__context__.code() == grpc.StatusCode.PERMISSION_DENIED: # type: ignore
                    raise ConfigException(self.PERMISSION_ERROR_MSG.format(action="list image tags in", permission="artifactregistry.tags.list"))
                if ex.__context__.code() == grpc.StatusCode.NOT_FOUND: # type: ignore
                    raise ConfigException(self.PERMISSION_ERROR_MSG.format(action="list image tags in", permission="artifactregistry.repositories.get"))
            raise ex

        try:
            client.delete_tag(name=filebeat_image_tag)
        except Exception as ex:
            logger.error("Error deleting an image tag %s", filebeat_image_tag, exc_info=ex)
            if ex.__cause__ is None and ex.__context__ is not None and hasattr(ex.__context__, 'code'):
                if ex.__context__.code() == grpc.StatusCode.PERMISSION_DENIED: # type: ignore
                    raise ConfigException(self.PERMISSION_ERROR_MSG.format(action="delete an image tags from", permission="artifactregistry.tags.delete")) 
            raise ex