import logging

from typing import List

from ec2init import Procedure
from pathlib import Path
from ec2init.utils.system import SystemUtils
from configparser import ConfigParser
from ec2init.utils.cloud import get_cloud_provider, Provider
import kubernetes
from kubernetes.utils.quantity import parse_quantity
from ec2init.utils.constants import Scripts
from ec2init.utils.exceptions import ConfigException

logger = logging.getLogger(__name__)

HOPSWORKS_NAMESPACE = "hopsworks"

class SetupK8S(Procedure):
    def __init__(self, name: str, check_file: Path, region: str, eks_cluster_name: str, config: ConfigParser, cloud_provider: str) -> None:
        self.name = name
        self.check_file = check_file
        self.region = region
        self.eks_cluster_name =  eks_cluster_name
        self.config = config
        self.cloud_provider = cloud_provider

    def pre(self) -> None:
        pass

    def execute(self) -> List[Procedure]:
        #Setup
        self.setup_k8s()

        #Every thing is OK. 
        if self.check_file is not None:
            self.check_file.touch()

        return []

    def setup_k8s(self) -> None:
        kubernetes.config.load_kube_config()
        core_v1 = kubernetes.client.CoreV1Api()
        # update coredns 
        get_cloud_provider(self.cloud_provider).update_k8s_coredns(self.config)
        # Clean up old EKS setup
        self.clean_up_old_cluster(core_v1)
        # Setup docker max allocation vars
        self.update_max_docker_allocation(core_v1)
        self.configure_hopsfmount()

    def clean_up_old_cluster(self, core_v1) -> None:
        apps_api = kubernetes.client.AppsV1Api()
        
        try:
            # remove old prefixed daemon sets and namespaces if we have permission to do so 
            all_daemon_sets = apps_api.list_daemon_set_for_all_namespaces()
            all_daemon_sets_names = [(e.metadata.name, e.metadata.namespace) for e in all_daemon_sets.items]
            for name, namespace in all_daemon_sets_names:
                if name.startswith('hopsworks-prepull-') and namespace.startswith(f'{HOPSWORKS_NAMESPACE}-'):
                    logger.info("%s - deleting old daemonset (%s, %s)", self.name, name, namespace)
                    apps_api.delete_namespaced_daemon_set(name, namespace)
                    core_v1.delete_namespace(namespace)
        except kubernetes.client.rest.ApiException as ex:
            logger.warn(ex)

        # remove old daemon sets
        all_daemon_sets = apps_api.list_namespaced_daemon_set(HOPSWORKS_NAMESPACE)
        for e in all_daemon_sets.items:
            if e.metadata.name.startswith('hopsworks-prepull-'):
                logger.info("%s - deleting old daemonset (%s, %s)", self.name, e.metadata.name)
                apps_api.delete_namespaced_daemon_set(e.metadata.name, HOPSWORKS_NAMESPACE)

    def update_max_docker_allocation(self, core_v1) -> None:
        kube_docker_max_memory_allocation = 8192
        kube_docker_max_cores_allocation = 4
        kube_docker_cores_fraction = 1.0
        kube_docker_max_gpus_allocation = 0
        try:
            nodes = core_v1.list_node().items 
            if(len(nodes) > 0):
                capacityDict = nodes[0].status.capacity
                if 'cpu' in capacityDict:
                    kube_docker_max_cores_allocation = int(capacityDict['cpu'])
                if 'memory' in capacityDict:
                    kube_docker_max_memory_allocation = int(int(parse_quantity(capacityDict['memory'])) / 1024 / 1024)
                if 'nvidia.com/gpu' in capacityDict:
                    kube_docker_max_gpus_allocation = int(capacityDict['nvidia.com/gpu'])
        except Exception as ex:
            logger.warning("%s - Exception %s", self.name, ex)

        # update variables
        SystemUtils.execute_command([Scripts.MYSQL_CLIENT, "hopsworks", "-e", f"REPLACE INTO variables (id, value, visibility) VALUES ('kube_docker_max_memory_allocation', '{kube_docker_max_memory_allocation}', '1')"])
        SystemUtils.execute_command([Scripts.MYSQL_CLIENT, "hopsworks", "-e", f"REPLACE INTO variables (id, value, visibility) VALUES ('kube_docker_max_cores_allocation', '{kube_docker_max_cores_allocation}', '1')"])
        SystemUtils.execute_command([Scripts.MYSQL_CLIENT, "hopsworks", "-e", f"REPLACE INTO variables (id, value, visibility) VALUES ('kube_docker_cores_fraction', '{kube_docker_cores_fraction}', '1')"])
        SystemUtils.execute_command([Scripts.MYSQL_CLIENT, "hopsworks", "-e", f"REPLACE INTO variables (id, value, visibility) VALUES ('kube_docker_max_gpus_allocation', '{kube_docker_max_gpus_allocation}', '1')"])
    
    def configure_hopsfmount(self):
        # disable the apparmor profile so python jobs can run 
        # If we to enable apparmor then we need to also to configure the profile on k8s by applying the following yml file
        # https://github.com/logicalclocks/kube-hops-chef/blob/master/templates/default/hopsfsmount-apparmor-profile.yml.erb
        SystemUtils.execute_command([Scripts.MYSQL_CLIENT, "hopsworks", "-e", "REPLACE INTO variables (id, value) VALUES ('apply_hopsfsmount_apparmor_profile_kube', 'false')"])
        installation_dir = self.config.get('general', 'installation_dir', fallback='/root/ec2init')
        SystemUtils.execute_command_capture_with_print_error(["kubectl", "apply", "-f", f"{installation_dir}/smarter-device-manager-plugin.yml"], timeout=120, env={"KUBECONFIG" : "/root/.kube/config", "PATH": "/usr/local/bin:/usr/bin"})

    def rollback(self) -> None:
        pass
    

class SetupEKS(SetupK8S):
    def __init__(self, name: str, check_file: Path, region: str, eks_cluster_name: str, config: ConfigParser, is_primary_head : bool) -> None:
        super().__init__(name, check_file, region, eks_cluster_name, config, Provider.AWS)
        self.is_primary_head = is_primary_head

    def execute(self) -> List[Procedure]:
        try :
            aws_command = "aws eks --region " + self.region + " update-kubeconfig --name " + self.eks_cluster_name
            SystemUtils.execute_command(["sh", "-c", aws_command])
            SystemUtils.execute_command(["sudo", "-H", "-u", "glassfish", "sh", "-c", aws_command])
            if self.is_primary_head:
                SystemUtils.execute_command([Scripts.MYSQL_CLIENT, "hopsworks", "-e", "REPLACE INTO variables (id, value) VALUES ('kube_type', 'eks')"])
                super().execute()
            else:
                if self.check_file is not None:
                    self.check_file.touch()
        except Exception as ex:
            logger.warning("%s - Exception %s", self.name, ex)
            raise ex

        return []


class SetupAKS(SetupK8S):
    def __init__(self, name: str, check_file: Path, region: str, eks_cluster_name: str, config: ConfigParser, azure_managed_identity_id: str, azure_resource_group: str) -> None:
        self.azure_managed_identity_id = azure_managed_identity_id
        self.azure_resource_group = azure_resource_group
        super().__init__(name, check_file, region, eks_cluster_name, config, Provider.AZURE)

    def get_aks_resource_group(self) -> str:
        azure_command = f"az aks list --query \"[?name=='{self.eks_cluster_name}'].resourceGroup\" -o tsv"
        output = SystemUtils.execute_command(["sh", "-c", azure_command], timeout=60)
        aks_resource_group = output.stdout.strip().decode("utf-8")
        if aks_resource_group != self.azure_resource_group:
            logger.info("%s - AKS cluster runs in a different resource group (%s) than the Hopsworks cluster (%s)", self.name, aks_resource_group, self.azure_resource_group)
        
        if aks_resource_group == "":
            raise Exception(f"No AKS cluster found with the name {self.eks_cluster_name}. Make sure that the AKS cluster exists and the correct permissions are already set for the user-assigned-managed identity with client id {self.azure_managed_identity_id}.")
        
        return aks_resource_group

    def execute(self) -> List[Procedure]:
        try :
            SystemUtils.execute_command(["sh", "-c", f"az login --identity --username {self.azure_managed_identity_id}"], timeout=40)
            aks_resource_group = self.get_aks_resource_group()
            logger.info(f"az aks get-credentials --resource-group {aks_resource_group} --name {self.eks_cluster_name}")
            azure_command = f"az aks get-credentials --resource-group {aks_resource_group} --name {self.eks_cluster_name}"
            SystemUtils.execute_command(["sh", "-c", azure_command], timeout=40)
            SystemUtils.execute_command(["sudo", "-H", "-u", "glassfish", "sh", "-c", f"az login --identity --username {self.azure_managed_identity_id}"])
            SystemUtils.execute_command(["sudo", "-H", "-u", "glassfish", "sh", "-c", azure_command])
            SystemUtils.execute_command([Scripts.MYSQL_CLIENT, "hopsworks", "-e", "REPLACE INTO variables (id, value) VALUES ('kube_type', 'aks')"])
            SystemUtils.execute_command([Scripts.MYSQL_CLIENT, "hopsworks", "-e", "REPLACE INTO variables (id, value) VALUES ('kube_keystore_path', '/usr/lib/jvm/java-8-openjdk-amd64/jre/lib/security/cacerts')"])
            SystemUtils.execute_command([Scripts.MYSQL_CLIENT, "hopsworks", "-e", "REPLACE INTO variables (id, value) VALUES ('kube_keystore_key', 'changeit')"])
            super().execute()
        except Exception as ex:
            logger.warning("%s - Exception %s", self.name, ex)
            raise ex

        return []

class SetupGKE(SetupK8S):
    def __init__(self, name: str, check_file: Path, region: str, gke_cluster_name: str, config: ConfigParser, gcp_project_id: str) -> None:
        super().__init__(name, check_file, region, gke_cluster_name, config, Provider.GCP)
        self.gcp_project_id = gcp_project_id

    PERMISSION_ERROR_MSG = "We were not able to integrate with GKE. Make sure that your service account has the Kubernetes Engine Developer role (roles/container.developer) and retry. " + \
                    "See our documentation for more details: https://docs.hopsworks.ai/latest/setup_installation/gcp/gke_integration/#step-1-attach-kuberentes-developer-role-to-the-service-account-for-cluster-instances"
    
    def _execute_command(self, command : List[str], home_dir: str = "/root"):
        logger.info("%s - executing command: %s", self.name, command)
        result = SystemUtils.execute_command_capture_with_print_error(command, timeout=60, capture_output=True, env={"HOME": home_dir})
        if result.returncode > 0:
            raise ConfigException(SetupGKE.PERMISSION_ERROR_MSG)
        return result.stdout.strip().decode("utf-8")

    def execute(self) -> List[Procedure]:
        try :
            stdout = self._execute_command(["sh", "-c", f"gcloud container clusters list --project {self.gcp_project_id} --format='csv[no-heading](name,location,autopilot.enabled)'"])
            clusters = [line.strip().split(',') for line in stdout.split("\n")]
            logger.info("%s - found the following clusters: %s", self.name, clusters)
            cluster = [cluster for cluster in clusters if cluster[0] == self.eks_cluster_name]
            if len(cluster) == 0:
                raise ConfigException(f"We have not found any GKE cluster named {self.eks_cluster_name} under the project {self.gcp_project_id}")
            
            cluster_location = cluster[0][1]
            cluster_autopilot = cluster[0][2].lower() == "true"

            command = f"gcloud container clusters get-credentials {self.eks_cluster_name} --project {self.gcp_project_id} --location {cluster_location} --internal-ip"
            self._execute_command(["sh", "-c", command])
            self._execute_command(["sudo", "-H", "-u", "glassfish", "sh", "-c", command], home_dir="/home/glassfish")
    
            SystemUtils.execute_command([Scripts.MYSQL_CLIENT, "hopsworks", "-e", "REPLACE INTO variables (id, value) VALUES ('kube_type', 'gke')"])

            # autopilot clusters do not support configuring smarter-device-manager-plugin so we should disable hopsfsmount related functionality 
            if cluster_autopilot:
                SystemUtils.execute_command([Scripts.MYSQL_CLIENT, "hopsworks", "-e", "REPLACE INTO variables (id, value) VALUES ('mount_hopsfs_in_python_job', 'false')"])
                SystemUtils.execute_command([Scripts.MYSQL_CLIENT, "hopsworks", "-e", "REPLACE INTO variables (id, value) VALUES ('jupyter_remote_fs_driver', 'hdfscontentsmanager')"])

            super().execute()
            if self.check_file is not None:
                self.check_file.touch()
        except Exception as ex:
            logger.warning("%s - Exception %s", self.name, ex)
            raise ex

        return []