import logging
import pathlib
from configparser import ConfigParser

from typing import List

from ec2init import Procedure
from ec2init.utils.system import SystemUtils

logger = logging.getLogger(__name__)


class SetupUpdater(Procedure):
    def __init__(self, name: str, config: ConfigParser, instance_id: str, node_id: str, instance_type: str,
                 uid: str, gateway: str, api_key: str, cloud_provider: str, has_public_ip: bool, k8s_enabled: bool) -> None:
        self.name = name
        self.config = config
        self.instance_id = instance_id
        self.uid = uid
        self.gateway = gateway
        self.api_key = api_key
        self.node_id = node_id
        self.instance_type = instance_type
        self.cloud_provider = cloud_provider
        self.has_public_ip = has_public_ip
        self.k8s_enabled = k8s_enabled

    def pre(self) -> None:
        if self.report_callback is not None:
            try:
                self.report_callback.execute()
            except Exception as ex:
                logger.warning(
                    "%s - Could not execute pre-execute: %s", self.name, ex)

    def execute(self) -> List[Procedure]:
        logger.info("%s - Setting up updater script", self.name)

        installation_dir = self.config.get(
            'general', 'installation_dir', fallback='/root/ec2init')
        updater_config_file = pathlib.Path(installation_dir, "ec2update.ini")
        # Create configuration for updater
        updater_config = ConfigParser()
        updater_config['general'] = {
            'instance_id': self.instance_id,
            'user_id': self.uid,
            'gateway': self.gateway,
            'api_key': self.api_key,
            'node_id': self.node_id,
            'instance_type': self.instance_type,
            'cloud_provider': self.cloud_provider,
            'has_public_ip': str(self.has_public_ip),
            'k8s_enabled': str(self.k8s_enabled)
        }
        with open(str(updater_config_file), 'w') as fd:
            updater_config.write(fd)

        # Enable systemd service
        logger.info("Enabling ec2update systemd service")
        command = ["systemctl", "enable", "ec2update"]
        SystemUtils.execute_command(command)
        command = ["systemctl", "daemon-reload"]
        SystemUtils.execute_command(command)
        logger.info("Finished setting up ec2update")

        return []

    def rollback(self) -> None:
        pass
