
from configparser import ConfigParser
import pathlib
from typing import List, Sequence
from ec2init.plans.init.generate_hops_certs import GenerateHopsworksEndpoint
from ec2init.plans.init.glassfish_config_node import ConfigureConfigNode
from ec2init.plans.init.onlinefs_plan import GenerateOnlinefsApiKey
from ec2init.plans.multinode.master_init_plan import MasterInitPlan
from ec2init import Procedure
from ec2init.plans.multinode.register_generate_certs import SecondaryRegisterGenerateCerts
from ec2init.plans.multinode.secondary_seds import SecondarySeds
from ec2init.plans.multinode.wait_for_primary_head import WaitForPrimaryHeadToStart
from ec2init.plans.multinode.create_service_api_key import WorkerCreateServiceApiKey
from ec2init.utils import cloud

import logging

logger = logging.getLogger(__name__)

class SecondaryInitPlan(MasterInitPlan):
    def init(self, name: str, config: ConfigParser, **kwargs: str) -> None:
        super().init(name, config, **kwargs)
        mip = kwargs.get("master_ip")
        if not mip:
            raise Exception("It is SecondaryInitPlan but Master IP is missing! Have you passed master-ip argument?")
        self.master_ip = str(mip)
        self.instance_type = "secondary"

    def create(self) -> Sequence[Procedure]:
        plan: List[Procedure] = []

        sed_check = self.add_secondary_seds(plan)
        wait_for_master_check = self.add_wait_for_primary_head_to_start(plan)

        if self.cloud_provider == cloud.Provider.AWS:
            self.add_setup_eks(plan, is_primary_head=False)

        check_files = [sed_check, wait_for_master_check]
        if self.plan_utils.check_file_exists(wait_for_master_check):
            self.add_start_consul(plan)
            check_files.append(self.add_generate_host_certificates(plan))
            check_files.append(self.add_generate_onlinefs_api_key(plan))
            check_files.append(self.add_glassfish_instance(plan))

            self.add_generate_hopsworks_internal_x509(plan)

        self.add_enable_ec2update(plan)

        if self._has_ran_all_secondary(check_files):
            self.add_start_enable_services(plan)

        return plan 

    def add_wait_for_primary_head_to_start(self, plan :List[Procedure]) -> pathlib.Path:
        wait_for_master_running_check = self.plan_utils.get_procedure_check_file(WaitForPrimaryHeadToStart.__name__)
        if not self.plan_utils.check_file_exists(wait_for_master_running_check):
            logger.info("%s - Adding procedure %s", self.name, WaitForPrimaryHeadToStart.__name__)
            wait_for_master_running_task = WaitForPrimaryHeadToStart(WaitForPrimaryHeadToStart.__name__, self.config, wait_for_master_running_check, self.gateway, self.uid, self.instance_id, self.api_key)
            wait_for_master_running_task.friendly_name = "Waiting for the primary head node to start.."
            wait_for_master_running_task.report_callback = self.plan_utils.create_callback_procedure("starting", stage="Waiting for the primary head node to start..", instance_type = self.instance_type)
            plan.append(wait_for_master_running_task)
        return wait_for_master_running_check
    
    def add_secondary_seds(self, plan :List[Procedure]) -> pathlib.Path:
        sed_check = self.plan_utils.get_procedure_check_file(SecondarySeds.__name__)
        if not self.plan_utils.check_file_exists(sed_check):
            logger.info("%s - Adding procedure %s", self.name, SecondarySeds.__name__)
            sed_task = SecondarySeds(SecondarySeds.__name__, self.config, sed_check, self.cloud_provider,
                self.instance_id, self.search_domain, self.node_id, self.cluster_domain_prefix, self.gateway, self.uid, self.instance_id, self.api_key, self.deactivate_log_report, self.master_ip)
            sed_task.friendly_name = "Updating addresses"
            sed_task.report_callback = self.plan_utils.create_callback_procedure("starting", stage="Updating addresses", instance_type = self.instance_type)
            plan.append(sed_task)
        return sed_check

    def add_generate_host_certificates(self, plan :List[Procedure]) -> pathlib.Path:
        register_generate_cert_check = self.plan_utils.get_procedure_check_file(SecondaryRegisterGenerateCerts.__name__)
        if not self.plan_utils.check_file_exists(register_generate_cert_check):
            logger.info("%s - Adding procedure %s", self.name, SecondaryRegisterGenerateCerts.__name__)
            p3 = SecondaryRegisterGenerateCerts(SecondaryRegisterGenerateCerts.__name__, self.config, register_generate_cert_check,
                                                   self.master_ip, self.cloud_provider, self.api_key, self.gateway, self.uid, self.instance_id,
                                                    self.node_id, self.search_domain, self.services_list)
            p3.friendly_name = "Generating Hops x.509"
            p3.report_callback = self.plan_utils.create_callback_procedure("starting", stage="Generating Hops x.509", instance_type = self.instance_type)
            plan.append(p3)
        return register_generate_cert_check

    def add_create_api_key(self, plan: List[Procedure]):
        if WorkerCreateServiceApiKey.should_run(self.plan_utils) \
            and self.plan_utils.has_started_enabled_consul():
            # Although we are secondary Head nodes, we should skip replacing the API key in the database
            # as this has been done by the primary. We only need to replace the API key in ssl-server.xml
            # which is what Workers do
            logger.info(f"{self.name} - Adding procedure {WorkerCreateServiceApiKey.__name__}")
            p = WorkerCreateServiceApiKey(WorkerCreateServiceApiKey.__name__, self.config, self.plan_utils, self.cloud_provider, None)
            p.friendly_name = "Creating internal API key"
            p.report_callback = self.plan_utils.create_callback_procedure("starting", stage="Creating internal API key", instance_type=self.instance_type)
            plan.append(p)

    def add_generate_onlinefs_api_key(self, plan: List[Procedure]) -> pathlib.Path:
        gen_check = self.plan_utils.get_procedure_check_file(GenerateOnlinefsApiKey.__name__)
        if not self.plan_utils.check_file_exists(gen_check):
            logger.info("%s - Adding procedure %s", self.name, GenerateOnlinefsApiKey.__name__)
            gofs = GenerateOnlinefsApiKey(GenerateOnlinefsApiKey.__name__, self.config, self.api_key, gen_check)
            gofs.friendly_name = "Generating Online FS API key"
            gofs.report_callback = self.plan_utils.create_callback_procedure("starting",\
                stage="Generating online FS API key", instance_type=self.get_instance_type())
            plan.append(gofs)
        return gen_check  

    def add_glassfish_instance(self, plan: List[Procedure]) -> pathlib.Path:
        create_instance = self.plan_utils.get_procedure_check_file(ConfigureConfigNode.__name__)
        if self.plan_utils.is_ha(self.name, self.config, self.gateway, self.uid, self.instance_id, self.api_key):
            has_configured_instance = self.plan_utils.has_configured_glassfish_instance()
            if not self.plan_utils.check_file_exists(create_instance):
                logger.info("%s - Adding procedure %s", self.name, ConfigureConfigNode.__name__)
                instance = ConfigureConfigNode(ConfigureConfigNode.__name__, self.config, create_instance, has_configured_instance, self.cloud_provider, self.master_ip)
                instance.friendly_name = "Create glassfish instance"
                instance.report_callback = self.plan_utils.create_callback_procedure("starting",\
                    stage="Create glassfish instance", instance_type=self.get_instance_type())
                plan.append(instance)
        return create_instance  
    
    def add_generate_hopsworks_internal_x509(self, plan: List[Procedure]) -> None:
        gen_check = self.plan_utils.get_procedure_check_file(GenerateHopsworksEndpoint.__name__)
        create_instance = self.plan_utils.get_procedure_check_file(ConfigureConfigNode.__name__)
        if not self.plan_utils.check_file_exists(gen_check) \
                and self.plan_utils.check_file_exists(create_instance):
            logger.info("%s - Adding procedure %s", self.name, GenerateHopsworksEndpoint.__name__)
            gen_task = GenerateHopsworksEndpoint(
                GenerateHopsworksEndpoint.__name__, self.config, gen_check, self.cloud_provider, self.search_domain, self.node_id)
            gen_task.friendly_name = "Generating Hopsworks internal endpoint X.509"
            plan.append(gen_task)

    def _has_ran_all_secondary(self, checks : List[pathlib.Path]) -> bool:
        return all([self.plan_utils.check_file_exists(check) for check in checks]) and \
            self.plan_utils.has_started_enabled_consul() and self.plan_utils.has_enabled_ec2update_service() \
            and not WorkerCreateServiceApiKey.should_run(self.plan_utils)
    
    def _has_docker_ran(self) -> bool:
        return True