import configparser
import logging
import os
import json
import re

from pathlib import Path
from typing import Dict, List, Optional

from ec2init import Procedure
from ec2init.plans.plan_utils import PlanUtils
from ec2init.utils.services import Service, ServiceUtils
from ec2init.utils.system import OSType, SystemUtils
from ec2init.utils.cloud import get_cloud_provider
from ec2init.utils.offline_backup_information import OffLineBackupInformation
import yaml

from ec2init.utils.systemd import SystemdUtils

logger = logging.getLogger(__name__)

class Seds(Procedure):
    MASTER_IP = "MASTER_IP"
    MASTER_HOSTNAME = "MASTER_HOSTNAME"
    NODE_HOSTNAME= "NODE_HOSTNAME"
    NODE_IP = "NODE_IP"
    PUBLIC_DOMAIN_NAME = "PUBLIC_DOMAIN_NAME"
    SEEK_AND_REPLACE = "sudo find {0} -type f -exec sed -i {1} {{}} \\;"
    REPLACE = "sudo sed -i {0} {1}"
    REPLACE_NO_SPACES = "sudo&sed&-i&{0}&{1}"

    def __init__(self, name: str, config: configparser.ConfigParser, cloud_provider: str, search_domain: str, node_id: str, gateway: str, uid: str, instance_id: str, api_key: str, offline_backup_info: Optional[OffLineBackupInformation] = None) -> None :
        self.name = name
        self.config = config
        self.cloud_provider = cloud_provider
        self.offline_backup_info = offline_backup_info
        if OffLineBackupInformation.is_backup(offline_backup_info) and offline_backup_info is not None:
            if offline_backup_info.ip:
                self.MASTER_IP = offline_backup_info.ip
                # we only backup the master so the values are the same
                self.NODE_IP = offline_backup_info.ip
            if offline_backup_info.hostname:
                self.MASTER_HOSTNAME = offline_backup_info.hostname
                # we only backup the master so the values are the same
                self.NODE_HOSTNAME = offline_backup_info.hostname
        self.search_domain = search_domain
        self.node_id = node_id
        self.gateway = gateway
        self.uid = uid
        self.instance_id = instance_id
        self.api_key = api_key

    def pre(self) -> None:
        logger.debug("%s - Getting my local ip", self.name)
        self.local_ip = get_cloud_provider(self.cloud_provider).get_private_ip(self.config)

        self.fqdn = get_cloud_provider(self.cloud_provider).get_fqdn(self.config, self.search_domain, self.node_id)

        if self.report_callback is not None:
            try:
                self.report_callback.execute()
            except Exception as ex:
                logger.warning("%s - Could not execute pre-execute: %s", self.name, ex)

    def execute(self) -> List[Procedure]:

        self._replaceNodeFqdnConsulBashrc(self.fqdn)
        self._replaceNodeFqdnConsulService(self.fqdn)
        self._replaceNodeFqdnHopsifyStateStore(self.fqdn)
        self._replaceNodeIpDnsmasq(self.local_ip)
        if SystemUtils.get_os_type() == OSType.UBUNTU:
            self._replaceNodeIpSystemdResolved(self.local_ip)

        self.ha_logstash()
        return []

    def _replaceNodeFqdnConsulService(self, fqdn: str) -> None:
        old_hostname_config = "\$(hostname -f | tr -d '\[:space:\]')"
        if OffLineBackupInformation.is_backup(self.offline_backup_info):
            old_hostname_config = self.NODE_HOSTNAME
        subString = "s/{0}/{1}/g".format(old_hostname_config, fqdn)
        cmd = Seds.REPLACE_NO_SPACES.format(subString, '/lib/systemd/system/consul.service')
        logger.debug("%s - Replacing hostname -f in /lib/systemd/system/consul.service", self.name)
        SystemUtils.execute_command(cmd.split("&"))
        logger.info("%s - Replaced %s in /lib/systemd/system/consul.service", self.name, fqdn)

    def _replaceNodeFqdnConsulBashrc(self, fqdn: str) -> None:
        subString = "s/{0}/{1}/g".format(self.NODE_HOSTNAME, fqdn)
        cmd = Seds.REPLACE.format(subString, '/srv/hops/consul/.bashrc')
        logger.debug("%s - Replacing hostname in /srv/hops/consul/.bashrc", self.name)
        SystemUtils.execute_command(cmd.split(" "))
        logger.info("%s - Replaced %s in /srv/hops/consul/.bashrc", self.name, fqdn)

    def _replaceNodeFqdnHopsifyStateStore(self, fqdn: str) -> None:
        kagent_conf_dir = self.config.get('general', 'kagent_conf_dir', fallback="/srv/hops/kagent/etc")
        kagent_conf_file = Path(kagent_conf_dir, 'config.ini')
        kagent_conf = configparser.ConfigParser()
        kagent_conf.read(str(kagent_conf_file))
        state_store_location = kagent_conf.get('agent', 'state-store', fallback='/srv/hops/kagent/etc/state_store')
        hopsify_state_store = os.path.join(state_store_location, "crypto_material_state.json")

        logger.debug("%s - Replacing hostname in %s", self.name, str(hopsify_state_store))
        with open(hopsify_state_store, 'r') as fd:
            cms = json.load(fd)

        service_discovery_dn = re.compile(".+\.service\.[a-zA-Z0-9]+$")
        for s in cms["crypto_state_store"]:
            cn = cms["crypto_state_store"][s]["common_name"]
            if not service_discovery_dn.match(cn):
                cms["crypto_state_store"][s]["common_name"] = fqdn

        with open(hopsify_state_store, 'w') as fd:
            json.dump(cms, fd)

        logger.info("%s - Replaced %s in %s", self.name, fqdn, str(hopsify_state_store))

    def _replaceNodeIpDnsmasq(self, local_ip: str) -> None:
        subString = "s/{0}/{1}/g".format(self.NODE_IP, local_ip)
        cmd = Seds.REPLACE.format(subString, '/etc/dnsmasq.d/default')
        logger.debug("%s - Replacing IP in /etc/dnsmasq.d/default", self.name)
        SystemUtils.execute_command(cmd.split(" "))
        logger.info("%s - Replaced %s in /etc/dnsmasq.d/default", self.name, local_ip)

    def _replaceNodeIpSystemdResolved(self, local_ip: str) -> None:
        subString = f"s/DNS=.*/DNS={local_ip}/"
        cmd = Seds.REPLACE.format(subString, '/etc/systemd/resolved.conf')
        logger.debug("%s - Replacing IP in /etc/systemd/resolved.conf", self.name)
        SystemUtils.execute_command(cmd.split(" "))
        logger.info("%s - Replaced %s in /etc/systemd/resolved.conf", self.name, local_ip)
        SystemUtils.execute_command(["systemctl", "restart", "systemd-resolved"])

    # And the award to the most ambiguous method name goes to sed_almost_similar
    # Change the same files but for Master node change to LOCAL-IP and for Workers change to MASTER-IP
    def sed_almost_similar(self, ip2replace: str, hostname2replace: str) -> None:
        subString = f"s/{self.MASTER_IP}/{ip2replace}/g"

        cmd = Seds.REPLACE.format(subString, '/srv/hops/hadoop/etc/hadoop/mapred-site.xml')
        logger.debug("%s - Replacing IP in /srv/hops/hadoop/etc/hadoop/mapred-site.xml", self.name)
        SystemUtils.execute_command(cmd.split(" "))
        logger.info("%s - Replaced %s with %s in /srv/hops/hadoop/etc/hadoop/mapred-site.xml", self.name, self.MASTER_IP, ip2replace)

        subString = f"s/{self.MASTER_HOSTNAME}/{hostname2replace}/g"

        cmd = Seds.REPLACE.format(subString, '/srv/hops/hadoop/etc/hadoop/hdfs-site.xml')
        logger.debug("%s - Replacing hostname in /srv/hops/hadoop/etc/hadoop/hdfs-site.xml", self.name)
        SystemUtils.execute_command(cmd.split(" "))
        logger.info("%s - Replaced %s with %s in /srv/hops/hadoop/etc/hadoop/hdfs-site.xml", self.name, self.MASTER_HOSTNAME, hostname2replace)

        cmd = Seds.REPLACE.format(subString, '/srv/hops/hadoop/etc/hadoop/container-executor.cfg')
        logger.debug("%s - Replacing hostname in /srv/hops/hadoop/etc/hadoop/container-executor.cfg", self.name)
        SystemUtils.execute_command(cmd.split(" "))
        logger.info("%s - Replaced %s with %s in /srv/hops/hadoop/etc/hadoop/container-executor.cfg", self.name, self.MASTER_HOSTNAME, hostname2replace)

        cmd = Seds.REPLACE.format(subString, '/srv/hops/opensearch-dashboards/config/opensearch_dashboards.yml')
        logger.debug("%s - Replacing hostname in /srv/hops/opensearch-dashboards/config/opensearch_dashboards.yml", self.name)
        SystemUtils.execute_command(cmd.split(" "))
        logger.info("%s - Replaced %s with %s in /srv/hops/opensearch-dashboards/config/opensearch_dashboards.yml", self.name, self.MASTER_HOSTNAME, hostname2replace)

    def ha_logstash(self) -> None:
        cluster_status = PlanUtils.get_cluster_status(self.name, self.config, self.gateway, self.uid, self.instance_id, self.api_key)
        logstach_nodes = [node for node in cluster_status["payload"] if node["type"] == "master" or (node["type"] == "secondary" and ServiceUtils.containsValue(node["services"], Service.Logstash))]
        if len(logstach_nodes) > 1:
            filebeat_dir = "/srv/hops/filebeat"
            filebeat_config_files = [ f for f in os.listdir(filebeat_dir) if f.startswith("filebeat-") and f.endswith(".yml")]
            for f_config in filebeat_config_files:
                filebeat_config_file = f"{filebeat_dir}/{f_config}"
                logger.debug("%s - set logstash ha for %s", self.name, filebeat_config_file)
                with open(filebeat_config_file, 'r') as f:
                    filebeat_config : Dict  = yaml.load(f, Loader=yaml.loader.SafeLoader)
                if "logstash" in filebeat_config["output"]:
                    filebeat_config["output"]["logstash"]["worker"] = len(logstach_nodes)
                    filebeat_config["output"]["logstash"]["loadbalance"] = True
                
                with open(filebeat_config_file, 'w') as f:
                    yaml.dump(filebeat_config, f, sort_keys=False, default_flow_style=False)

    
    def ha_alertmanager(self) -> None:
        cluster_status = PlanUtils.get_cluster_status(self.name, self.config, self.gateway, self.uid, self.instance_id, self.api_key)
        alertmanager_nodes = [node for node in cluster_status["payload"] if node["type"] == "master" or (node["type"] == "secondary" and ServiceUtils.containsValue(node["services"], Service.Alertmanager))]
        if len(alertmanager_nodes) > 1:
            listen_address = r'1! s/\(--storage.path=\)/--cluster.listen-address=0.0.0.0:9194 \\\n \1/'
            peers = r'1! s/\(--storage.path=\)/--cluster.peer=alertmanager.prometheus.service.consul:9194 \\\n \1/'
            path = "/lib/systemd/system/alertmanager.service"
            listen_address_set=False
            peers_set=False

            with open(path) as f:
                alertmanager_service = f.read()
                if '--cluster.listen-address=0.0.0.0:9194' in alertmanager_service:
                    listen_address_set=True
                if '--cluster.peer=alertmanager.prometheus.service.consul:9194' in alertmanager_service:
                    peers_set=True

            if not listen_address_set:
                logger.debug("%s - Adding cluster.listen-address in %s", self.name, path)
                command = ["sudo", "sed", "-i", listen_address, path]
                SystemUtils.execute_command(command)
                logger.info("%s - Added cluster.listen-address in %s", self.name, path)
            if not peers_set:
                logger.debug("%s - Adding cluster.peer in %s", self.name, path)
                command = ["sudo", "sed", "-i", peers, path]
                SystemUtils.execute_command(command)
                logger.info("%s - Added cluster.peer in %s", self.name, path)

            if not listen_address_set or not peers_set:
                SystemdUtils.reload()
    
    def rollback(self) -> None:
        pass
