import logging
import pathlib

from configparser import ConfigParser
from typing import Sequence, List

from ec2init import Procedure
from ec2init.plans.init.base_init_plan import BaseInitPlan
from ec2init.plans.multinode.worker_seds import WorkerSeds
from ec2init.plans.multinode.register_generate_certs import WorkerRegisterGenerateCerts
from ec2init.utils.cloud import Provider
from ec2init.utils.bucket_configuration import BucketConfiguration

logger = logging.getLogger(__name__)

class WorkerInitPlan(BaseInitPlan):
    def init(self, name: str, config: ConfigParser, **kwargs: str) -> None:
        super().init(name, config, **kwargs)
        mip = kwargs.get("master_ip")
        if not mip:
            raise Exception("It is WorkerInitPlan but Master IP is missing! Have you passed master-ip argument?")
        self.master_ip = str(mip)
        self.instance_type = "worker"
        self.cloud_provider = kwargs['cloud_provider']
        self.bucket_configuration = BucketConfiguration.get_bucket_configuration(self.cloud_provider, **kwargs)
        self.azure_storage_account = kwargs.get("azure_storage_account", "")
        self.azure_managed_identity_id = kwargs.get("azure_managed_identity_id", "")

    def create(self) -> Sequence[Procedure]:
        plan: List[Procedure] = []

        self.add_fix_hdfs_site_for_bucket(plan)

        # Setup ECR Docker images 
        if self.cloud_provider == Provider.AWS:
            self.add_setup_ecr_docker_images(plan)
        elif self.cloud_provider == Provider.AZURE:
            self.add_setup_azure_docker_images(plan)
        elif self.cloud_provider == Provider.GCP:
            self.add_setup_gcr_docker_images(plan)
        
        sed_check = self.plan_utils.get_procedure_check_file(WorkerSeds.__name__)
        if not self.plan_utils.check_file_exists(sed_check):
            logger.info("%s - Adding procedure %s", self.name, WorkerSeds.__name__)
            sed_task = WorkerSeds(WorkerSeds.__name__, self.config, sed_check, self.master_ip, self.cloud_provider, self.search_domain, self.node_id, self.gateway, self.uid, self.instance_id, self.api_key)
            sed_task.report_callback = self.plan_utils.create_callback_procedure("starting", stage="Updating addresses", instance_type = self.instance_type)
            plan.append(sed_task)

        self.add_start_consul(plan)

        # Regenerate Hops host certificate
        register_generate_cert_check = self.plan_utils.get_procedure_check_file(WorkerRegisterGenerateCerts.__name__)
        if not self.plan_utils.check_file_exists(register_generate_cert_check):
            logger.info("%s - Adding procedure %s", self.name, WorkerRegisterGenerateCerts.__name__)
            gen_x509 = WorkerRegisterGenerateCerts(WorkerRegisterGenerateCerts.__name__, self.config, register_generate_cert_check,
                                                   self.master_ip, self.cloud_provider, self.api_key, self.gateway, self.uid, self.instance_id, self.node_id, self.search_domain)
            gen_x509.friendly_name = "Generating Hops x.509"
            gen_x509.report_callback = self.plan_utils.create_callback_procedure("starting", stage="Generating Hops x.509", instance_type = self.instance_type)
            plan.append(gen_x509)
        
        self.add_enable_ec2update(plan)

        self.add_create_api_key(plan)
        
        return plan

    def _has_ran_all(self, sed_check: pathlib.Path, gen_x509: pathlib.Path, service_jwt_check: pathlib.Path) -> bool:
        return self.plan_utils.check_file_exists(sed_check) and self.plan_utils.check_file_exists(gen_x509) \
            and self.plan_utils.has_started_enabled_consul() and self.plan_utils.check_file_exists(service_jwt_check)
