import logging
import configparser

from typing import Sequence, List, Optional

from ec2init import Procedure
from ec2init.plans.init.base_init_plan import BaseInitPlan
from ec2init.plans.ndb.start_agent import StartNdbAgent
from ec2init.plans.init.user_init_script import UserInitScript
from ec2init.plans.init.callback_hopsworks import CallbackHopsworks

logger = logging.getLogger(__name__)


class NdbNodeInitPlan(BaseInitPlan):
    def init(self, name: str, config: configparser.ConfigParser, **kwargs: str) -> None:
        super().init(name, config, **kwargs)
        self.extra_mysqld_vms = kwargs.get("extra_mysqld_vms", "1")
        logger.info("%s - extra_mysqld_vms = %s", self.name, self.extra_mysqld_vms)

    def create(self) -> Sequence[Procedure]:
        plan: List[Procedure] = []

        self.add_callback_instance_booted(plan, user_email=self.user_email)
        callback_hopsworks_check = self.plan_utils.get_procedure_check_file(
            "{}_booted".format(CallbackHopsworks.__name__)
        )
        if self.plan_utils.check_file_exists(callback_hopsworks_check):
            self.add_user_init_script(plan)
        return plan

    def add_start_ndb_agent(self, plan: List[Procedure]) -> None:
        user_init_check = self.plan_utils.get_procedure_check_file(
            UserInitScript.__name__
        )
        if (
            self.plan_utils.check_file_exists(user_init_check)
            and not StartNdbAgent.has_run()
        ):
            logger.info(f"{self.name} - Adding procedure {StartNdbAgent.__name__}")
            sna = StartNdbAgent(
                StartNdbAgent.__name__,
                self.config,
                self.api_key,
                False,
                self.gateway,
                self.uid,
                self.instance_id,
                self.extra_mysqld_vms,
            )
            sna.friendly_name = "Setting NDB agent"
            sna.report_callback = self.plan_utils.create_callback_procedure(
                "ndb_starting",
                stage="Setting NDB agent",
                instance_type=self.get_instance_type(),
            )
            plan.append(sna)


class MgmNodeInitPlan(NdbNodeInitPlan):
    def init(self, name: str, config: configparser.ConfigParser, **kwargs: str) -> None:
        super().init(name, config, **kwargs)
        self.instance_type = "ndb_mgm"
        self.is_bootstrap = (
            True if kwargs.get("mgm_bootstrap", "false").lower() == "true" else False
        )
        if self.is_bootstrap:
            self.instance_type += "/bootstrap"

    def create(self) -> Sequence[Procedure]:
        plan: List[Procedure] = []

        parentPlan: Sequence[Procedure] = super().create()
        plan += list(parentPlan)

        self.add_start_ndb_agent(plan)
        return plan

    def add_start_ndb_agent(self, plan: List[Procedure]) -> None:
        if not StartNdbAgent.has_run():
            logger.info(f"{self.name} - Adding procedure {StartNdbAgent.__name__}")
            sna = StartNdbAgent(
                StartNdbAgent.__name__,
                self.config,
                self.api_key,
                self.is_bootstrap,
                self.gateway,
                self.uid,
                self.instance_id,
                self.extra_mysqld_vms,
            )
            sna.friendly_name = "Setting NDB agent"
            sna.report_callback = self.plan_utils.create_callback_procedure(
                "ndb_starting",
                stage="Setting NDB agent",
                instance_type=self.get_instance_type(),
            )
            plan.append(sna)


class NdbdNodeInitPlan(NdbNodeInitPlan):
    def init(self, name: str, config: configparser.ConfigParser, **kwargs: str) -> None:
        super().init(name, config, **kwargs)
        self.instance_type = "ndb_ndbd"

    def create(self) -> Sequence[Procedure]:
        plan: List[Procedure] = []

        parentPlan: Sequence[Procedure] = super().create()
        plan += list(parentPlan)
        self.add_start_ndb_agent(plan)

        return plan


class MysqldNodeInitPlan(NdbNodeInitPlan):
    def init(self, name: str, config: configparser.ConfigParser, **kwargs: str) -> None:
        super().init(name, config, **kwargs)
        self.instance_type = "mysqld"

    def create(self) -> Sequence[Procedure]:
        plan: List[Procedure] = []

        parentPlan: Sequence[Procedure] = super().create()
        plan += list(parentPlan)
        self.add_start_ndb_agent(plan)
        return plan


class ApiNodeInitPlan(NdbNodeInitPlan):
    def init(self, name: str, config: configparser.ConfigParser, **kwargs: str) -> None:
        super().init(name, config, **kwargs)
        self.instance_type = "ndb_api"

    def create(self) -> Sequence[Procedure]:
        plan: List[Procedure] = []

        parentPlan: Sequence[Procedure] = super().create()
        plan += list(parentPlan)
        self.add_start_ndb_agent(plan)
        return plan


class RonDBAllInOneInitPlan(MgmNodeInitPlan):
    def init(self, name: str, config: configparser.ConfigParser, **kwargs: str) -> None:
        super().init(name, config, **kwargs)
        self.instance_type = "rondb_all-in-one"
        self.is_bootstrap = (
            True if kwargs.get("mgm_bootstrap", "false").lower() == "true" else False
        )
