import logging
import time

from typing import List

from ec2init import Procedure
from ec2init.plans.init.start_services import StartServices
from ec2init.utils.system import SystemUtils
from ec2init.utils.services import Service
from ec2init.utils.systemd import SystemdUtils

logger = logging.getLogger(__name__)

class StartSideServices(Procedure):

    SIDE_CAR_SERVICES = [Service.NodeExporter, Service.MySQLdExporter, Service.Kagent, Service.FilebeatService]
    SIDE_CAR_SERVICES_WITH_FLYINGDUCK = SIDE_CAR_SERVICES + [Service.FlyingDuck]

    def __init__(self, name: str, enable_flyingduck: bool) -> None:
        self.name = name
        self.enable_flyingduck = enable_flyingduck

    def pre(self) -> None:
        pass

    def rollback(self) -> None:
        pass

    def execute(self) -> List[Procedure]:
        logger.info(f"{self.name} - Starting execution")
        for service_enum in StartSideServices._get_side_car_services(self.enable_flyingduck):
            service = service_enum.value
            if SystemdUtils.is_service_installed(service):
                logger.info(f"{self.name} - Starting {service}")
                # We need to reload consul for now and relay on the fact that it is passing the first time, we wouldn't need this once we have SAN in certificates
                if service_enum == Service.FlyingDuck:
                    SystemUtils.execute_command(["systemctl", "reload", "consul"])
                SystemUtils.execute_command(StartServices.SYSTEMCTL_ENABLE.format(service=service).split(" "))
                SystemUtils.execute_command(StartServices.SYSTEMCTL_START.format(service=service).split(" "), timeout=120)
                logger.info(f"{self.name} - Started {service}")

        time.sleep(5)
        return []

    @staticmethod
    def _get_side_car_services(enable_flyingduck: bool) -> List[Service]:
        if enable_flyingduck:
            return StartSideServices.SIDE_CAR_SERVICES_WITH_FLYINGDUCK 
        return StartSideServices.SIDE_CAR_SERVICES
    
    @staticmethod
    def has_started_services(enable_flyingduck: bool) -> bool:
        for service_enum in StartSideServices._get_side_car_services(enable_flyingduck):
            service = service_enum.value
            if SystemdUtils.is_service_installed(service):
                if not SystemdUtils.is_enabled(service) or not SystemdUtils.is_active(service):
                    logger.warn(f"Service {service} is not active/running")
                    return False
        return True
