import configparser
import pathlib
import base64
import json
from configparser import ConfigParser
import logging

from typing import Any, Dict, Optional

from ec2init.plans.init.callback_hopsworks import CallbackHopsworks
from ec2init.plans.init.start_services import StartServices
from ec2init.utils import cloud
from ec2init.utils.services import Service, ServiceUtils

from ec2init.utils.cloud import get_cloud_provider
from ec2init.utils.http import HttpUtils
from ec2init.utils.systemd import SystemdUtils

REPORT_ENDPOINT = (
    "POST", "{gateway}/eventsv2/instance-state/users/{uid}/instances/{instance_id}")

CLUSTER_STATUS_ENDPOINT = (
    "GET", "{gateway}/eventsv2/instance-state/users/{uid}/instances/{instance_id}")

logger = logging.getLogger(__name__)

class PlanUtils:
    def __init__(self, checks_dir: pathlib.Path) -> None:
        self.checks_dir = checks_dir

    def get_procedure_check_file(self, procedure_name: str) -> pathlib.Path:
        return self.checks_dir / procedure_name

    def check_file_exists(self, check_file: pathlib.Path) -> bool:
        return check_file.is_file()

    def remove_check_file(self, check_file: pathlib.Path) -> None:
        if self.check_file_exists(check_file):
            check_file.unlink()
        
    def get_procedures_check_directory(self) -> pathlib.Path:
        return self.checks_dir

    @staticmethod
    def create_checks_directory(config: ConfigParser) -> pathlib.Path:
        installation_dir = config.get('general', 'installation_dir', fallback='/root/ec2init')
        ec2init_dir = pathlib.Path(installation_dir, "ec2init_checks")

        if not ec2init_dir.is_dir():
            ec2init_dir.mkdir()
        return ec2init_dir

    @staticmethod
    def report_error(plan_name :str, config: configparser.ConfigParser, gateway: str, user_id: str, instance_id: str, api_key :str, error_message: str, node_id: Optional[str]=None, instance_type: Optional[str]=None) -> None:
        method, url = REPORT_ENDPOINT
        url = url.format(gateway=gateway, uid=user_id,
                         instance_id=instance_id)
        data = {"instanceState": "error", "error": error_message}
        if node_id:
            data['nodeId'] = node_id
        if instance_type:
            data['instanceType'] = instance_type
        logger.debug(
            "%s - Reporting an unrecoverable error to %s", plan_name, url)
        logger.debug("%s - Error is %s", plan_name, data)
        headers: Dict[str, str] = {}
        if api_key:
            headers = {"x-api-key": api_key}
        try:
            HttpUtils.maybe_api_call(config, url, method, data=data, headers=headers)
            logger.debug("%s - Reported error", plan_name)
        except Exception as ex:
            logger.error(
                "%s - Could not report error to Hopsworks Cloud: %s", plan_name, ex)

    @staticmethod
    def get_cluster_status(plan_name: str, config: configparser.ConfigParser, gateway: str, user_id: str, instance_id: str, api_key :str) -> Dict[str, Any]:
        method, url = CLUSTER_STATUS_ENDPOINT
        url = url.format(gateway=gateway, uid=user_id, instance_id=instance_id)
        headers: Dict[str, str] = {}
        if api_key:
            headers = {"x-api-key": api_key}
        try:
            logger.info("%s - get cluster status", plan_name)
            return HttpUtils.do_api_call(config, url, method, headers=headers)
        except Exception as ex:
            logger.error(
                "%s - Could not get cluster status from Hopsworks Cloud: %s", plan_name, ex)
        return {}

    @staticmethod
    def get_secondary_node_id(plan_name: str, config: configparser.ConfigParser, gateway: str, user_id: str, instance_id: str, api_key :str, cloud_provider :str) -> str:
        local_ip = cloud.get_cloud_provider(cloud_provider).get_private_ip(config)
        cluster_status = PlanUtils.get_cluster_status(plan_name, config, gateway, user_id, instance_id, api_key)
        secondary_node_id_arr = [node["secondaryNodeId"] for node in cluster_status["payload"] if node["type"] == "secondary" and node['ip'] == local_ip]
        my_node_id = "0"
        if len(secondary_node_id_arr) > 0:
            my_node_id = str(secondary_node_id_arr[0])
        return my_node_id
    
    @staticmethod
    def is_ha(plan_name: str, config: configparser.ConfigParser, gateway: str, user_id: str, instance_id: str, api_key :str):
        cluster_status = PlanUtils.get_cluster_status(plan_name, config, gateway, user_id, instance_id, api_key)
        glassfish_instanses = [node for node in cluster_status["payload"] if node["type"] == "secondary" and ServiceUtils.containsValue(node["services"], Service.GlassfishInstance)]
        return len(glassfish_instanses) > 0
    
class HopsPlanUtils(PlanUtils):

    def __init__(self, config: ConfigParser, user_password: Optional[str], **kwargs: str) -> None:
        super().__init__(PlanUtils.create_checks_directory(config))
        self.config = config
        self.gateway = kwargs.get("gateway", "gateway")
        self.instance_id = kwargs.get("instance_id", "instance_id")
        self.uid = kwargs.get("uid", "uid")
        self.user_email = kwargs.get("user_email", "user_email")
        self.user_password = user_password
        self.api_key = kwargs.get("api_key", "")
        self.node_id = kwargs.get("node_id", "node_id")
        self.cloud_provider = kwargs.get("cloud_provider", "cloud_provider")
        self.issue_lets_encrypt = True if kwargs.get("issue_lets_encrypt", "true").lower() == "true" else False
        self.has_public_ip = True if kwargs.get("has_public_ip", "true").lower() == "true" else False
        self.user_mgm_type = kwargs.get("user_management_type", "")
        self.cluster_domain_prefix = kwargs.get("cluster_domain_prefix", "")
        # Default is true if not set. create_pwd_user will only be set for OAuth
        self.create_pwd_user = False if kwargs.get("create_pwd_user", "").lower() == "false" else True

    def base64_to_json(self, conf: str) -> Dict[str, str]:
        if conf:
            decoded = base64.b64decode(conf)
            return json.loads(decoded)
        return {}

    def get_cloud_tmp_test_file(self) -> pathlib.Path:
        path = self.checks_dir / "test-obj"
        f = open(path, 'w+')
        f.write("S3/Azure test file")
        f.close()
        return path

    def get_cloud_versioning_enabled_file(self) -> pathlib.Path:
        return self.checks_dir / "versioning-enabled" 

    def create_callback_procedure(self, state: str, stage: str = None, check_file: pathlib.Path = None, instance_type: str = "") -> CallbackHopsworks:
        return CallbackHopsworks(CallbackHopsworks.__name__, self.config, self.gateway, self.instance_id,  self.node_id,
                                 instance_type, state, self.uid, REPORT_ENDPOINT, self.has_public_ip, self.user_email, self.user_password, stage, check_file, self.api_key, 
                                 self.cloud_provider, user_mgm_type=self.user_mgm_type, create_pwd_user=self.create_pwd_user)

    def has_issued_certificate(self) -> bool:
        # Mock it if it's unmanaged instance or instance has no internet connectivity and issue
        mock_msg = "Mock has_issued_certificate check because"
        if self.config.getboolean('general', 'unmanaged', fallback=False):
            logger.warning(f"{mock_msg} it is an unmanaged instance")
            return True
        if not self.issue_lets_encrypt:
            logger.warning(f"{mock_msg} flag issue_lets_encrypt is set to False")
            return True
        domain_name = get_cloud_provider(
            self.cloud_provider).get_public_domain_name(self.config, self.instance_id, self.cluster_domain_prefix)
        certificate = pathlib.Path(
            '/etc', 'letsencrypt', 'live', domain_name, 'cert.pem')
        return certificate.is_file()

    def has_configured_glassfish(self) -> bool:
        if self.config.getboolean('general', 'unmanaged', fallback=False)\
            or not self.issue_lets_encrypt:
            return SystemdUtils.is_enabled('glassfish-domain1') \
                and SystemdUtils.is_active('glassfish-domain1')
        lets_encrypt_dir = self.config.get(
            'initplan', 'lets_encrypt_dir', fallback="/etc/letsencrypt")
        script = pathlib.Path(lets_encrypt_dir, 'renewal-hooks',
                              'deploy', 'deploy2glassfish_hook.sh')
        return script.is_file()
    
    def has_configured_glassfish_instance(self) -> bool:
        return SystemdUtils.is_enabled(Service.GlassfishInstance.value) and SystemdUtils.is_active(Service.GlassfishInstance.value)

    def has_started_enabled_consul(self) -> bool:
        return SystemdUtils.is_active("consul") and SystemdUtils.is_enabled("consul")

    def has_started_enabled_kagent(self) -> bool:
        return SystemdUtils.is_active("kagent") and SystemdUtils.is_enabled("kagent")

    def has_enabled_ec2update_service(self) -> bool:
        if self.config.getboolean('general', 'unmanaged', fallback=False):
            return True
        return SystemdUtils.is_enabled("ec2update")

    def has_started_enabled_cloudwatch_agent(self) -> bool:
        return SystemdUtils.is_active("amazon-cloudwatch-agent")\
            and SystemdUtils.is_enabled("amazon-cloudwatch-agent")
            
    @staticmethod
    def clean_hopsworks_version(version: str) -> str:
        return version.replace('-SNAPSHOT', '')
