import logging
from configparser import ConfigParser
import pathlib

from typing import Sequence, List

from ec2init import Procedure
from ec2init.plans.init.base_init_plan import BaseInitPlan
from ec2init.plans.multinode.generate_opensearch_users_passwords import GenerateOpensearchUsersPasswords, UpdateOpensearchUsersPasswords
from ec2init.plans.postinit.post_init import PostInit
from ec2init.plans.init.user_init_script import UserInitScript 
from ec2init.plans.postinit.store_backup_info import StoreBackupInfo
from ec2init.plans.postinit.expat import Expat 
from ec2init.plans.init.start_services import StartServices

logger = logging.getLogger(__name__)


class BasePostInitPlan(BaseInitPlan):
    def init(self, name: str, config: ConfigParser, **kwargs: str) -> None:
        super().init(name, config, **kwargs)

    def create(self) -> Sequence[Procedure]:
        plan: List[Procedure] = []
        self.add_set_custom_hostname(plan)
        return plan

#for master but can't be renamed for backward compatibility reasons
class PostInitPlan(BasePostInitPlan):
    def init(self, name: str, config: ConfigParser, **kwargs: str) -> None:
        super().init(name, config, **kwargs)
        self.instance_type = "master"

    def create(self) -> Sequence[Procedure]:
        plan: List[Procedure] = []
        parentPlan: Sequence[Procedure] = super().create()
        plan += list(parentPlan)

        # Generate passwords for Opensearch internal users
        generate_opensearch_users_passwords = self.add_generate_opensearch_users_passwords(plan)

        offline_backup_info_check = self.plan_utils.get_procedure_check_file(StoreBackupInfo.__name__)
        if not self.plan_utils.check_file_exists(offline_backup_info_check):
            backupInfoPlan = StoreBackupInfo(StoreBackupInfo.__name__, self.config, self.plan_utils.get_procedure_check_file(
                "backup_info"), offline_backup_info_check, self.cloud_provider, self.user_registry_url, self.search_domain, self.node_id)
            backupInfoPlan.friendly_name = "Store backup Info"
            backupInfoPlan.fail_plan = True
            backupInfoPlan.report_callback = self.plan_utils.create_callback_procedure("starting", stage="Store backup Info", instance_type=self.get_instance_type())
            plan.append(backupInfoPlan)

        post_init_check = self.plan_utils.get_procedure_check_file(PostInit.__name__)
        if not self.plan_utils.check_file_exists(post_init_check):
            registry_url = f"{self.user_registry_url}/{self.instance_id}" if self.user_registry_url != "" else None
            logger.info("%s - Adding procedure %s", self.name, PostInit.__name__)
            postInitPlan = PostInit(PostInit.__name__, self.config, check_file=post_init_check, bucket_configuration=self.bucket_configuration, is_upgrade=self.is_upgrade,
                            user_management_type=self.user_mgm_type, create_pwd_user=self.create_pwd_user, is_backup=self.is_backup, user_password=self.user_password, registry_url=registry_url)
            postInitPlan.friendly_name = "Post Cluster Init."
            postInitPlan.fail_plan = True
            postInitPlan.report_callback = self.plan_utils.create_callback_procedure("starting", stage="Post Cluster Init", instance_type=self.get_instance_type())
            plan.append(postInitPlan)

        expatProcCondition = True
        if self.is_upgrade:
            expat_check = self.plan_utils.get_procedure_check_file(Expat.__name__)
            expatProcCondition = self.plan_utils.check_file_exists(expat_check)
            if not expatProcCondition and self.plan_utils.check_file_exists(post_init_check):
                logger.info("%s - Adding procedure %s", self.name, Expat.__name__)
                expatProc = Expat(Expat.__name__, self.config, check_file=expat_check, source_version=self.source_version, upgrade_version=self.version,
                 cloud_provider=self.cloud_provider, search_domain=self.search_domain, node_id=self.node_id, default_registry_url=self.default_registry_url,
                 python_image_name=self.python_image_name, managed_containers=bool(self.user_registry_url and self.main_registry_url))
                expatProc.friendly_name = "Post Cluster Upgrade Init"
                expatProc.fail_plan = True
                expatProc.report_callback = self.plan_utils.create_callback_procedure("starting", stage="Post Cluster Upgrade Init", instance_type=self.get_instance_type())
                plan.append(expatProc)
        
        if self.plan_utils.check_file_exists(post_init_check) and expatProcCondition:
            self.add_user_init_script(plan)

        user_init_check = self.plan_utils.get_procedure_check_file(UserInitScript.__name__)
        if self.plan_utils.check_file_exists(user_init_check) and self.plan_utils.check_file_exists(post_init_check) and expatProcCondition and self.plan_utils.check_file_exists(generate_opensearch_users_passwords):
            self.add_report_all_started(plan)

        return plan
    
    def add_generate_opensearch_users_passwords(self, plan: List[Procedure]) -> pathlib.Path:
        gen_check = self.plan_utils.get_procedure_check_file(
            GenerateOpensearchUsersPasswords.__name__)
        if not self.plan_utils.check_file_exists(gen_check) \
            and self.plan_utils.has_started_enabled_consul():
            logger.info("%s - Adding procedure %s", self.name, GenerateOpensearchUsersPasswords.__name__)
            generate_pass = GenerateOpensearchUsersPasswords(GenerateOpensearchUsersPasswords.__name__, self.config, gen_check)
            generate_pass.friendly_name = "Configuring Opensearch users"
            generate_pass.report_callback = self.plan_utils.create_callback_procedure(
                "starting", stage="Configuring Opensearch users", instance_type=self.instance_type)
            plan.append(generate_pass)
        return gen_check


class WorkerPostInitPlan(BasePostInitPlan):
    def init(self, name: str, config: ConfigParser, **kwargs: str) -> None:
        super().init(name, config, **kwargs)
        self.instance_type = "worker"
     
    def create(self) -> Sequence[Procedure]:    
        plan: List[Procedure] = []
        parentPlan: Sequence[Procedure] = super().create()
        plan += list(parentPlan)

        self.add_start_enable_services(plan)
        start_services_checked = self.plan_utils.get_procedure_check_file("{}__started".format(StartServices.__name__))

        if self.plan_utils.check_file_exists(start_services_checked):
            self.add_user_init_script(plan)

        user_init_check = self.plan_utils.get_procedure_check_file(UserInitScript.__name__)
        if self.plan_utils.check_file_exists(user_init_check):
            self.add_report_all_started(plan)

        return plan


class SecondaryPostInitPlan(BasePostInitPlan):
    def init(self, name: str, config: ConfigParser, **kwargs: str) -> None:
        super().init(name, config, **kwargs)
        self.instance_type = "secondary"
     
    def create(self) -> Sequence[Procedure]:    
        plan: List[Procedure] = []
        parentPlan: Sequence[Procedure] = super().create()
        plan += list(parentPlan)

        update_opensearch_users_pass_check = self.add_update_opensearch_users_passwords(plan)

        self.add_user_init_script(plan)

        user_init_check = self.plan_utils.get_procedure_check_file(UserInitScript.__name__)
        if self.plan_utils.check_file_exists(user_init_check) and self.plan_utils.check_file_exists(update_opensearch_users_pass_check):
            self.add_report_all_started(plan)

        return plan

    def add_update_opensearch_users_passwords(self, plan: List[Procedure]) -> pathlib.Path:
        gen_check = self.plan_utils.get_procedure_check_file(
            UpdateOpensearchUsersPasswords.__name__)
        if not self.plan_utils.check_file_exists(gen_check) \
            and self.plan_utils.has_started_enabled_consul():
            logger.info("%s - Adding procedure %s", self.name, UpdateOpensearchUsersPasswords.__name__)
            generate_pass = UpdateOpensearchUsersPasswords(UpdateOpensearchUsersPasswords.__name__, self.config, gen_check, self.services_list)
            generate_pass.friendly_name = "Configuring Opensearch users"
            generate_pass.report_callback = self.plan_utils.create_callback_procedure(
                "starting", stage="Configuring Opensearch users", instance_type=self.instance_type)
            plan.append(generate_pass)
        return gen_check