import configparser
import logging
import pathlib

from typing import List

from ec2init import Procedure
from ec2init.plans.plan_utils import HopsPlanUtils
from ec2init.plans.init.start_services import ServiceNotEnabledException, StartServices
from ec2init.utils.exceptions import Ec2InitFatalException

logger = logging.getLogger(__name__)

class ReportAllStarted(Procedure):
    def __init__(self, name: str, config: configparser.ConfigParser, plan_utils: HopsPlanUtils, check_file: pathlib.Path, gateway: str, uid: str, instance_id: str, api_key: str, instance_type: str) -> None:
        self.name = name
        self.config = config
        self.check_file = check_file
        self.plan_utils = plan_utils
        self.gateway = gateway
        self.uid = uid
        self.instance_id = instance_id
        self.api_key = api_key
        self.instance_type = instance_type

    def execute(self) -> List[Procedure]:
        procedures: List[Procedure] = []
        logger.info(f"{self.name} - Executing")
        try:
            self._check_all_services_already_started()
        except Exception as e:
            # Service not enabled is fatal, it does not worth retrying as
            # nothing will change this state
            if isinstance(e, ServiceNotEnabledException):
                logger.error(f"{self.name} - {str(e)}")
                raise Ec2InitFatalException("Fatal error") from e
            raise e
        logger.info(f"{self.name} - Making callback that all services have started")
        if not self.plan_utils.check_file_exists(self.check_file):
            proc = self.plan_utils.create_callback_procedure("started", check_file=self.check_file, instance_type=self.instance_type)
            proc.friendly_name = "Report instance has started"
            procedures.append(proc)

        return procedures


    def _check_all_services_already_started(self) -> None:
        logger.debug(f"{self.name} - Checking if all services are enabled and started")
        cluster_status = self.plan_utils.get_cluster_status(self.name, self.config, self.gateway, self.uid, self.instance_id, self.api_key)
        all_head_nodes = [node for node in cluster_status["payload"] if node["type"] == "master" or node["type"] == "secondary"]
        # if one or more head nodes are running, then there is no need to check if all services are running
        # some services could be restarting (kafka/zookeeper for example) waiting for other services in other head nodes to join
        if len(all_head_nodes) == 1:
            StartServices.all_services_started_enabled(self.config)
        logger.info(f"{self.name} - All Hopsworks services are enabled and started")

    def rollback(self) -> None:
        pass