import pathlib
from ec2init.plans.plan_utils import HopsPlanUtils
import logging
import json
import configparser

from typing import List

from ec2init import Procedure
from ec2init.utils.system import OSType, SystemUtils
from ec2init.plans.plan_utils import HopsPlanUtils
from ec2init.utils.cloud import get_cloud_provider, Azure
import time
from ec2init.utils.hopsfs import HopsFSUtils

logger = logging.getLogger(__name__)

class NoDataVolumeException(Exception):
    pass

class WrongLabelForDataVolumeException(Exception):
    pass 

class GrowDataVolume(Procedure):

    DATA_VOLUME_LABEL = "hw_data_01"

    def __init__(self, name: str, config: configparser.ConfigParser, plan_utils: HopsPlanUtils, is_backup: bool) -> None:
        self.name = name
        self.config = config
        self.plan_utils = plan_utils
        self.is_backup = is_backup

    def execute(self) -> List[Procedure]:
        logger.info(f"{self.name} - Starting execution")
        mount_point = self._get_data_mount_point()

        try:
            device = self._get_data_block_device(mount_point)

            logger.info(f"{self.name} - Growing partition 1 of device {device}")
            SystemUtils.execute_command(['growpart', device, '1'], timeout=120)

            logger.info(f"{self.name} - Growing filesystem {mount_point}")
            SystemUtils.execute_command(['xfs_growfs', '-d', mount_point], timeout=120)

            if self.is_backup:
                HopsFSUtils.empty_datanode()

            logger.info(f"{self.name} - Finished growing partition")
            self.done()
            return []
        except NoDataVolumeException as ex:
            logger.info(f"{self.name} - {str(ex)}")
            self.done()
            return []

    def done(self) -> None:
        check_file = self.plan_utils.get_procedure_check_file(GrowDataVolume.__name__)
        check_file.touch()

    def rollback(self) -> None:
        pass

    def _get_data_mount_point(self) -> str:
        return self.config.get('initplan', 'data_mount_point', fallback='/hopsworks_data')

    def _get_data_block_device(self, mount_point: str) -> str:
        os_type = SystemUtils.get_os_type()
        if os_type == OSType.UBUNTU:
            return self._ubuntu_get_data_block_device(mount_point)
        elif os_type == OSType.CENTOS:
            return self._centos_get_data_block_device(mount_point)
        raise Exception(f"Unknown OS type {os_type}")

    def _get_data_filesystem(self, mount_point: str) -> str:
        os_type = SystemUtils.get_os_type()
        if os_type == OSType.UBUNTU:
            return self._ubuntu_get_data_filesystem(mount_point)
        elif os_type == OSType.CENTOS:
            return self._centos_get_data_filesystem(mount_point)
        raise Exception(f"Unknown OS type {os_type}")

    def _ubuntu_get_data_block_device(self, mount_point: str) -> str:
        filesystem = self._ubuntu_get_data_filesystem(mount_point)
        result = SystemUtils.execute_command(['lsblk', '-J', '-no', 'label,pkname', filesystem])
        dev_out = json.loads(result.stdout)
        return f"/dev/{dev_out['blockdevices'][0]['pkname']}"

    def _ubuntu_get_data_filesystem(self, mount_point: str) -> str:
        result = SystemUtils.execute_command(['findmnt', '-J', '-o', 'label,source', mount_point], check=False)
        if result.returncode != 0:
            raise NoDataVolumeException(f"No mount point {mount_point}")
        fs_out = json.loads(result.stdout)
        filesystems = fs_out['filesystems']
        if len(filesystems) == 0:
            # That's strange but it's better to be on the safe side
            error_msg = f"{self.name} - Could not format data volume, mount point {mount_point} was found but with zero filesystems"
            logger.error(error_msg)
            raise Exception(error_msg)

        data_filesystem = filesystems[0]
        if data_filesystem['label'] != GrowDataVolume.DATA_VOLUME_LABEL:
            # That is also strange
            error_msg = f"{self.name} - Could not format data volume, mount point {mount_point} has label {data_filesystem['label']} while expecting {GrowDataVolume.DATA_VOLUME_LABEL}"
            logger.error(error_msg)
            raise WrongLabelForDataVolumeException(error_msg)

        filesystem = data_filesystem['source']
        return filesystem

    def _centos_get_data_block_device(self, mount_point: str) -> str:
        filesystem = self._centos_get_data_filesystem(mount_point)

        result = SystemUtils.execute_command(['lsblk', '-no', 'pkname', filesystem])
        dev = result.stdout.decode('utf-8').strip()
        if dev == "":
            error_msg = f"{self.name} - Could not format data volume. Device of filesystem {filesystem} is empty"
            logging.error(error_msg)
            raise Exception(error_msg)
        return f"/dev/{dev}"

    def _centos_get_data_filesystem(self, mount_point: str) -> str:
        result = SystemUtils.execute_command(['findmnt', '-n', '-o', 'label,source', mount_point], check=False)
        if result.returncode != 0:
            raise NoDataVolumeException(f"No mount point {mount_point}")
        stdout = result.stdout.decode('utf-8').strip()
        out = stdout.split(' ')
        if len(out) != 2:
            error_msg = f"{self.name} - Could not format data volume. Unexpected format of findmnt output -> {result}"
            logger.error(error_msg)
            raise Exception(error_msg)
        
        label = out[0]
        filesystem = out[1]
        if label != GrowDataVolume.DATA_VOLUME_LABEL:
            error_msg = f"{self.name} - Could not format data volume, mount point {mount_point} has label {label} while expecting {GrowDataVolume.DATA_VOLUME_LABEL}"
            logger.error(error_msg)
            raise WrongLabelForDataVolumeException(error_msg)
        return filesystem

"""
When we take a backup of a cluster we essentially create a snapshot of the data volume of the
Head node.

In AWS, when we restore the backup, we use the upstream Image and replace the existing data volume
with the backed up user data snapshot (internally it's converted to volume).

In Azure the process is the same but it does not allow us to replace the data volume of an Image.
To restore a backup, we create a new VM starting from a gecko Hopsworks Gallery Image and we attach
the Data Volume that we snapshot during backup. At this point the VM will have two Data Volumes.

Data volume filesystems are labeled with 'hw_data_01' In /etc/fstab we mount any filesystem with
label 'hw_data_01' under '/hopsworks_data' and this will be the Data Volume in use.

Remember that after restoring a snapshot volume, we endup with two data volumes - one from the gecko
image and another from the user's backup. Both have the same filesystem label so need to change the
label of the gecko data volume and remout the filesystems. This will result with only the user's
filesystem (from the snapshot) having the label 'hw_data_01' and this will be the volume mounted
under '/hopsworks_data'
"""
class FixBackupDataVolume(Procedure):
    # Gecko data volume SHOULD ALWAYS have LUN 0
    GECKO_DATA_LUN = "/dev/disk/azure/scsi1/lun0"
    GECKO_FS_NEW_LABEL = "hw_gecko_0"

    def __init__(self, name: str, config: configparser.ConfigParser, plan_utils: HopsPlanUtils) -> None:
        self.name = name
        self.config = config
        self.plan_utils = plan_utils
        
    def execute(self) -> List[Procedure]:
        logger.info(f"{self.name} - Starting execution")
        cloud_provider = get_cloud_provider(self.plan_utils.cloud_provider)
        if cloud_provider.__name__ != Azure.__name__:
            logger.warn(f"{self.name} - Skip fixing data volume because cloud provider is not Azure")
            self.done()
            return []

        gdv = GrowDataVolume(FixBackupDataVolume.__name__, self.config, self.plan_utils, True)
        mount_point = gdv._get_data_mount_point()
        try:
            device_name = gdv._get_data_block_device(mount_point)
            gecko_device = pathlib.Path(FixBackupDataVolume.GECKO_DATA_LUN).resolve()
            if device_name != str(gecko_device):
                logger.info(f"{self.name} - No unmount, just changing the fs label. {mount_point} points to {device_name} while Image Data Volume device is {gecko_device}")
                SystemUtils.execute_command(f"xfs_admin -L {FixBackupDataVolume.GECKO_FS_NEW_LABEL} {gecko_device}1".split(" "))
                self.done()
                return []

            # Now we have established that the filesystem mounted under /hopsworks_data is the Data Volume
            # of the Image instead of the backup snapshot
            # We must unmount the filesystem, change the label from 'hw_data_01' to something else
            # and re-mount the filesystems
            filesystem = gdv._get_data_filesystem(mount_point)
            logger.info(f"{self.name} - Changing the label of {filesystem} to {FixBackupDataVolume.GECKO_FS_NEW_LABEL}")
            # But first we need to stop Docker which is running and is keeping open file descriptors to /hopsworks_data
            SystemUtils.execute_command("systemctl stop docker".split(" "), timeout=60)
            # Retry to avoid cases when umount fails silently 
            for _ in range(10):
                SystemUtils.execute_command(f"umount {mount_point}".split(" "), timeout=30)
                try:
                    time.sleep(5)
                    gdv._get_data_filesystem(mount_point)
                except NoDataVolumeException as ex:
                    logger.debug(ex)
                    break
            SystemUtils.execute_command(f"xfs_admin -L {FixBackupDataVolume.GECKO_FS_NEW_LABEL} {filesystem}".split(" "))

            # For some reason on AZURE mount -a mounted the image disk instead of the data disk
            # So, we need to make sure that mount is taking fstab changes in consideration
            for _ in range(10):
                time.sleep(5)
                wrongLabelException = None
                SystemUtils.execute_command("mount -a".split(" "), timeout=40)
                try:
                    gdv._get_data_filesystem(mount_point)
                    break
                except WrongLabelForDataVolumeException as ex:
                    wrongLabelException = ex
                    logger.debug(ex)
                    SystemUtils.execute_command(f"umount {mount_point}".split(" "), timeout=30)
            
            if wrongLabelException is not None:
                raise wrongLabelException

            SystemUtils.execute_command("systemctl start docker".split(" "), timeout=360)
            logger.info(f"{self.name} - Finished fixing data volume for restoring backup")
            self.done()
            return []
        except NoDataVolumeException as ex:
            logger.info(f"{self.name} - VM does not have data volume: ${str(ex)}")
            self.done()
            return []

        return []

    def done(self) -> None:
        check_file = self.plan_utils.get_procedure_check_file(FixBackupDataVolume.__name__)
        check_file.touch()

    def rollback(self) -> None:
        pass
