import typing
from ec2init.plans.init.revoke_certificates import RevokeCertificates
from ec2init.utils.bucket_configuration import BucketConfiguration, S3BucketEncryptionConfiguration, S3GeneralConfiguration
from ec2init.utils.database import databaseUtils
import logging
import configparser
import pathlib
import subprocess
import os
import xml.etree.ElementTree as ET
import json
import time
import glob
import shutil
import base64


from pathlib import Path

from typing import List, Tuple, Dict, Optional

from ec2init import Procedure
from ec2init.utils.services import Service, ServiceUtils
from ec2init.utils.system import OSType, SystemUtils
from ec2init.utils.cloud import CloudProvider, get_cloud_provider
from ec2init.plans.plan_utils import PlanUtils
from ec2init.plans.multinode.master_seds import MasterSeds
from ec2init.plans.multinode.register_generate_certs import RegisterGenerateCerts
from ec2init.utils.hopsfs import HopsFSUtils
from ec2init.utils.offline_backup_information import OffLineBackupInformation
from ec2init.utils.shell import ShellUtils
from ec2init.plans.init.setup_ecr_docker_images import SetupDockerImages
from ec2init.utils.docker import DockerImageName
from ec2init.utils.constants import Scripts
from ec2init.utils.versions import VersionUtils

logger = logging.getLogger(__name__)
    
class BasePreInit(Procedure):
    def __init__(self, name: str, config: configparser.ConfigParser, check_file: pathlib.Path,
        bucket_configuration: BucketConfiguration, region: str, cloud_provider: str, search_domain: str, node_id: str):
        self.name = name
        self.config = config
        self.check_file = check_file
        self.bucket_configuration = bucket_configuration
        self.region = region
        self.cloud_provider = cloud_provider
        self.search_domain = search_domain
        self.node_id = node_id

    def pre(self) -> None:
        if self.report_callback is not None:
            try:
                self.report_callback.execute()
            except Exception as ex:
                logger.warning("%s - Could not execute pre-execute: %s", self.name, ex)

    def execute(self) -> List[Procedure]:
        return []

    def rollback(self) -> None:
        pass

    def get_local_name(self) -> str:
        try:
            return get_cloud_provider(self.cloud_provider).get_fqdn(self.config, self.search_domain, self.node_id)
        except Exception as ex:
            self.friendly_name = str(ex)
            raise ex

    def get_local_ip(self) -> str:
        try:
            return get_cloud_provider(self.cloud_provider).get_private_ip(self.config)
        except Exception as ex:
            self.friendly_name = "Cluster initialization failed, could not get local IP address"
            raise ex

    def sysexec(self, cmd: List[str], timeout: int = 120) -> subprocess.CompletedProcess:
        return SystemUtils.execute_command(cmd, timeout=timeout)

    def fix_hopsfs_s3_conf_attrs(self) -> None:

        if self.bucket_configuration.has_bucket():
            HopsFSUtils.fix_hdfs_site("dfs.replication", "1")
            HopsFSUtils.fix_hdfs_site("dfs.enable.cloud.persistence", "true"); 
            HopsFSUtils.fix_hdfs_site("dfs.cloud.aws.s3.region", self.region); 
            HopsFSUtils.fix_hdfs_site("dfs.cloud.aws.s3.bucket", self.bucket_configuration.name); 
            HopsFSUtils.fix_hdfs_site("dfs.disable.non.cloud.storage.policies", "true")
            if self.bucket_configuration.general_configuration:
                s3_general_conf = typing.cast(S3GeneralConfiguration, self.bucket_configuration.general_configuration)
                if s3_general_conf.acl_bucket_full_control:
                    HopsFSUtils.fix_hdfs_site("dfs.cloud.aws.amz.acl.bucket.owner.full.control.enable", str(s3_general_conf.acl_bucket_full_control).lower())

            if self.bucket_configuration.encryption_configuration.is_encryption_enabled():
                bucket_encryption_conf = typing.cast(S3BucketEncryptionConfiguration, self.bucket_configuration.encryption_configuration)
                HopsFSUtils.fix_hdfs_site("dfs.cloud.aws.server.side.encryption.enable", "true")
                HopsFSUtils.fix_hdfs_site("dfs.cloud.aws.server.side.encryption.type", bucket_encryption_conf.encryption_mode)
                if bucket_encryption_conf.is_kms_enabled():
                    HopsFSUtils.fix_hdfs_site("dfs.cloud.aws.server.side.encryption.bucket.key.enable", str(bucket_encryption_conf.bucket_key).lower())
                    if not bucket_encryption_conf.is_managed_kms():
                        HopsFSUtils.fix_hdfs_site("dfs.cloud.aws.server.side.encryption.key.arn", bucket_encryption_conf.user_key_arn)
        else:
            HopsFSUtils.fix_hdfs_site("dfs.enable.cloud.persistence", "false"); 
            HopsFSUtils.fix_hdfs_site("dfs.cloud.aws.s3.region", ""); 
            HopsFSUtils.fix_hdfs_site("dfs.cloud.aws.s3.bucket", "");

    def sed_stmt(self, regex: str, file: str) -> str:
        return f"sed -i '{regex}' {file}"

    def sed_cmd(self, regex: str, file: str) -> List[str]:
        return ["sh", "-c", self.sed_stmt(regex, file)]

    def touch_check_file(self) -> None:
        if self.check_file is not None:
            self.check_file.touch()

    def update_kagent_password(self, password) -> None:
        self._config_update_kagent_password(password)

    def _config_update_kagent_password(self, password) -> None:
        kagent_conf_dir = self.config.get(
            'general', 'kagent_conf_dir', fallback="/srv/hops/kagent/etc")
        kagent_conf = Path(kagent_conf_dir, "config.ini")
        self.sysexec(self.sed_cmd(f"s/password[[:space:]]=.*/password = {password}/g", str(kagent_conf)))

    def delete_datanode_directories(self) -> None:
        data_dir = Path(self.config.get('initplan', 'hops_data_dir', fallback='/srv/hops/hopsdata'))
        dn_base_dir = data_dir / "hdfs" / "dn"

        cloud_dn_dir = dn_base_dir / "cloud" / "*"
        dirs_to_delete: List[str] = []
        dirs_to_delete += glob.glob(str(cloud_dn_dir))

        disk_dn_dir =  dn_base_dir / "disk" / "*"
        dirs_to_delete += glob.glob(str(disk_dn_dir))

        for d in dirs_to_delete:
            if os.path.isdir(d):
                shutil.rmtree(d)
            else:
                os.remove(d)

class BrainsPreInit(BasePreInit):
    def __init__(self, name: str, config: configparser.ConfigParser, check_file: pathlib.Path,
        bucket_configuration: BucketConfiguration, region: str, ldap: str, oauth: str, password_file: pathlib.Path, user_email: str,
        gateway: str, user_id: str, instance_id: str, api_key: str, cloud_provider: str, is_upgrade: bool, search_domain: str, node_id: str, user_mgm_type: str,
        create_pwd_user: bool, is_backup: bool, offline_backup_info: OffLineBackupInformation, cluster_domain_prefix: str):
        super().__init__(name, config, check_file, bucket_configuration, region, cloud_provider, search_domain, node_id)
        self.ldap = ldap
        self.oauth = oauth
        self.password_file = password_file
        self.user_email = user_email
        self.gateway = gateway
        self.user_id = user_id
        self.instance_id = instance_id
        self.api_key = api_key
        self.is_upgrade = is_upgrade
        self.plan_utils = PlanUtils(PlanUtils.create_checks_directory(config))
        self.offline_backup_info = offline_backup_info
        self.is_backup = is_backup
        logger.info("is_backup: " + str(self.is_backup))
        self.user_mgm_type = user_mgm_type
        self.create_pwd_user = create_pwd_user
        self.cluster_domain_prefix = cluster_domain_prefix
        if OffLineBackupInformation.is_backup(self.offline_backup_info):
            kagent_conf_dir = self.config.get('general', 'kagent_conf_dir', fallback="/srv/hops/kagent/etc")
            services_file = pathlib.Path(kagent_conf_dir, "services")
            services = configparser.ConfigParser()
            services.read(str(services_file))
            services_to_skip = [
                'service_does_not_exist',
                'glassfish-domain1',
                'docker',
                # Skip database. The configuration is updated and the database is restarted in database_init.py
                'ndb_mgmd',
                'ndbmtd',
                'mysqld'
            ]
            for section in services.sections():
                service = services.get(section, 'service', fallback='service_does_not_exist')
                #do not stop
                if service not in services_to_skip:
                    logger.debug("%s - Stopping service: %s", self.name, service)
                    self._stop_service(service)
                    logger.info("%s - Stopped service: %s", self.name, service)
            lets_encrypt_dir = self.config.get('initplan', 'lets_encrypt_dir', fallback="/etc/letsencrypt")
            script = pathlib.Path(lets_encrypt_dir, 'renewal-hooks', 'deploy', 'deploy2glassfish_hook.sh')
            if script.is_file() :
                script.unlink()
            SystemUtils.execute_command(["systemctl", "stop", "consul"], timeout=120)
            SystemUtils.execute_command(["systemctl", "disable", "ec2update"], timeout=120)

            
    def substitute_in_db(self, local_name: str, local_ip: str) -> None:
        subs: List[Tuple[str, str]] = [
            ("hopsworks_endpoint", local_name),
            ("yarn_ui_ip", local_ip),
            ("elastic_ip", local_ip),
            ("kibana_ip", local_ip),
            ("localhost", "false")
        ]

        if not self.is_upgrade:
            ldap_enabled = "true" if self.ldap else "false"
            oauth_enabled = "true" if self.oauth else "false"
            subs.append(("ldap_auth", ldap_enabled))
            subs.append(("oauth_enabled", oauth_enabled))

        for s in subs:
            k, v = s
            databaseUtils.update_variable(k, v)

    def fix_hopsfs_s3_db_attrs(self) -> None:
        if self.bucket_configuration.has_bucket():
            databaseUtils.run_command(["hops", "-e", "truncate hdfs_cloud_buckets"])
            databaseUtils.run_command(["hops", "-e", f"insert into hdfs_cloud_buckets values (1, '{self.bucket_configuration.name}')"])
            databaseUtils.run_command(["hops", "-e", f"UPDATE hdfs_inodes SET storage_policy=3 where is_dir = 1 and storage_policy != 0"])
            databaseUtils.run_command(["hops", "-e", f"UPDATE hdfs_inodes SET storage_policy=3 WHERE id=1"])

            databaseUtils.update_variable("hdfs_base_storage_policy", "CLOUD")
            databaseUtils.update_variable("hdfs_log_storage_policy", "CLOUD")
        else:
            databaseUtils.run_command(["hops", "-e", "truncate hdfs_cloud_buckets"])
            databaseUtils.run_command(["hops", "-e", f"UPDATE hdfs_inodes SET storage_policy=7 WHERE id=1"])

            databaseUtils.update_variable("hdfs_base_storage_policy", "DB")
            databaseUtils.update_variable("hdfs_log_storage_policy", "HOT")

    def update_hopsfs_time_stamps(self) -> None:
           ts = int(round(time.time() * 1000))
           logger.info("Setting inodes and block time stamp to " + str(ts))
           databaseUtils.run_command(["hops", "-e", f"UPDATE hdfs_inodes SET access_time = {ts}, modification_time = {ts}"])
           databaseUtils.run_command(["hops", "-e", f"UPDATE hdfs_block_infos SET time_stamp = {ts}"])

    def generate_new_jwt_signing_key(self) -> None:
        if OffLineBackupInformation.is_backup(self.offline_backup_info):
            return
        logger.info("Replacing JWT signing key")
        rand = os.urandom(64)
        encoded_rand = base64.b64encode(rand).decode('utf-8')
        sql_update = f"REPLACE INTO jwt_signing_key (secret, name) values ('{encoded_rand}', 'apiKey')"
        databaseUtils.run_command(["hopsworks", "-e", sql_update])

    def create_hopsworks_user(self) -> None:
        if self.is_backup or self.is_upgrade:
            #if we are recovering from a backup the user should already exist in the database
            #also if we are upgrading, since this will be handled by the upgrade_cluster proc
            logger.info("Skip creating new Hopsworks user during backup restore and upgrade")

            # We may still need to do some call with the admin user. For this purpose we reactivate it
            # and update its password, it will be deactivated at the end of the process.
            databaseUtils.run_command(["hopsworks", "-e","UPDATE users SET status=2 where username='meb10000'"])
            databaseUtils.update_variable("disable_password_login", "false")

            process=self.sysexec(["openssl", "rand", "-base64", "32"])
            password=process.stdout.strip().decode("utf-8")
            with open(self.password_file, "w") as fd:
                fd.write(password)

            salt=self.get_existing_user_salt("meb10000")
            sha=self.password_digest(password, salt)
            databaseUtils.run_command(["hopsworks", "-e",f"UPDATE users SET password='{sha}' where username='meb10000'"])
            return
        if self.config.getboolean('general', 'unmanaged', fallback=False):
            logger.warning("Skip creating new Hopsworks user because it's an unmanaged AMI")
            return
        process = self.sysexec(["openssl", "rand", "-base64", "32"])
        password = process.stdout.strip().decode("utf-8")
        with open(self.password_file, "w") as fd:
            fd.write(password)

        salt = self.get_existing_user_salt("meb10000")
        sha = self.password_digest(password, salt)
        
        # self.create_pwd_user is true by default if not set and not set for user management == managed and disabled
        if self.user_mgm_type.upper() != 'MANAGED' and self.create_pwd_user:
            process = SystemUtils.execute_command([f"{Scripts.MYSQL_CLIENT} hopsworks -e 'select max(uid) from users' -sN | tail -1"],
                 timeout=120, shell=True)
            uuid = process.stdout.strip().decode("utf-8")
            uuid = str(int(uuid) + 1)
            databaseUtils.run_command(["hopsworks",  "-e",
                 "INSERT INTO users (uid, username, password, email, fname, lname, activated, false_login, isonline, secret, mode, password_changed, status, max_num_projects, two_factor, tours_state , salt)" +
                f"VALUES ('{uuid}', 'meb{uuid}', '{sha}', '{self.user_email}', 'Admin', 'Admin', '2015-05-15 10:22:36', -1, 0, 'V3WBPS4G2WMQ53VA', 0, '2015-04-28 15:18:42', 2, 30, 0, 3, '{salt}')"])

            databaseUtils.run_command(["hopsworks", "-e",
                 f"INSERT INTO account_audit (initiator, target, action, action_timestamp, message, outcome, ip, useragent) values ({uuid}, {uuid}, 'REGISTRATION', CURRENT_TIMESTAMP, 'Admin account was enabled', 'SUCCESS', '127.0.0.1', 'Firefox')"])

            databaseUtils.run_command(["hopsworks", "-e",
                 f"REPLACE INTO user_group (uid, gid) VALUES ({uuid}, 1005)"])

        databaseUtils.run_command(["hopsworks", "-e",
             f"UPDATE users SET password='{sha}' where username='meb10000'"])

    def update_kagent_password(self, password) -> None:
        self._db_update_kagent_password(password)
        self._config_update_kagent_password(password)
    
    def update_onlinefs_password(self, password) -> None:
        self._db_update_onlinefs_password(password)

    def _db_update_kagent_password(self, password) -> None:
        if self.is_backup:
            return
        self._db_update_user_password("agent007", password)

    def _db_update_onlinefs_password(self, password) -> None:
        if self.is_backup:
            return
        self._db_update_user_password("onlinefs", password)
    
    def _db_update_user_password(self, username, password) -> None:
        salt = self.get_existing_user_salt(username)
        sha = self.password_digest(password, salt)
        sql_update = f"UPDATE users SET password='{sha}' where username='{username}'"
        databaseUtils.run_command(["hopsworks", "-e", sql_update])

    def get_existing_user_salt(self, username) -> str:
        process = SystemUtils.execute_command([f"{Scripts.MYSQL_CLIENT} hopsworks -e \"select salt from users where username='{username}'\" -sN | tail -1"], timeout=120, shell=True)
        return process.stdout.strip().decode("utf-8")

    def password_digest(self, password, salt) -> str:
        salted_password = password + salt
        process = SystemUtils.execute_command([f"echo -n {salted_password} | sha256sum | cut -d \" \" -f 1"], timeout=120, shell=True)
        return process.stdout.strip().decode("utf-8")

    def update_cloud_callback(self) -> None:
        # add cloud_events_endpoint to the variables table"
        databaseUtils.run_command(["hopsworks", "-e",
             f"REPLACE INTO variables (id, value) values ('cloud_events_endpoint', '{self.gateway}/eventsv2/ops/users/{self.user_id}/instances/{self.instance_id}')"])

        databaseUtils.run_command(["hopsworks", "-e",
             f"REPLACE INTO variables (id, value) values ('cloud_events_endpoint_api_key', '{self.api_key}')"])

    def update_hopsworks_public_host(self) -> None:
        public_domain_name = CloudProvider.get_public_domain_name(self.config, self.instance_id, self.cluster_domain_prefix)
        databaseUtils.run_command(["hopsworks", "-e",
             f"REPLACE INTO variables (id, value) values ('hopsworks_public_host', '{public_domain_name}')"])

    def restart_spark_flink_history_servers(self) -> None:
        #flink and spark history server are already started by systemd and need to restart to use new conf
        self.sysexec(["systemctl", "stop", "flinkhistoryserver"])
        self.sysexec(["systemctl", "stop", "sparkhistoryserver"])

    def _stop_service(self, service: str) -> None:
        try:
            SystemUtils.execute_command(["systemctl", "stop", service], timeout=120)
        except Exception as ex:
            logger.error("%s - Could not stop service: %s", service, ex)
    
    def deny_outside_access_to_yarn_ui(self) -> None:
        logger.info("Deny outside access to YARN")
        try:
            SystemUtils.execute_command(["iptables", "-A", "INPUT", "-i", "lo", "-p", "tcp", "--dport", "8088", "-j", "ACCEPT"])
            SystemUtils.execute_command(["iptables", "-A", "INPUT", "-p", "tcp", "--dport", "8088", "-j", "DROP"])
            os_type = SystemUtils.get_os_type()
            logger.info(f"{self.name} - OS type is: {os_type}")
            if os_type == OSType.UBUNTU:
                SystemUtils.execute_command(["iptables-save | tee /etc/iptables/rules.v4"], shell=True)
                SystemUtils.execute_command(["ip6tables-save | sudo tee /etc/iptables/rules.v6"], shell=True)
            elif os_type == OSType.CENTOS:
                cmd = "systemctl enable iptables"
                SystemUtils.execute_command(cmd.split(" "))
                cmd = "/usr/libexec/iptables/iptables.init save"
                SystemUtils.execute_command(cmd.split(" "))
            logger.info(f"{self.name} - Finished setting iptables rules")
        except Exception as ex: 
            logger.error("Could not add iptables rules to deny access to YARN web ui: %s", ex)
            raise ex
    
class MasterPreInit(BrainsPreInit):

    def __init__(self, name: str, config: configparser.ConfigParser, check_file: pathlib.Path,
        bucket_configuration: BucketConfiguration, region: str, ldap: str, oauth: str, password_file: pathlib.Path, user_email: str,
        gateway: str, user_id: str, instance_id: str, api_key: str, cloud_provider: str, user_mgm_type: str, create_pwd_user: bool,
        is_upgrade: bool, source_version: str, upgrade_version: str, search_domain: str, node_id: str, is_backup: bool, offline_backup_info: OffLineBackupInformation,
        cluster_domain_prefix: str, services_list : List[Service], k8s_enabled: bool, enable_flyingduck: bool, nlb_domain: str):
        super().__init__(name, config, check_file, bucket_configuration, region, ldap, oauth, password_file, user_email,
            gateway, user_id, instance_id, api_key, cloud_provider, is_upgrade, search_domain, node_id, user_mgm_type, create_pwd_user,
            is_backup, offline_backup_info, cluster_domain_prefix)
        self.source_version = source_version
        self.upgrade_version = upgrade_version
        self.services_list = services_list
        self.k8s_enabled = k8s_enabled
        self.enable_flyingduck = enable_flyingduck
        self.nlb_domain = nlb_domain

    def execute(self) -> List[Procedure]:
        local_name = self.get_local_name()
        local_ip = self.get_local_ip()

        if OffLineBackupInformation.is_backup(self.offline_backup_info):
            sed_check = self.plan_utils.get_procedure_check_file(MasterSeds.__name__)
            self.plan_utils.remove_check_file(sed_check)
            register_generate_cert_check = self.plan_utils.get_procedure_check_file(RegisterGenerateCerts.__name__)
            self.plan_utils.remove_check_file(register_generate_cert_check)
            revoke_previous_certificates = self.plan_utils.get_procedure_check_file(RevokeCertificates.__name__)
            self.plan_utils.remove_check_file(revoke_previous_certificates)
            
        self.run_flyway_on_upgrade()
        self.substitute_in_db(local_name, local_ip)

        databaseUtils.run_command(["hopsworks", "-e", f"DELETE FROM hopsworks.hosts"])
        databaseUtils.run_command(["hopsworks", "-e",
             f"INSERT INTO hopsworks.hosts(hostname, host_ip) VALUES ('{local_name}', '{local_ip}') ON DUPLICATE KEY UPDATE cores=NULL;"])

        if not self.is_upgrade:
            # Disable registration (UI)
            if self.user_mgm_type.upper() == 'LDAP':
                databaseUtils.update_variable("disable_registration", "true")
            elif self.user_mgm_type.upper() == 'MANAGED':
                databaseUtils.update_variable("disable_registration", "true")

        self.fix_hopsfs_s3_conf_attrs()
        self.fix_hopsfs_s3_db_attrs()
        self.fix_hopsfs_sizing_attrs()
        if not self.k8s_enabled:
            self.fix_docker_limit_attrs()

        if not self.is_backup and not self.is_upgrade:
            self.update_hopsfs_time_stamps()

        self.create_hopsworks_user()
        self.update_kagent_password(self.api_key)
        self.update_onlinefs_password(self.api_key)
        self.generate_new_jwt_signing_key()
        self.update_cloud_callback()
        self.restart_spark_flink_history_servers()
        self.deny_outside_access_to_yarn_ui()
        self.update_hopsworks_public_host()
        
        self.remove_unused_services()
        if not OffLineBackupInformation.is_backup(self.offline_backup_info) and not self.is_upgrade:
            self.clean_up_kafka_data_dir()
        self.clean_up_prometheus_data_dir()

        self.update_pki_ca_configuration_on_upgrade()
        
        # Set FlyingDuck feature flag
        databaseUtils.update_variable("enable_flyingduck", str(self.enable_flyingduck).lower())

        databaseUtils.update_variable("loadbalancer_external_domain", self.nlb_domain)

        self.touch_check_file()
        return []

    def get_oauth_variables_values(self) -> Dict[str, str]:
        oauth_values = {}
        for v in ["oauth_account_status", "oauth_redirect_uri", "oauth_logout_redirect_uri", "oauth_enabled", "oauth_group_mapping", "managed_cloud_redirect_uri", "remote_auth_need_consent"]:
            value = SystemUtils.execute_mysql_statement(f"SELECT value FROM variables WHERE id='{v}'")
            if len(value) > 0:
                oauth_values[v] = value[0][0]
        return oauth_values

    def set_oauth_variables_values(self, values : Dict[str, str]) -> None:
        for k, v in values.items():
            SystemUtils.execute_mysql_statement(f"UPDATE variables SET value='{v}' WHERE id='{k}'")

    def materialize_flyway_schemas(self, flyway_dir : str, hopsworks_versions : List[str]) -> None:
        ShellUtils.rmtree(f"{flyway_dir}/sql", keep_root=True)
        flyway_versions = SystemUtils.execute_mysql_statement("SELECT version, description FROM flyway_schema_history")
        for flyway_version in flyway_versions:
            version, description = flyway_version
            if description.strip() == "initial tables":
                shutil.copy2(f"{flyway_dir}/all/sql/V{version}__initial_tables.sql", f"{flyway_dir}/sql")
            else:
                shutil.copy2(f"{flyway_dir}/all/sql/V{version}__hopsworks.sql", f"{flyway_dir}/sql")
        
        source_version_with_zero_patch = VersionUtils.get_version_with_zero_patch_version(self.source_version)
        start_flyway_migration = False
        for version in hopsworks_versions:
            if source_version_with_zero_patch == version:
                start_flyway_migration = True
                continue
            if start_flyway_migration:
                version_with_zero_patch = VersionUtils.get_version_with_zero_patch_version(version)
                shutil.copy2(f"{flyway_dir}/all/sql/V{version_with_zero_patch}__hopsworks.sql", f"{flyway_dir}/sql")
                
            if self.upgrade_version == version:
                break

    def run_flyway_on_upgrade(self) -> None:
        if self.is_upgrade:
            try:
                logger.info("Upgrading HopsFS Schema")
                SystemUtils.execute_command(['/srv/hops/ndb-hops/flyway/flyway', 'migrate'], timeout=300)
                logger.info("Upgrading Hopsworks Schema")
                hopsworks_versions = self.config.get('upgradeplan', 'hopsworks_versions').split(',')
                hopsworks_flyway_dir = '/srv/hops/domains/domain1/flyway'
                self.materialize_flyway_schemas(hopsworks_flyway_dir, hopsworks_versions)
                # Validation during migrate is very strict for our needs. If we found a bug
                # on a released branch and we fix it, next time someone upgrades to released branch+1
                # they will get a validation error because the files checksum do not match                
                SystemUtils.execute_command([f'{hopsworks_flyway_dir}/flyway', '-validateOnMigrate=false', 'migrate'], timeout=300)

                # Run all dml files during upgrade
                oauth_values = self.get_oauth_variables_values()
                for version in hopsworks_versions:
                    if VersionUtils.less_than_or_equal(version, "3.4.0"):
                        version_with_zero_patch = version
                    else:
                        version_with_zero_patch = VersionUtils.get_version_with_zero_patch_version(version)
                    SystemUtils.execute_command([f'{Scripts.MYSQL_CLIENT} hopsworks < {hopsworks_flyway_dir}/dml/V{version_with_zero_patch}__hopsworks.sql'], shell=True, timeout=300)
                self.set_oauth_variables_values(oauth_values)

                # Run upgrade on hive DB
                logger.info("Upgrading Hive Schema")
                hive_schematool = self.config.get('upgradeplan', 'hive_schematool', fallback='/srv/hops/apache-hive/bin/schematool')
                SystemUtils.execute_command([hive_schematool, '-dbType', 'mysql', '-upgradeSchema'], timeout=300)

            except subprocess.CalledProcessError as ex:
                stdout = "None" if ex.stdout is None else ex.stdout.decode("utf-8")
                stderr = "None" if ex.stderr is None else ex.stderr.decode("utf-8")
                logger.error("Failed to run flyway during upgrade: %s\n stdout: %s \nstderr: %s\n", ex, stdout, stderr)
                raise ex
            except Exception as ex:
                logger.error("Failed to run flyway during upgrade: %s", ex)
                raise ex

    def clean_up_kafka_data_dir(self) -> None:
        logger.info("%s - clean up kafka log.dir and zookeeper data dir", self.name)
        ShellUtils.rmtree("/srv/hops/zookeeper/zookeeper/data", keep_root=True)
        ShellUtils.rmtree("/srv/hops/kafka-logs", keep_root=True)

    def clean_up_prometheus_data_dir(self) -> None:
        logger.info("%s - clean up prometheus data dir", self.name)
        ShellUtils.rmtree("/srv/hops/prometheus/prometheus-data", keep_root=True)
        
    def remove_unused_services(self) -> None:
        if len(self.services_list) > 0:
            logger.info("%s - keep only services - %s", self.name, ','.join([s.value for s in self.services_list]))
            self._remove_unused_kagent()
            self._remove_unused_consul()
        else:
            logger.info("%s - services list is empty - no need to remove any", self.name)
    
    def _remove_unused_kagent(self) -> None:
        kagent_conf_dir = self.config.get(
            'general', 'kagent_conf_dir', fallback="/srv/hops/kagent/etc")
        services_file = pathlib.Path(kagent_conf_dir, "services")
        services = configparser.ConfigParser()
        services.read(str(services_file))

        for service_section in services.sections():
            service = services.get(service_section, 'service', fallback='service_does_not_exist')
            if service != 'service_does_not_exist' and not ServiceUtils.contains(self.services_list, service):
                del services[service_section]
                SystemUtils.execute_command(["systemctl", "stop", service])
                SystemUtils.execute_command(["systemctl", "disable", service])
                SystemUtils.execute_command(["systemctl", "mask", service])
            else:
                logger.info("%s - keep service %s", self.name, str(service))
        
        with open(str(services_file), 'w') as fd:
            services.write(fd) 

    def _remove_unused_consul(self) -> None:
        consul_to_keep = ["consul.hcl", "systemd_env_vars"]
        for service in self.services_list:
            if service == Service.Namenode:
                consul_to_keep.append("nn-consul.hcl")
            elif service == Service.Datanode:
                consul_to_keep.append("dn-consul.hcl")
            elif service == Service.MySQLd:
                consul_to_keep.append("mysql-consul.hcl")
            elif service == Service.MySQLdExporter:
                consul_to_keep.append("mysql-exporter-consul.hcl")
            elif service == Service.ElasticExporter:
                consul_to_keep.append("elastic-consul.hcl")
            elif service == Service.NodeExporter:
                consul_to_keep.append("node-exporter-consul.hcl")
            elif service == Service.Resourcemanager:
                consul_to_keep.append("rm-consul.hcl")
            elif service == Service.Zookeeper:
                consul_to_keep.append("zk-consul.hcl")
            elif service == Service.Kafka:
                consul_to_keep.append("kafka-consul.hcl")
            elif service == Service.Onlinefs:
                consul_to_keep.append("onlinefs.hcl")
            elif service == Service.Hiveserver2:
                consul_to_keep.append("hiveserver2-consul.hcl")
            elif service == Service.Hivemetastore:
                consul_to_keep.append("metastore-consul.hcl")
            elif service == Service.Logstash:
                consul_to_keep.append("logstash-consul.hcl")
            elif service == Service.AirflowWebserver:
                consul_to_keep.append("airflow-consul.hcl")
            elif service == Service.Alertmanager:
                consul_to_keep.append("alertmanager-consul.hcl")
            elif service == Service.Grafana:
                consul_to_keep.append("grafana-consul.hcl")
            elif service == Service.SparkHistoryserver:
                consul_to_keep.append("historyserver-consul.hcl")
            elif service == Service.OpensearchDashboards:
                consul_to_keep.append("kibana-consul.hcl")
            elif service == Service.Livy:
                consul_to_keep.append("livy-consul.hcl")
            elif service == Service.Prometheus:
                consul_to_keep.append("prometheus-consul.hcl")
            elif service == Service.Pushgateway:
                consul_to_keep.append("pushgateway-consul.hcl")
            elif service == Service.Glassfish:
                consul_to_keep.append("glassfish-consul.hcl")
                # does not have a service by itself
                consul_to_keep.append("registry-consul.hcl")
            elif service == Service.GlassfishInstance:
                consul_to_keep.append("glassfish-consul.hcl")
            elif service == Service.ePipe:
                consul_to_keep.append("epipe-consul.hcl")
            
                
        consul_services = os.listdir("/srv/hops/consul/consul.d")
        for consul_config in consul_services:
            if consul_config not in consul_to_keep:
                os.remove(f"/srv/hops/consul/consul.d/{consul_config}")
            else:
                logger.info("%s - keey service consul %s", self.name, consul_config)

    def fix_hopsfs_sizing_attrs(self) -> None:
        multiplier = self.config.getfloat('initplan', 'subtree_executor_limit_multiplier', fallback=4)
        available_cpus = SystemUtils.get_num_cpus()
        subtree_executor_limit = int(multiplier * available_cpus)
        HopsFSUtils.fix_hdfs_site("dfs.namenode.subtree-executor-limit", str(subtree_executor_limit));    

    def fix_docker_limit_attrs(self) -> None:
        databaseUtils.run_command(["hopsworks", "-e", f"REPLACE INTO variables (id, value) values ('docker_cgroup_memory_limit_gb', '{self._get_docker_memory_size()}GB')"])
        databaseUtils.run_command(["hopsworks", "-e", f"REPLACE INTO variables (id, value) values ('docker_cgroup_soft_limit_memory_gb', '{self._get_docker_memory_size_soft()}GB')"])

    def _get_docker_memory_size(self) -> int:
        multiplier = self.config.getfloat('initplan', 'docker_memory_size_multiplier', fallback=0.2)
        available_memory_mb = SystemUtils.get_system_memory()
        docker_memory = int(multiplier * available_memory_mb)
        if docker_memory > available_memory_mb:
            docker_memory = available_memory_mb
        
        docker_memory = max(6*1024, docker_memory)
        return round(docker_memory/1024)

    def _get_docker_memory_size_soft(self) -> int:
        multiplier = self.config.getfloat('initplan', 'docker_memory_size_soft_multiplier', fallback=0.07)
        available_memory_mb = SystemUtils.get_system_memory()
        docker_memory = int(multiplier * available_memory_mb)
        if docker_memory > available_memory_mb:
            docker_memory = available_memory_mb
        
        docker_memory = max(2*1024, docker_memory)
        return round(docker_memory/1024)
    
    def update_pki_ca_configuration_on_upgrade(self) -> None:
        rows = SystemUtils.execute_mysql_statement(f"SELECT ca, subject FROM pki_certificate WHERE ca=0 AND subject='C=SE,ST=Sweden,L=Stockholm,O=SICS,CN=HopsRootCA'")
        if self.is_upgrade and len(rows) > 0:
            config = {
                "rootCA": {
                    "x509Name": "C=SE, ST=Sweden, L=Stockholm, O=SICS, CN=HopsRootCA",
                    "validityDuration": "3650d"
                },
                "intermediateCA" : {
                    "x509Name": "C=SE, ST=Sweden, O=SICS, CN=HopsIntermediateCA",
                    "validityDuration": "3650d"
                },
                "kubernetesCA": {}
            }
            databaseUtils.update_variable("pki_ca_configuration", json.dumps(config))

class WorkerPreInit(BasePreInit):
    def __init__(self, name: str, config: configparser.ConfigParser, check_file: pathlib.Path,
        bucket_configuration: BucketConfiguration, region: str, api_key:str, cloud_provider: str, search_domain: str, node_id: str):
        super().__init__(name, config, check_file, bucket_configuration, region, cloud_provider, search_domain, node_id)
        self.api_key = api_key

    def execute(self) -> List[Procedure]:
        self.fix_hopsfs_s3_conf_attrs()
        self.delete_datanode_directories()
        self.update_kagent_password(self.api_key)
        self.touch_check_file()

        return []

class SecondaryPreInit(MasterPreInit):

    def __init__(self, name: str, config: configparser.ConfigParser, check_file: pathlib.Path,
        bucket_configuration: BucketConfiguration, region: str, ldap: str, oauth: str, password_file: pathlib.Path, user_email: str,
        gateway: str, user_id: str, instance_id: str, api_key: str, cloud_provider: str, user_mgm_type: str, create_pwd_user: bool,
        is_upgrade: bool, source_version: str, upgrade_version: str, search_domain: str, node_id: str,
        is_backup: bool, offline_backup_info: OffLineBackupInformation, cluster_domain_prefix: str, services_list : List[Service], k8s_enabled: bool, enable_flyingduck: bool, nlb_domain: str):

        super().__init__(name, config, check_file, bucket_configuration, region, ldap, oauth, password_file, user_email,
        gateway, user_id, instance_id, api_key, cloud_provider, user_mgm_type, create_pwd_user,
        is_upgrade, source_version, upgrade_version, search_domain, node_id, is_backup, offline_backup_info, cluster_domain_prefix, services_list, k8s_enabled, enable_flyingduck, nlb_domain)
        
        self.services_list = services_list

    def execute(self) -> List[Procedure]:

        self.fix_hopsfs_s3_conf_attrs()

        self.remove_unused_services()
        
        self._config_update_kagent_password(self.api_key)

        self.deny_outside_access_to_yarn_ui()

        self.clean_up_opensearch_data_dir()
        
        self.clean_up_kafka_data_dir()
        
        self.delete_datanode_directories()

        self.touch_check_file()
        return []

    def clean_up_opensearch_data_dir(self) -> None:
        logger.info("%s - clean up opensearch data dir")
        ShellUtils.rmtree("/srv/hops/opensearch-data", keep_root=True)


class MysqlNodePreInit(Procedure):

    FLYING_DUCK_CONSUL = "/srv/hops/consul/consul.d/flyingduck.hcl"
    JOIN_HOPSWORKS_CLUSTER_CONFIG="/srv/hops/ndb-agent/ndb-agent/templates/join_hopsworks_cluster.sh"

    def __init__(self, name: str, check_file: pathlib.Path, config: configparser.ConfigParser, enable_flyingduck: bool, docker_repos: Dict[DockerImageName, str], cloud_provider: str, search_domain :str, node_id :str):
        self.name = name
        self.config = config 
        self.check_file = check_file
        self.enable_flyingduck = enable_flyingduck
        self.docker_repos = docker_repos
        self.cloud_provider = cloud_provider
        self.search_domain = search_domain
        self.node_id = node_id

    def pre(self) -> None:
        pass 

    def rollback(self) -> None:
        pass 

    def execute(self) -> List[Procedure]:
        if self.enable_flyingduck:
            self._enable_flyingduck()
        else:
            self._disable_flyingduck()
            self._set_default_mysqld_memory_limits()
        
        if self.check_file is not None:
            self.check_file.touch()
        return []

    def _enable_flyingduck(self) -> None:
        logger.info(f"{self.name} - Use MySQLNodeWithFlyingDuckJoinCluster plan")
        SystemUtils.execute_command([f"sed -i s/NdbNodeJoinCluster/MySQLNodeWithFlyingDuckJoinCluster/g {MysqlNodePreInit.JOIN_HOPSWORKS_CLUSTER_CONFIG}"], shell=True)

        # set memory limit for flyingduck, set to 50% of the available memory - at least 8 GB
        available_memory = SystemUtils.get_system_memory()/1024
        available_cpus = SystemUtils.get_num_cpus()
        logger.info(f"{self.name} - available_memory = {available_memory} , available_cpus = {available_cpus}")

        flyingduck_service_file = f"/lib/systemd/system/{Service.FlyingDuck.value}.service"
        flyingduck_memory = max(8, int(0.5 * available_memory))
        flyingduck_cpu = int(0.5 * available_cpus)

        logger.info(f"{self.name} - Set flyingduck memory limit to {flyingduck_memory} GB")
        SystemUtils.execute_command([f"sed -i s/'FLYINGDUCK_MEMORY=.*'/'FLYINGDUCK_MEMORY={flyingduck_memory}'/g {flyingduck_service_file}"], shell=True)
        SystemUtils.execute_command([f"sed -i s/'FLYINGDUCK_CPU_LIMIT=.*'/'FLYINGDUCK_CPU_LIMIT=--cpuset-cpus=\"0-{flyingduck_cpu-1}\"'/g {flyingduck_service_file}"], shell=True)

        fqdn = get_cloud_provider(self.cloud_provider).get_fqdn(self.config, self.search_domain, self.node_id)
        SystemUtils.execute_command([f"sed -i 's/--fqdn=.*/--fqdn={fqdn} \\\/'g {flyingduck_service_file}"], shell=True)

        mysqld_memory_max = int(0.4 * available_memory)
        self._set_mysqld_limits(mysqld_memory_max, f"{flyingduck_cpu}-{available_cpus-1}")
    
    def _set_mysqld_limits(self, memory_max: int, cpu_affinity: Optional[str] ) -> None:
        mysqld_service_file = f"/lib/systemd/system/{Service.MySQLd.value}.service"
        logger.info(f"{self.name} - Set mysqld limits to MemoryMax={memory_max}G CPUAffinity={cpu_affinity}")

        mysqld_config = SystemUtils.read_config_case_sensitive(mysqld_service_file)

        mysqld_config.set("Service", "MemoryMax", f"{memory_max}G")

        if cpu_affinity is not None:
            mysqld_config.set("Service", "CPUAffinity", cpu_affinity)

        with open(str(mysqld_service_file), 'w') as fd:
            mysqld_config.write(fd)
        
        SystemUtils.execute_command(["systemctl", "daemon-reload"])

    def _set_default_mysqld_memory_limits(self) -> None:
        available_memory = SystemUtils.get_system_memory()/1024
        mysqld_memory_max = int(0.9 * available_memory)
        self._set_mysqld_limits(mysqld_memory_max, None)


    def _disable_flyingduck(self) -> None:
        self._remove_kagent_service(Service.FlyingDuck)
        if os.path.exists(MysqlNodePreInit.FLYING_DUCK_CONSUL):
            logger.info("%s - removing %s", self.name, MysqlNodePreInit.FLYING_DUCK_CONSUL)
            os.remove(MysqlNodePreInit.FLYING_DUCK_CONSUL)
        for img, tag in self.docker_repos.items():
            if img == DockerImageName.FlyingDuck:
                SetupDockerImages.remove_image(f"{img.value}:{tag}")

    def _remove_kagent_service(self, service_to_remove : Service) -> None:
        kagent_conf_dir = self.config.get(
            'general', 'kagent_conf_dir', fallback="/srv/hops/kagent/etc")
        services_file = pathlib.Path(kagent_conf_dir, "services")
        services = configparser.ConfigParser()
        services.read(str(services_file))

        for service_section in services.sections():
            service = services.get(service_section, 'service', fallback='service_does_not_exist')
            if service != 'service_does_not_exist' and service == service_to_remove.value:
                del services[service_section]
                SystemUtils.execute_command(["systemctl", "stop", service])
                SystemUtils.execute_command(["systemctl", "disable", service])
                SystemUtils.execute_command(["systemctl", "mask", service])
                logger.info("%s - removing service from kagent %s", ServiceUtils.__name__, str(service))
        
        with open(str(services_file), 'w') as fd:
            services.write(fd) 