import configparser
import logging
import os

from typing import List

from ec2init import Procedure
from ec2init.plans.plan_utils import HopsPlanUtils
from ec2init.utils.cloud import Provider, get_cloud_provider
from ec2init.utils.hopsfs import HopsFSUtils
from ec2init.utils.system import SystemUtils

logger = logging.getLogger(__name__)


class SetCustomHostname(Procedure):
    EC2_DEFAULT_DOMAIN_NAME = "default"

    def __init__(self, name: str, config: configparser.ConfigParser, plan_utils: HopsPlanUtils, cloud_provider: str, domain_name: str) -> None:
        self.name = name
        self.config = config
        self.plan_utils = plan_utils
        self.cloud_provider = cloud_provider
        self.domain_name = domain_name

    def execute(self) -> List[Procedure]:
        cloud_provider = get_cloud_provider(self.cloud_provider)
        try:
            hostname = cloud_provider.get_fqdn(self.config, self.domain_name, '')
            logger.info(f"Setting custom hostname {hostname}")
            SystemUtils.execute_command(['hostnamectl', 'set-hostname', hostname])
            cloud_cfg_file = "/etc/cloud/cloud.cfg"
            if os.path.exists(cloud_cfg_file):
                cmd = ['sudo', 'sed', '-i', 's/preserve_hostname:[[:space:]]false/preserve_hostname: true/g', cloud_cfg_file]
                SystemUtils.execute_command(cmd)
            logger.info("Finished setting custom hostname")
        except Exception as ex:
            logger.error(f"Failed to set custom hostname. Reason: {ex}")
            raise Exception("Failed to set custom hostname") from ex

        try :
            self._hopsfs_disable_hostname_check()
        except FileNotFoundError:
            logger.debug("Skip disabling hostname verification in HopsFS because hdfs-site.xml does not exist")
        except Exception as ex:
            logger.error(f"Failed to disable hostname verification in HopsFS. Reason: {ex}")
            raise Exception("Failed to disable hostname verification in HopsFS") from ex

        self._done()
        return []

    def _hopsfs_disable_hostname_check(self) -> None:
        property_name = "dfs.namenode.datanode.registration.ip-hostname-check"
        HopsFSUtils.fix_hdfs_site(property_name, "false")

    def _done(self) -> None:
        check_file = self.plan_utils.get_procedure_check_file(SetCustomHostname.__name__)
        check_file.touch()

    def rollback(self) -> None:
        pass

    @staticmethod
    def should_run(plan_utils: HopsPlanUtils, cloud_provider: str, search_domain: str) -> bool:
        check_file = plan_utils.get_procedure_check_file(SetCustomHostname.__name__)
        ## Currently we set custom hostname only in AWS
        return cloud_provider == Provider.AWS and not plan_utils.check_file_exists(check_file)\
            and search_domain.lower() != SetCustomHostname.EC2_DEFAULT_DOMAIN_NAME