import logging

from typing import List
from ec2init import Procedure, MultiThreadProcedure
from ec2init.utils.system import SystemUtils
from ec2init.utils.shell import ShellUtils

import time 
import os
import shutil

logger = logging.getLogger(__name__)

class ParallelCp(Procedure):
    def __init__(self, name: str, src_dir: str, dst_dir: str, max_workers: int=6, timeout: int=3600, overwrite:bool=True) -> None:
        self.name = name
        self.src_dir = src_dir
        self.dst_dir = dst_dir
        self.max_workers = max_workers
        self.timeout = timeout
        self.overwrite = overwrite
    
    def execute(self) -> List['Procedure']:
        if self.overwrite:
            ShellUtils.rmtree(self.dst_dir, keep_root=True)
        
        for d in os.listdir(self.src_dir):
            src_d = os.path.join(self.src_dir, d)
            if os.path.isfile(src_d) or len(os.listdir(src_d)) <= 1:
                ParallelCpTask.copy(src_d, self.dst_dir, timeout=self.timeout)
            else:
                pcp = ParallelCpTask(self.name, src_d, os.path.join(self.dst_dir, d), self.max_workers, self.timeout)
                pcp.execute()
                pcp.wait_to_finish()

        return []

    def rollback(self) -> None:
        pass

class ParallelCpTask(MultiThreadProcedure):
    def __init__(self, name: str, src_dir: str, dst_dir: str, max_workers: int, timeout: int) -> None:
        self.name = name
        self.src_dir = src_dir
        self.dst_dir = dst_dir
        self.tasks = [os.path.join(self.src_dir, sub) for sub in os.listdir(self.src_dir)]
        self.max_workers = min(max_workers, len(self.tasks))
        self.start_time = time.time()
        self.timeout = timeout
        if not os.path.exists(self.dst_dir):
            os.mkdir(self.dst_dir)
            os.chmod(self.dst_dir, os.stat(self.src_dir).st_mode)

    def pre(self) -> None:
        logger.info("%s start copying %s to %s using %d tasks", self.name, self.src_dir, self.dst_dir, len(self.tasks))

    def task_function(self, task):
        ParallelCpTask.copy(task, self.dst_dir, timeout=self.timeout)

    def task_rollback_function(self, task):
        pass

    def rollback(self) -> None:
        pass

    def completed_callback(self):
        elapsed_time = time.time() - self.start_time
        logger.info("%s took %d seconds to copy %s to %s", self.name, elapsed_time, self.src_dir, self.dst_dir)
    
    @staticmethod
    def copy(src_p: str, dst_parent_p: str, timeout: int=30) -> None:
        SystemUtils.execute_command(['cp', '-r', '-p', src_p, dst_parent_p], timeout=timeout)