import logging
import shutil
import os
from typing import List
from ec2init import Procedure
from ec2init.utils.system import SystemUtils
from pathlib import Path
from configparser import ConfigParser

from ec2init.plans.plan_utils import HopsPlanUtils
from ec2init.utils.shell import ShellUtils

logger = logging.getLogger(__name__)


class UpgradeCluster(Procedure):
    def __init__(self, name: str, config: ConfigParser, check_file: Path, region: str, cloud_provider: str,
                 source_version: str, upgrade_version: str, managed_docker: bool, default_registry_url: str, python_image_name: str, mgmd_ip: str) -> None:
        self.name = name
        self.config = config
        self.check_file = check_file
        self.region = region
        self.cloud_provider = cloud_provider
        self.source_version = HopsPlanUtils.clean_hopsworks_version(source_version)
        self.upgrade_version_uncleaned = upgrade_version
        self.upgrade_version = HopsPlanUtils.clean_hopsworks_version(upgrade_version)
        self.managed_docker = managed_docker
        self.default_registry_url = default_registry_url
        self.python_image_name = python_image_name
        self.base_image_name = "base"
        self.install_dir = "/srv/hops"
        self.stage = 'Cluster upgrade. starting...'
        self.mgmd_ip = mgmd_ip

    def execute(self) -> List[Procedure]:
        try:
            self.cleanup_hopsfs()

            self.fixes_for_3_1()

            self.fixes_for_3_2()

            self.fixes_for_3_3()

            # Every thing is OK.
            if self.check_file is not None:
                self.check_file.touch()
        except Exception as ex:
            self.friendly_name = f"{self.stage}\nError details: {ex}"
            logger.error("%s - Failed at %s", self.name, self.friendly_name)
            raise ex
        return []

    def cleanup_hopsfs(self) -> None:
        logger.info("Deleting datanodes data dirs")
        ShellUtils.rmtree(f"{self.install_dir}/hopsdata/hdfs/dn/cloud/current/")
        ShellUtils.rmtree(f"{self.install_dir}/hopsdata/hdfs/dn/disk/current/")
        
    def fixes_for_3_1(self) -> None:
        staging_data_dir = os.readlink(f"{self.install_dir}/staging")
        if not os.path.exists(staging_data_dir):
            os.makedirs(staging_data_dir)
            SystemUtils.execute_command([f"chown -R glassfish:glassfish {staging_data_dir}"], shell=True)
            SystemUtils.execute_command([f"chmod 775 {staging_data_dir}"], shell=True)

            private_dirs = f"{staging_data_dir}/private_dirs"
            if not os.path.exists(private_dirs):
                os.makedirs(private_dirs)
                SystemUtils.execute_command([f"chown -R yarnapp:glassfish {private_dirs}"], shell=True)
                SystemUtils.execute_command([f"chmod 370 {private_dirs}"], shell=True)

            serving_dir = f"{staging_data_dir}/serving"
            if not os.path.exists(serving_dir):
                os.makedirs(serving_dir)
                SystemUtils.execute_command([f"chown -R glassfish:glassfish {serving_dir}"], shell=True)
                SystemUtils.execute_command([f"chmod 730 {serving_dir}"], shell=True)

            tensorboard_dir = f"{staging_data_dir}/tensorboard"
            if not os.path.exists(tensorboard_dir):
                os.makedirs(tensorboard_dir)
                SystemUtils.execute_command([f"chown -R anaconda:glassfish {tensorboard_dir}"], shell=True)
                SystemUtils.execute_command([f"chmod 770 {tensorboard_dir}"], shell=True)

        # Change ownership on the old CA.certs.pem so that Hopsworks CA can initialize itself.
        ca_cert = f"{self.install_dir}/certs-dir/certs/ca.cert.pem"
        ca_key = f"{self.install_dir}/certs-dir/private/ca.key.pem"
        for ca_file in [ca_cert, ca_key]:
            if os.path.exists(ca_file):
                shutil.chown(ca_file, user="glassfish")

        # remove old host certificates
        subjects_to_revoke = ["consul", "elastic", "elkadmin", "flink", "glassfish", "hdfs",
            "hive", "hopsmon", "kafka", "kagent", "livy", "mapred", "mysql", "rmyarn",
            "yarn", "onlinefs"]

        certs_dir = f"{self.install_dir}/certs-dir/intermediate/certs"
        if os.path.exists(certs_dir):
            for file in os.listdir(certs_dir):
                for subject in subjects_to_revoke:
                    if f"__{subject}__" in file:
                        logger.info("%s Remove old certificate in order to avoid migrating it with expat. The host certificate will be normally re-generated by the init plan %s", self.name, file)
                        os.remove(os.path.join(certs_dir, file))

    def fixes_for_3_2(self) -> None:
        zookeeper_data_dir = os.readlink(f"{self.install_dir}/zookeeper/zookeeper/data")
        if not os.path.exists(zookeeper_data_dir):
            os.makedirs(zookeeper_data_dir)
            SystemUtils.execute_command([f"chown -R zookeeper:zookeeper {zookeeper_data_dir}"], shell=True)
            SystemUtils.execute_command([f"chmod 750 {zookeeper_data_dir}"], shell=True)
    
    def fixes_for_3_3(self) -> None:
        consul_log_dir = os.readlink(f"{self.install_dir}/consul/logs")
        if not os.path.exists(consul_log_dir):
            os.makedirs(consul_log_dir)
            SystemUtils.execute_command([f"chown -R consul:consul {consul_log_dir}"], shell=True)
            SystemUtils.execute_command([f"chmod 750 {consul_log_dir}"], shell=True)
        
        dnsmasq_log_dir = os.readlink( f"{self.install_dir}/dnsmasq/logs")
        if not os.path.exists(dnsmasq_log_dir):
            os.makedirs(dnsmasq_log_dir)
            SystemUtils.execute_command([f"chmod 750 {dnsmasq_log_dir}"], shell=True)

        # Upgrade from Zookeeper 3.4 to 3.7 fails on missing snapshot file 
        # https://issues.apache.org/jira/browse/ZOOKEEPER-3056
        zookeeper_data_dir = f"{self.install_dir}/zookeeper/zookeeper/data/version-2"
        if os.path.exists(zookeeper_data_dir):
            snapshot_files = [file for file in os.listdir(zookeeper_data_dir) if file.startswith("snapshot.")]
            if len(snapshot_files) == 0:
                SystemUtils.execute_command([f'echo "snapshot.trust.empty=true" >> {self.install_dir}/zookeeper/zookeeper/conf/zoo.cfg'], shell=True)
                
    def rollback(self) -> None:
        pass

    def updateHWStage(self, stage: str) -> None:
        self.stage = stage
        if self.report_callback is not None:
            try:
                self.report_callback.set_stage(stage) # type: ignore
                self.report_callback.execute()
            except Exception as ex:
                logger.warning("%s - Could not execute pre-execute: %s", self.name, ex)
