import logging
import os

from .system import SystemUtils
from .http import HttpUtils
from configparser import ConfigParser

from typing import Type
from typing import Dict
import kubernetes
import json 

logger = logging.getLogger(__name__)

_ec2_meta_server = "http://169.254.169.254"
_ec2_metadata_version = "2016-09-02"

_AWS_TOKEN_ENDPOINT = (
    "PUT", f"{_ec2_meta_server}/latest/api/token")
_AWS_PUBLIC_IP_ENDPOINT = (
    "GET", f"{_ec2_meta_server}/{_ec2_metadata_version}/meta-data/public-ipv4")
_AWS_LOCAL_IP_ENDPOINT = (
    "GET", f"{_ec2_meta_server}/{_ec2_metadata_version}/meta-data/local-ipv4")
_AWS_FQDN_ENDPOINT = (
    "GET", f"{_ec2_meta_server}/{_ec2_metadata_version}/meta-data/local-hostname")
_AWS_INSTANCE_ID_ENDPOINT = (
    "GET", f"{_ec2_meta_server}/{_ec2_metadata_version}/meta-data/instance-id")
_AWS_AVAILABILITY_ZONE = (
    "GET", f"{_ec2_meta_server}/{_ec2_metadata_version}/meta-data/placement/availability-zone")


_azure_meta_server = "http://169.254.169.254"
_azure_metadata_version = "2019-11-01"
_azure_http_headers = {"Metadata": "true"}

_AZURE_INSTANCE_ID_ENDPOINT = (
    "GET", f"{_azure_meta_server}/metadata/instance/compute/name?format=text&api-version={_azure_metadata_version}")
_AZURE_PUBLIC_IP_ENDPOINT = (
    "GET", f"{_azure_meta_server}/metadata/instance/network/interface/0/ipv4/ipAddress/0/publicIpAddress?format=text&api-version={_azure_metadata_version}")
_AZURE_PRIVATE_IP_ENDPOINT = (
    "GET", f"{_azure_meta_server}/metadata/instance/network/interface/0/ipv4/ipAddress/0/privateIpAddress?format=text&api-version={_azure_metadata_version}")
_AZURE_AVAILABILITY_ZONE = (
    "GET", f"{_azure_meta_server}/metadata/instance/compute/zone?format=text&api-version={_azure_metadata_version}")

_DEFAULT_AZURE_SEARCH_DOMAIN = ".internal.cloudapp.net"
_DEFAULT_AWS_SEARCH_DOMAIN = "default"

_gcp_meta_server = "http://metadata.google.internal"
_GCP_FQDN_ENDPOINT = ("GET", f"{_gcp_meta_server}/computeMetadata/v1/instance/hostname")
_GCP_PUBLIC_IP_ENDPOINT = ("GET", f"{_gcp_meta_server}/computeMetadata/v1/instance/network-interfaces/0/access-configs/0/external-ip")
_GCP_PRIVATE_IP_ENDPOINT = ("GET", f"{_gcp_meta_server}/computeMetadata/v1/instance/network-interfaces/0/ip")
_GCP_INSTANCE_ID_ENDPOINT = ("GET", f"{_gcp_meta_server}/computeMetadata/v1/instance/name")
_GCP_AVAILABILITY_ZONE = ("GET", f"{_gcp_meta_server}/computeMetadata/v1/instance/zone")
_GCP_ACCESS_TOKEN_URL = "{}/computeMetadata/v1/instance/service-accounts/{}/token"

class CloudProvider:
    @staticmethod
    def get_fqdn(config: ConfigParser, search_domain: str, node_id: str) -> str:
        pass

    @staticmethod
    def get_hosted_zone(config: ConfigParser) -> str:
        hosted_zone = config.get('initplan', 'hosted_zone', fallback='cloud.hopsworks.ai')
        custom_hosted_zone = os.environ.get('CUSTOM_HOSTED_ZONE')
        if custom_hosted_zone:
            logger.info(f"Using custom Hosted Zone: {custom_hosted_zone}")
            return custom_hosted_zone
        if SystemUtils.is_production_environment():
            return hosted_zone
        tokens = hosted_zone.split(".")
        tokens[0] = "dev-cloud"
        return ".".join(tokens)

    @staticmethod
    def get_public_domain_name(config: ConfigParser, cluster_id: str, cluster_domain_prefix: str) -> str:
        hosted_zone = CloudProvider.get_hosted_zone(config)
        name = cluster_id
        if cluster_domain_prefix!="":
            name = cluster_domain_prefix
        if hosted_zone.startswith('.'):
            return f"{name}{hosted_zone}"
        return f"{name}.{hosted_zone}"

    @staticmethod
    def get_public_ip(config: ConfigParser) -> str:
        pass

    @staticmethod
    def get_private_ip(config: ConfigParser) -> str:
        pass

    @staticmethod
    def get_node_id(config: ConfigParser) -> str:
        pass

    @staticmethod
    def get_availability_zone(config: ConfigParser) -> str:
        pass

    @staticmethod
    def select_dns_name(fqdn: str, search_domain: str) -> str:
        if "\n" not in fqdn:
            return fqdn
        candidates = list(filter(lambda x: search_domain_with_trailing_dot(search_domain) in x, fqdn.split("\n")))
        if len(candidates) < 1:
            return _remove_trailing_dot(fqdn.split("\n")[0])
        return _remove_trailing_dot(candidates[0])
    

    @staticmethod
    def update_k8s_coredns(config: ConfigParser) -> None:
        pass

    @staticmethod
    def default_search_domain() -> str:
        pass

def _call_metadata_server(config: ConfigParser, endpoint: tuple, headers: Dict[str, str] = {}, data: Dict[str, str] = None) -> Dict[str, str]:
    method, url = endpoint
    response = HttpUtils.do_api_call(
        config, url, method, headers=headers, data=data)
    return response

def _call_metadata_server_with_response(config: ConfigParser, endpoint: tuple, headers: Dict[str, str] = {}, data: Dict[str, str] = None) -> str:
    response = _call_metadata_server(config, endpoint, headers, data)
    return response['response']

def search_domain_with_trailing_dot(search_domain: str) -> str:
    if search_domain.startswith("."):
        return search_domain
    else :
        return "." + search_domain

def _remove_trailing_dot(fqdn: str) -> str:
    if fqdn.endswith("."):
        return fqdn[:-1]
    return fqdn

class Azure(CloudProvider):
    @staticmethod
    def get_fqdn(config: ConfigParser, search_domain: str, node_id: str) -> str:
        try:
            fqdn = f"{node_id}{search_domain_with_trailing_dot(search_domain)}"
            #check that the provided search_domain and host_name give a resolvable fqdn
            ip = SystemUtils.execute_command(["dig", fqdn, "+short"]).stdout.strip().decode("utf-8").replace('"','')
            logger.debug(f"fqdn is {fqdn} and ip is {ip} while private ip is {Azure.get_private_ip(config)}")
            private_ip = Azure.get_private_ip(config)
        except Exception as ex:
            raise Exception("Cluster initialization failed, could not get local hostname")
        if ip != private_ip:
            raise Exception(f"the provided host name ({node_id}) and search domain ({search_domain}) do not form a resolvable fqdn")
        return fqdn

    @staticmethod
    def get_public_ip(config: ConfigParser) -> str:
        return _call_metadata_server_with_response(config, _AZURE_PUBLIC_IP_ENDPOINT, _azure_http_headers)

    @staticmethod
    def get_private_ip(config: ConfigParser) -> str:
        return _call_metadata_server_with_response(config, _AZURE_PRIVATE_IP_ENDPOINT, _azure_http_headers)

    @staticmethod
    def get_node_id(config: ConfigParser) -> str:
        return _call_metadata_server_with_response(config, _AZURE_INSTANCE_ID_ENDPOINT, _azure_http_headers)

    @staticmethod
    def get_availability_zone(config: ConfigParser) -> str:
        return _call_metadata_server_with_response(config, _AZURE_AVAILABILITY_ZONE, _azure_http_headers)

    @staticmethod
    def update_k8s_coredns(config: ConfigParser) -> None:
        kubernetes.config.load_kube_config()
        core_v1 = kubernetes.client.CoreV1Api()
        consul_server = Azure.get_private_ip(config)
        coredns_config_map = core_v1.read_namespaced_config_map("coredns-custom", "kube-system")
        if coredns_config_map.data is None:
            coredns_config_map.data = dict()
        coredns_config_map.data['consul.server'] = "consul:53 {\n    errors\n    cache 30\n    forward . " + consul_server + "\n}"
        core_v1.patch_namespaced_config_map("coredns-custom", "kube-system", coredns_config_map) 

    @staticmethod
    def default_search_domain() -> str:
        return _DEFAULT_AZURE_SEARCH_DOMAIN

class AWS(CloudProvider):
    
    _token_header = ""

    @staticmethod
    def get_token_header(config: ConfigParser) -> Dict[str, str]:
        # ec2-init does not run for more than a few minutes so we should not need to renew the token
        if AWS._token_header == "" :
            test = _call_metadata_server_with_response(config, _AWS_TOKEN_ENDPOINT, {"X-aws-ec2-metadata-token-ttl-seconds": "21600"})
            AWS._token_header = test
        return {"X-aws-ec2-metadata-token": AWS._token_header}
    
    @staticmethod
    def get_fqdn(config: ConfigParser, search_domain: str, node_id: str) -> str:
        try:
            sanitized_domain_name = _remove_trailing_dot(search_domain).strip()
            if sanitized_domain_name.lower() == _DEFAULT_AWS_SEARCH_DOMAIN:
                return _call_metadata_server_with_response(config, _AWS_FQDN_ENDPOINT, AWS.get_token_header(config))
            instance_id = _call_metadata_server_with_response(config, _AWS_INSTANCE_ID_ENDPOINT, AWS.get_token_header(config))
            return f"{instance_id}.{sanitized_domain_name}"
        except Exception as ex:
            raise Exception("Cluster initialization failed, could not get local hostname") from ex

    @staticmethod
    def get_public_ip(config: ConfigParser) -> str:
        return _call_metadata_server_with_response(config, _AWS_PUBLIC_IP_ENDPOINT, AWS.get_token_header(config))

    @staticmethod
    def get_private_ip(config: ConfigParser) -> str:
        return _call_metadata_server_with_response(config, _AWS_LOCAL_IP_ENDPOINT, AWS.get_token_header(config))

    @staticmethod
    def get_node_id(config: ConfigParser) -> str:
        return _call_metadata_server_with_response(config, _AWS_INSTANCE_ID_ENDPOINT, AWS.get_token_header(config))
    
    @staticmethod
    def get_availability_zone(config: ConfigParser) -> str:
        return _call_metadata_server_with_response(config, _AWS_AVAILABILITY_ZONE, AWS.get_token_header(config))

    @staticmethod
    def update_k8s_coredns(config: ConfigParser) -> None:
        kubernetes.config.load_kube_config()
        core_v1 = kubernetes.client.CoreV1Api()
        consul_server = AWS.get_private_ip(config)
        coredns_config_map = core_v1.read_namespaced_config_map("coredns", "kube-system")
        consul_index = coredns_config_map.data['Corefile'].find("consul:53")
        if consul_index > 0:
            coredns_config_map.data['Corefile'] = coredns_config_map.data['Corefile'][:consul_index]
        
        coredns_config_map.data['Corefile'] = coredns_config_map.data['Corefile'] + "consul:53 {\n    errors\n    cache 30\n    forward . " + consul_server + "\n}"
        core_v1.patch_namespaced_config_map("coredns", "kube-system", coredns_config_map)

    @staticmethod
    def default_search_domain() -> str:
        return _DEFAULT_AWS_SEARCH_DOMAIN

class GCP(CloudProvider):
    @staticmethod
    def get_fqdn(config: ConfigParser, search_domain: str, node_id: str) -> str:
        try:
            return _call_metadata_server_with_response(config, _GCP_FQDN_ENDPOINT, { "Metadata-Flavor": "Google" })
        except Exception as ex:
            raise Exception("Cluster initialization failed, could not get local hostname")

    @staticmethod
    def get_public_ip(config: ConfigParser) -> str:
        return _call_metadata_server_with_response(config, _GCP_PUBLIC_IP_ENDPOINT, { "Metadata-Flavor": "Google" })

    @staticmethod
    def get_private_ip(config: ConfigParser) -> str:
        return _call_metadata_server_with_response(config, _GCP_PRIVATE_IP_ENDPOINT, { "Metadata-Flavor": "Google" })

    @staticmethod
    def get_node_id(config: ConfigParser) -> str:
        return _call_metadata_server_with_response(config, _GCP_INSTANCE_ID_ENDPOINT, { "Metadata-Flavor": "Google" })
    
    @staticmethod
    def get_availability_zone(config: ConfigParser) -> str:
        return _call_metadata_server_with_response(config, _GCP_AVAILABILITY_ZONE, { "Metadata-Flavor": "Google" })

    @staticmethod
    def select_dns_name(fqdn: str, search_domain: str) -> str:
        return fqdn

    @staticmethod
    def update_k8s_coredns(config: ConfigParser) -> None:
        # update KubeDNS
        kubernetes.config.load_kube_config()
        core_v1 = kubernetes.client.CoreV1Api()
        consul_server = GCP.get_private_ip(config)
        kubedns_config_map = core_v1.read_namespaced_config_map("kube-dns", "kube-system")
        logger.info("got kube-dns %s", kubedns_config_map)
        if kubedns_config_map.data is None:
            kubedns_config_map.data = { "stubDomains" : "{}" }

        stub_domains = json.loads(kubedns_config_map.data["stubDomains"])
        if "consul" in stub_domains:
            stub_domains["consul"] = [ consul_server ]
        else:
            stub_domains = { "consul" : [ consul_server ] }

        kubedns_config_map.data["stubDomains"] = json.dumps(stub_domains)
        try:
            core_v1.patch_namespaced_config_map("kube-dns", "kube-system", kubedns_config_map)
        except kubernetes.client.rest.ApiException as ex:
            logger.warning(" failed to update kube-dns configuration with the following exception", exc_info=ex)

    @staticmethod
    def default_search_domain() -> str:
        return ""

    @staticmethod
    def get_access_token(config: ConfigParser, service_account: str) -> str:
        url = _GCP_ACCESS_TOKEN_URL.format(_gcp_meta_server, service_account)
        response = _call_metadata_server(config, ('GET', url), { "Metadata-Flavor": "Google" })
        return response['access_token']


class Provider():
    AZURE = 'azure'
    AWS = 'aws'
    GCP = 'gcp'


_cloud_provider_factory = {
    Provider.AZURE: Azure,
    Provider.AWS: AWS,
    Provider.GCP: GCP
}


def get_cloud_provider(provider: str) -> Type[CloudProvider]:
    # Fallback on AWS to be backwards compatible in places where cloud_provider is not set
    return _cloud_provider_factory.get(provider, AWS)
