import logging
import os
import shutil
import pathlib
import configparser
import subprocess
from ec2init.utils.services import Service

from ec2init.utils.system import SystemUtils
from typing import Any, Dict, List

from ec2init.utils.systemd import SystemdUtils

logger = logging.getLogger(__name__)

GLASSFISH_INSTALL_DIR = "/srv/hops/glassfish/versions/current"
ASADMIN = "/srv/hops/glassfish/versions/current/bin/asadmin"
PASSWORD_FILE = "/srv/hops/domains/domain1_admin_passwd"
ADMIN_USER = "adminuser"
ADMIN_PORT = "4848"
HTTPS_PORT = 443
ADMIN_HOST = "localhost"

NODE_DIR = "/srv/hops/nodes"
PAYARA_CONFIG = "hopsworks-config"
DEPLOYMENT_GROUP = "hopsworks-dg"

LDAP_RESOURCE_NAME = "ldap/LdapResource"
GLASSFISH_START_TIMEOUT = 900
GLASSFISH_SERVICE_GROUP = "glassfish_domain1"
DEFAULT_TARGET = "server"
HOPSWORKS_USER = "glassfish"
HOPSWORKS_GROUP = "glassfish"


class AsadminUtils:

    @staticmethod
    def exec_cmd(cmd: List[str], user: str = None, timeout=60):
        return SystemUtils.execute_command_capture(cmd, user=user, timeout=timeout)

    @staticmethod
    def exec_cmd_on_output(cmd: List[str], stdout: Any):
        # mypy error: No overload variant of "run" matches argument types "List[str]", "Any", "bool"
        # error is on capture_output=True which is a valid arg
        return subprocess.run(cmd, input=stdout, capture_output=True) # type: ignore

    @staticmethod
    def asadmin(cmd: List[str], user: str = HOPSWORKS_USER, host=ADMIN_HOST, port=ADMIN_PORT, timeout=60):
        command = [
            ASADMIN,
            "-I", "false",
            "-u", ADMIN_USER,
            "--host", host,
            "--port", port,
            "--passwordfile", PASSWORD_FILE
        ]
        asadmin_command = command + cmd
        logger.info("executing asadmin commnad - %s",
                    ' '.join(asadmin_command))
        return AsadminUtils.exec_cmd(asadmin_command, user=user, timeout=timeout)

    @staticmethod
    def start_server(server=Service.Glassfish.value):
        SystemUtils.execute_command(["systemctl", "start", server],
                                    timeout=GLASSFISH_START_TIMEOUT)

    @staticmethod
    def stop_server(server=Service.Glassfish.value):
        SystemUtils.execute_command(["systemctl", "stop", server],
                                    timeout=GLASSFISH_START_TIMEOUT)

    @staticmethod
    def restart_server(server=Service.Glassfish.value):
        SystemUtils.execute_command(["systemctl", "restart", server],
                                    timeout=GLASSFISH_START_TIMEOUT)

    @staticmethod
    def check_output(output: subprocess.CompletedProcess, result: str):
        return output.returncode == 0 and result in output.stdout.decode().strip()

    @staticmethod
    def check_output_error(output: subprocess.CompletedProcess, msg: str):
        if output.returncode > 0:
            raise Exception(msg + output.stdout.decode("utf-8"))

    @staticmethod
    def create_ldap_jndi_resource(jndiLookupName: str, properties: List[Any], target=DEFAULT_TARGET):
        command = [
            "create-jndi-resource",
            "--restype", "javax.naming.ldap.LdapContext",
            "--factoryclass", "com.sun.jndi.ldap.LdapCtxFactory",
            "--jndilookupname", jndiLookupName
        ]
        logger.info("Create ldap jndi resource properties %s", properties)
        if properties:
            props = []
            for prop in properties:
                value = prop['value'].replace(':', '\\:').replace('=', '\\=')
                props.append(f"{prop['name']}={value}")
            command.append("--property")
            command.append(":".join(props))
        command.append("--target")
        command.append(target)
        command.append(LDAP_RESOURCE_NAME)
        # create should be done from primary head node, so use default host and port
        output = AsadminUtils.asadmin(command)
        AsadminUtils.check_output_error(output, "Configure LDAP. Caused by ")
        return output

    @staticmethod
    def ldap_resource_created(target=DEFAULT_TARGET):
        command = [
            "list-jndi-resources",
            "--target", target
        ]
        # create should be done from primary head node, so use default host and port
        output = AsadminUtils.asadmin(command)
        return AsadminUtils.check_output(output, LDAP_RESOURCE_NAME)

    @staticmethod
    def create_deployment_group():
        if not AsadminUtils.deployment_group_created():
            command = [
                "create-deployment-group",
                DEPLOYMENT_GROUP
            ]
            output = AsadminUtils.asadmin(command)
            AsadminUtils.check_output_error(
                output, "Creating deployment group. Caused by ")
            return output
        return None

    @staticmethod
    def deployment_group_created():
        command = [
            "list-deployment-groups",
        ]
        output = AsadminUtils.asadmin(command)
        return AsadminUtils.check_output(output, DEPLOYMENT_GROUP)

    @staticmethod
    def add_instance_to_deployment_group(instance: str, host=ADMIN_HOST, port=ADMIN_PORT):
        command = [
            "add-instance-to-deployment-group",
            "--instance", instance,
            "--deploymentgroup",
            DEPLOYMENT_GROUP
        ]
        output = AsadminUtils.asadmin(command, host=host, port=port)
        AsadminUtils.check_output_error(
            output, f"Adding {instance} to deployment group. Caused by ")
        return output

    @staticmethod
    def set_property(key: str, value: str):
        command = [
            "set",
            f"{key}={value}",
        ]
        return AsadminUtils.asadmin(command)

    @staticmethod
    def get_property(key: str):
        command = [
            "get",
            key,
        ]
        return AsadminUtils.asadmin(command)

    @staticmethod
    def set_properties(properties: Dict[str, str]):
        for k, v in properties.items():
            AsadminUtils.set_property(k, v)

    @staticmethod
    def create_local_instance(instance: str, node_name: str, host=ADMIN_HOST, port=ADMIN_PORT):
        if not AsadminUtils.instance_created(instance, host=host, port=port):
            command = [
                "create-local-instance",
                "--config", PAYARA_CONFIG,
                "--node", node_name,
                "--nodedir", NODE_DIR,
                instance
            ]
            output = AsadminUtils.asadmin(command, host=host, port=port)
            AsadminUtils.check_output_error(
                output, "Creating local instance. Caused by ")
            return output
        return None

    @staticmethod
    def instance_created(instance: str, host=ADMIN_HOST, port=ADMIN_PORT):
        command = [
            "list-instances",
        ]
        output = AsadminUtils.asadmin(command, host=host, port=port)
        return AsadminUtils.check_output(output, instance)

    @staticmethod
    def create_system_properties(instance: str, property: str, host=ADMIN_HOST, port=ADMIN_PORT):
        command = [
            "create-system-properties",
            "--target", instance,
            property
        ]
        output = AsadminUtils.asadmin(command, host=host, port=port)
        AsadminUtils.check_output_error(
            output, "Creating system properties. Caused by ")
        return output

    @staticmethod
    def set_hazelcast_public_address(instance: str, ip: str, host=ADMIN_HOST, port=ADMIN_PORT):
        return AsadminUtils.create_system_properties(instance, f"hazelcast.local.publicAddress={ip}", host=host, port=port)

    @staticmethod
    def set_hazelcast_configuration(private_ip: str, interfaces: List[str], host=ADMIN_HOST, port=ADMIN_PORT):
        command = [
            "set-hazelcast-configuration",
            "--enabled", "true",
            "--publicaddress", private_ip,
            "--daspublicaddress", private_ip,
            "--autoincrementport", "true",
            "--interfaces", ','.join(interfaces),
            "--membergroup", PAYARA_CONFIG,
            "--target", PAYARA_CONFIG
        ]
        output = AsadminUtils.asadmin(command, host=host, port=port)
        AsadminUtils.check_output_error(
            output, "Setting hazelcast configuration. Caused by ")
        return output

    @staticmethod
    def get_node_index(host=ADMIN_HOST, port=ADMIN_PORT):
        command = [
            "-t",
            "list-nodes"
        ]
        output = AsadminUtils.asadmin(command, host=host, port=port)
        AsadminUtils.check_output_error(output, "Get node index. Caused by ")
        output = AsadminUtils.exec_cmd_on_output(
            ["wc", "-l"], stdout=output.stdout)
        AsadminUtils.check_output_error(
            output, "Get node index count output. Caused by ")
        return output.stdout.decode("utf-8").strip()

    @staticmethod
    def get_node_name(node_host: str, host=ADMIN_HOST, port=ADMIN_PORT):
        command = [
            "-t",
            "list-nodes",
            "-l"
        ]
        output = AsadminUtils.asadmin(command, host=host, port=port)
        AsadminUtils.check_output_error(output, "Get node name. Caused by ")
        output = AsadminUtils.exec_cmd_on_output(
            ["grep", node_host], stdout=output.stdout)
        AsadminUtils.check_output_error(output, "Grep node name. Caused by ")
        output = AsadminUtils.exec_cmd_on_output(
            ["awk", "{print $1}"], stdout=output.stdout)
        AsadminUtils.check_output_error(output, "awk node name. Caused by ")
        return output.stdout.decode("utf-8").strip()

    @staticmethod
    def create_node_config(node_host: str, host=ADMIN_HOST, port=ADMIN_PORT):
        if not AsadminUtils.node_created(node_host, host=host, port=port):
            index = AsadminUtils.get_node_index()
            node_name = f"worker{index}"
            command = [
                "create-node-config",
                "--nodehost", node_host,
                "--installdir", GLASSFISH_INSTALL_DIR,
                "--nodedir", NODE_DIR,
                node_name
            ]
            output = AsadminUtils.asadmin(command, host=host, port=port)
            AsadminUtils.check_output_error(
                output, f"Creating node on {node_host}. Caused by ")
            return output
        return None

    @staticmethod
    def node_created(node_host: str, host=ADMIN_HOST, port=ADMIN_PORT):
        command = [
            "list-nodes",
            "-l"
        ]
        output = AsadminUtils.asadmin(command, host=host, port=port)
        return AsadminUtils.check_output(output, node_host)

    @staticmethod
    def create_resource_ref(resource: str):
        if AsadminUtils.resource_ref_created(resource) and not AsadminUtils.resource_ref_created(resource, target=DEPLOYMENT_GROUP):
            command = [
                "create-resource-ref",
                "--target", DEPLOYMENT_GROUP,
                resource
            ]
            output = AsadminUtils.asadmin(command)
            AsadminUtils.check_output_error(
                output, f"Creating resource ref {resource}. Caused by ")
            return output
        return None

    @staticmethod
    def resource_ref_created(resource: str, target=DEFAULT_TARGET):
        command = [
            "list-resource-refs",
            target
        ]
        output = AsadminUtils.asadmin(command)
        return AsadminUtils.check_output(output, resource)

    @staticmethod
    def list_applications(component_name: str, target=DEFAULT_TARGET):
        command = [
            "list-applications",
            target
        ]
        output = AsadminUtils.asadmin(command)
        return AsadminUtils.check_output(output, component_name)

    @staticmethod
    def undeploy(component_name: str, target=DEFAULT_TARGET):
        command = [
            "undeploy",
            "--target", target,
            component_name
        ]
        return AsadminUtils.asadmin(command, timeout=120)

    @staticmethod
    def deploy(component_name: str, version: str, extension: str, contextroot="", target=DEFAULT_TARGET):
        name_with_version = f"{component_name}:{version}"
        if not AsadminUtils.list_applications(name_with_version, target=target):
            name_with_version_version = f"{name_with_version}-{version}"
            path_to_component = f"/srv/hops/domains/domain1/archives/{name_with_version}/versions/{name_with_version_version}/{name_with_version_version}.{extension}"
            command = [
                "deploy",
                "--target", target,
                "--name", name_with_version,
                "--enabled=true",
                "--upload=true",
                "--force=true",
                "--generatermistubs=false",
                "--availabilityenabled=false",
                "--lbenabled=true",
                "--keepstate=true",
                "--skipdsfailure=true",
                "--verify=false",
                "--precompilejsp=false",
                "--asyncreplication=false"
            ]
            if contextroot:
                command.append(f"--contextroot={contextroot}")
            command.append(path_to_component)
            output = AsadminUtils.asadmin(
                command, timeout=GLASSFISH_START_TIMEOUT)
            AsadminUtils.check_output_error(
                output, f"Deploying {name_with_version}. Caused by ")
            return output

    @staticmethod
    def link_log_dir_to_data_dir(node_name: str, instance_name: str, data_dir: pathlib.Path):
        log_dir = pathlib.Path(NODE_DIR, node_name, instance_name, "logs")
        data_volume_hopsworks = pathlib.Path(data_dir, "hopsworks")
        data_volume_node_dir = pathlib.Path(data_volume_hopsworks, node_name)
        data_volume_logs_dir = pathlib.Path(data_volume_node_dir, "logs")
        if not os.path.exists(data_volume_hopsworks):
            os.makedirs(data_volume_hopsworks)
            shutil.chown(data_volume_hopsworks, group=HOPSWORKS_GROUP, user=HOPSWORKS_USER)
        if not os.path.exists(data_volume_node_dir):
            os.makedirs(data_volume_node_dir)
            shutil.chown(data_volume_node_dir, group=HOPSWORKS_GROUP, user=HOPSWORKS_USER)
        if not os.path.exists(data_volume_logs_dir):
            os.makedirs(data_volume_logs_dir)
            shutil.chown(data_volume_logs_dir, group=HOPSWORKS_GROUP, user=HOPSWORKS_USER)
        if not os.path.islink(log_dir):
            os.symlink(data_volume_logs_dir, log_dir, target_is_directory=True)  

    @staticmethod
    def add_glassfish_instance_to_systemd_unit(glassfish_user=HOPSWORKS_USER, https_port=HTTPS_PORT, host=ADMIN_HOST, port=ADMIN_PORT):
        if not SystemdUtils.is_active(Service.GlassfishInstance.value):
            authbind = '/usr/bin/authbind --deep ' if int(
                ADMIN_PORT) < 1024 or https_port < 1024 else ''
            start_cmd = f"{authbind} {ASADMIN} --host {host} --port {port} --user {ADMIN_USER} --passwordfile {PASSWORD_FILE} start-local-instance --sync normal --nodedir {NODE_DIR}"
            stop_cmd = f"{authbind} {ASADMIN} --host {host} --port {port} --user {ADMIN_USER} --passwordfile {PASSWORD_FILE} stop-local-instance --nodedir {NODE_DIR}"
            restart_cmd = f"{authbind} {ASADMIN} --host {host} --port {port} --user {ADMIN_USER} --passwordfile {PASSWORD_FILE} restart-local-instance --nodedir {NODE_DIR}"
            SystemdUtils.create_service(Service.GlassfishInstance.value, "GlassFish instance",
                                        "syslog.target network.target remote-fs.target mysqld.service", glassfish_user, start_cmd, stop_cmd, restart_cmd)
            output = SystemdUtils.enable(Service.GlassfishInstance.value)
            AsadminUtils.check_output_error(
                output, f"Enabling systemd service {Service.GlassfishInstance.value}. Caused by ")
            output = SystemdUtils.start(Service.GlassfishInstance.value)
            AsadminUtils.check_output_error(
                output, f"Starting systemd service {Service.GlassfishInstance.value}. Caused by ")

    @staticmethod
    def add_glassfish_instance_to_kagent_service(kagent_services_file: pathlib.Path, node_name: str, instance_name: str, group=GLASSFISH_SERVICE_GROUP, service=Service.GlassfishInstance.value, fail_attempts=1):
        services = configparser.ConfigParser()
        services.read(str(kagent_services_file))
        section = f"{group}-{service}"
        log_file = f"{NODE_DIR}/{node_name}/{instance_name}/logs/server.log"
        if section not in services:
            services[section] = {
                'group': group,
                'service': service,
                'stdout-file': log_file,
                'config-file' : '',
                'fail-attempts': fail_attempts,
            }
            with open(str(kagent_services_file), 'w') as fd:
                services.write(fd)
            output = SystemdUtils.restart("kagent", timeout=90)
            AsadminUtils.check_output_error(output, "Restarting systemd service kagent. Caused by ")

    @staticmethod
    def configure_das_node(private_ip: str, head_node_ips: List[str], version: str, data_dir: pathlib.Path, kagent_services_file: pathlib.Path):
        node_name = "localhost-domain1"
        instance_name = "instance0"

        # disable monitoring and http-listeners on server-config
        glassfish_network_listener_conf = {
            "configs.config.server-config.network-config.network-listeners.network-listener.http-listener-2.enabled": "false",
            "configs.config.server-config.network-config.network-listeners.network-listener.https-int-list.enabled": "false",
            "configs.config.server-config.rest-monitoring-configuration.enabled": "false",
            "configs.config.server-config.monitoring-service.mbean-enabled": "false",
            "configs.config.server-config.monitoring-service.monitoring-enabled": "false",
            "configs.config.server-config.microprofile-metrics-configuration.enabled": "false"
        }
        # if http enabled on server disable it
        output = AsadminUtils.get_property(
            "configs.config.server-config.network-config.network-listeners.network-listener.http-int-list.enabled")
        if output.returncode == 0:
            glassfish_network_listener_conf["configs.config.server-config.network-config.network-listeners.network-listener.http-int-list.enabled"] = "false"

        AsadminUtils.set_properties(glassfish_network_listener_conf)
        AsadminUtils.set_hazelcast_configuration(private_ip, head_node_ips)
        AsadminUtils.create_deployment_group()
        AsadminUtils.create_local_instance(instance_name, node_name)
        AsadminUtils.add_instance_to_deployment_group(instance_name)
        # link instance log dir to data dir
        AsadminUtils.link_log_dir_to_data_dir(
            node_name, instance_name, data_dir)

        AsadminUtils.set_hazelcast_public_address(instance_name, private_ip)
        secondary_nodes = head_node_ips.copy()
        secondary_nodes.remove(private_ip)
        for ip in secondary_nodes:
            AsadminUtils.create_node_config(ip)

        # Undeploy hopsworks
        AsadminUtils.undeploy(f"hopsworks-ear:{version}")
        AsadminUtils.undeploy(f"hopsworks-ca:{version}")
        AsadminUtils.undeploy(f"hopsworks-web:{version}")

        # Resources created in default server, so create a reference to the resources in the new config
        glassfish_resources = [
            'concurrent/hopsThreadFactory',
            'concurrent/condaExecutorService',
            'concurrent/hopsExecutorService',
            'concurrent/jupyterExecutorService',
            'concurrent/condaScheduledExecutorService',
            'concurrent/hopsScheduledExecutorService',
            'concurrent/kagentExecutorService',
            'jdbc/airflow',
            'jdbc/featurestore',
            'jdbc/hopsworks',
            'jdbc/hopsworksTimers',
            'ldap/LdapResource',
            'mail/BBCMail'
        ]
        for resource in glassfish_resources:
            AsadminUtils.create_resource_ref(resource)

        AsadminUtils.restart_server()

        AsadminUtils.deploy("hopsworks-ear", version,
                            "ear", target=DEPLOYMENT_GROUP)
        AsadminUtils.deploy("hopsworks-ca", version, "war",
                            contextroot="/hopsworks-ca", target=DEPLOYMENT_GROUP)
        AsadminUtils.deploy("hopsworks-web", version, "war",
                            contextroot="/hopsworks", target=DEPLOYMENT_GROUP)

        # add glassfish-instance to services
        AsadminUtils.add_glassfish_instance_to_systemd_unit()

        # add glassfish-instance to kagent
        AsadminUtils.add_glassfish_instance_to_kagent_service(
            kagent_services_file, node_name, instance_name)

    @staticmethod
    def configure_config_node(primary_ip: str, node_ip: str, data_dir: pathlib.Path, kagent_services_file: pathlib.Path):
        node_name = AsadminUtils.get_node_name(node_ip, host=primary_ip)
        instance_name = node_name.replace("worker", "instance")
        # create node dir
        if not os.path.exists(NODE_DIR):
            os.makedirs(NODE_DIR)
            shutil.chown(NODE_DIR, group=HOPSWORKS_GROUP, user=HOPSWORKS_USER)

        AsadminUtils.create_local_instance(instance_name, node_name, host=primary_ip)

        # link instance log dir to data dir
        AsadminUtils.link_log_dir_to_data_dir(node_name, instance_name, data_dir)

        AsadminUtils.set_hazelcast_public_address(
            instance_name, node_ip, host=primary_ip)
        AsadminUtils.add_instance_to_deployment_group(
            instance_name, host=primary_ip)

        # add glassfish-instance to services
        AsadminUtils.add_glassfish_instance_to_systemd_unit(host=primary_ip)

        # add glassfish-instance to kagent
        AsadminUtils.add_glassfish_instance_to_kagent_service(
            kagent_services_file, node_name, instance_name)
