import configparser
import subprocess
import logging
import os
from pwd import getpwnam
import multiprocessing
from enum import Enum
from pathlib import Path
import time 
from typing import List, Dict, Optional, Callable, Any
from ec2init.utils.constants import Scripts

logger = logging.getLogger(__name__)

class OSType(Enum):
    UBUNTU = 1
    CENTOS = 2

class CaseSensitiveConfigParser(configparser.RawConfigParser):
    def optionxform(self, optionstr: str) -> str:
        return optionstr
    
class SystemUtils:

    UBUNTU_VERSION_FILE = Path("/etc/debian_version")
    CENTOS_RELEASE_FILE = Path("/etc/centos-release")

    @staticmethod
    def execute_command(command: List[str], check: bool = True, env: Dict[str, str] = None,
                        timeout: Optional[int] = 20, shell: bool = False) -> subprocess.CompletedProcess:
        logger.debug("Executing %s", command)

        return subprocess.run(command, stdout=subprocess.PIPE, check=check,
                              env=env, timeout=timeout, shell=shell)
    
    @staticmethod
    def execute_command_capture(command: List[str], user: str = None, timeout: Optional[int] = 20, capture_output: bool = True, env: Dict[str, str] = None) -> subprocess.CompletedProcess:
        if user:
            pwnam = getpwnam(user)
            user_uid = pwnam.pw_uid
            user_gid = pwnam.pw_gid
            logger.debug("Executing %s as %s", command, user)
            # mypy error: No overload variant of "run" matches argument types "List[str]", "Any", "bool"
            # error is on capture_output=True which is a valid arg
            return subprocess.run(command, timeout=timeout, capture_output=capture_output, preexec_fn=SystemUtils.demote(user_uid, user_gid), env=env) # type: ignore
        logger.debug("Executing %s", command)
        # mypy error: No overload variant of "run" matches argument types "List[str]", "Any", "bool"
        # error is on capture_output=True which is a valid arg
        return subprocess.run(command, timeout=timeout, capture_output=capture_output, env=env) # type: ignore
    
    @staticmethod
    def execute_command_capture_with_print_error(command: List[str], user: str = None, timeout: Optional[int] = 20, capture_output: bool = True, env: Dict[str, str] = None) -> subprocess.CompletedProcess:
        result = SystemUtils.execute_command_capture(command=command, user=user, timeout=timeout, capture_output=capture_output, env=env)
        if result.returncode > 0:
            stdout = "None" if result.stdout is None else result.stdout.decode("utf-8")
            stderr = "None" if result.stderr is None else result.stderr.decode("utf-8")
            logger.warning("failed to execute command %s with stdout=%s and stderr=%s", command, stdout, stderr)
        return result

    @staticmethod
    def demote(user_uid, user_gid):
        def result():
            os.setgid(user_gid)
            os.setuid(user_uid)
        return result

    @staticmethod
    def execute_mysql_statement(statement: str, database: str = "hopsworks", timeout: int = 120) -> List[List[str]]:
        logger.debug("Executing %s", statement)
        process = SystemUtils.execute_command([Scripts.MYSQL_CLIENT, database, "-e", statement, "-sN"], check=True, timeout=timeout, shell=False)
        out = process.stdout.decode('latin1')
        rows = [line.strip().split("\t") for line in out.split("\n") if line and not line.startswith('Using socket')]
        return rows

    @staticmethod
    def read_config(config_file: str, strict: bool = True) -> configparser.ConfigParser:
        config = configparser.ConfigParser(strict=strict)
        config.read(config_file)
        return config

    @staticmethod
    def read_config_case_sensitive(config_file: str, strict: bool = True) ->CaseSensitiveConfigParser:
        config = CaseSensitiveConfigParser(strict=strict)
        config.read(config_file)
        return config

    @staticmethod
    def is_production_environment() -> bool:
        env = SystemUtils.get_environment()
        return env is not None and env.lower() == "prod"

    @staticmethod
    def is_staging() -> bool:
        stag = SystemUtils.get_staging()
        return not SystemUtils.is_production_environment() or (stag is not None and stag.lower() == "true")

    @staticmethod
    def get_environment() -> Optional[str]:
        return os.environ.get('APP_ENV')

    @staticmethod
    def get_staging() -> Optional[str]:
        return os.environ.get('CERT_STAGING')

    # Get system memory in Mbytes
    @staticmethod
    def get_system_memory() -> int:
        mem_bytes = os.sysconf('SC_PAGE_SIZE') * os.sysconf('SC_PHYS_PAGES')
        return int(mem_bytes / (1024 * 1024))

    @staticmethod
    def get_num_cpus() -> int:
        return multiprocessing.cpu_count()

    @staticmethod
    def get_os_type() -> OSType:
        if SystemUtils.CENTOS_RELEASE_FILE.exists():
            return OSType.CENTOS
        if SystemUtils.UBUNTU_VERSION_FILE.exists():
            return OSType.UBUNTU
        raise Exception('Unable to determine OS type')

    @staticmethod
    def execute_hdfs_command(command: str, timeout: int=320) -> subprocess.CompletedProcess:
        return SystemUtils.execute_command([f"su - hdfs -c '/srv/hops/hadoop/bin/hdfs {command} '" ], timeout=timeout, shell=True)
    
    @staticmethod
    def retry_operation_on_exception(op_function: Callable, post_exception_action: Optional[Callable] = None, retries: int = 10, waitTime: int = 20) -> Any:
        original_exception: Optional[Exception] = None
        for retry in range(retries):
            try:
                op_resp = op_function()
            except Exception as ex:
                logger.warning("Operation failed. Retries %d", retry)
                if not original_exception:
                    original_exception = ex

                if post_exception_action:
                    try:
                        logger.info("Executing post-exception action")
                        post_exception_action()
                    except Exception as iex:
                        logger.error(f"Exception while executing post-exception action: {iex}")
                time.sleep(waitTime)
                continue
            return op_resp
        if original_exception:
            # The condition is needed for the type system
            # to understand the type is Exception and not Optional[Exception]
            raise original_exception

    @staticmethod
    def execute_command_with_retry(command: List[str], post_exception_action: Optional[Callable] = None, check: bool = True, env: Dict[str, str] = None,
                                   timeout: Optional[int] = 20, shell: bool = False, retries: int = 10, waitTime: int = 20) -> None:
        SystemUtils.retry_operation_on_exception(
            lambda: SystemUtils.execute_command(
                command=command, check=check, env=env, timeout=timeout, shell=shell),
            post_exception_action=post_exception_action, retries=retries, waitTime=waitTime
        )