import subprocess
from ec2init.utils.system import SystemUtils
import os
import pathlib

SYSTEMD_PATH="/lib/systemd/system"

class SystemdUtils:
    
    @staticmethod
    def is_active(name: str, unit='service'):
        output = SystemUtils.execute_command_capture(["systemctl", "is-active", "--quiet", f"{name}.{unit}"])
        return output.returncode == 0
    
    @staticmethod
    def is_enabled(name: str, unit='service') -> bool:
        output = SystemUtils.execute_command_capture(["systemctl", "is-enabled", "--quiet", f"{name}.{unit}"])
        return output.returncode == 0
    
    @staticmethod
    def check_output_error(output: subprocess.CompletedProcess, msg: str):
        if output.returncode > 0:
            raise Exception(msg + output.stdout.decode("utf-8"))

    @staticmethod
    def stop(name: str, timeout=90, unit='service'):
        return SystemUtils.execute_command_capture(["systemctl", "stop", f"{name}.{unit}"], timeout=timeout)

    @staticmethod
    def start(name: str, timeout=900, unit='service'):
        return SystemUtils.execute_command_capture(["systemctl", "start", f"{name}.{unit}"], timeout=timeout)

    @staticmethod
    def restart(name: str, timeout=900, unit='service'):
        return SystemUtils.execute_command_capture(["systemctl", "restart", f"{name}.{unit}"], timeout=timeout)

    @staticmethod
    def enable(name: str, unit='service'):
        return SystemUtils.execute_command_capture(["systemctl", "enable", f"{name}.{unit}"])

    @staticmethod
    def reload():
        return SystemUtils.execute_command_capture(["systemctl", "daemon-reload"])

    @staticmethod
    def remove(name: str, unit='service'):
        file = f"{SYSTEMD_PATH}/{name}.{unit}"
        if os.path.exists(file):
            os.remove(file)

    @staticmethod    
    def disable(name: str, timeout=120, unit='service'):
        return SystemUtils.execute_command(["systemctl", "disable", f"{name}.{unit}"], timeout=timeout)

    @staticmethod
    def is_service_installed(service: str, unit='service') -> bool:
        unit_file = pathlib.Path(f"{SYSTEMD_PATH}/{service}.{unit}")
        return unit_file.is_file()

    @staticmethod
    def create_service(name: str, description: str, after: str, user: str, start_cmd: str, stop_cmd: str, restart_cmd: str, start_timeout=900, stop_timeout=90, unit='service'):
        systemd_script = f'''[Unit]
Description={description}
After = {after}
StartLimitIntervalSec=0

[Install]
WantedBy = multi-user.target

[Service]
User={user}
ExecStart = {start_cmd}
ExecStop = {stop_cmd}
ExecReload = {restart_cmd}
Type = forking
TimeoutStartSec = {start_timeout}
TimeoutStopSec = {stop_timeout}
Restart=always
RestartSec=5'''
        if SystemdUtils.is_active(name):
            output = SystemdUtils.stop(name, unit=unit)
            SystemdUtils.check_output_error(output, f"Stop {name}. Caused by ")
            SystemdUtils.remove(name, unit=unit)
        with open(f"{SYSTEMD_PATH}/{name}.{unit}", 'w') as file:
            file.write(systemd_script)
        output = SystemdUtils.reload()
        SystemdUtils.check_output_error(output, f"Reload {name}. Caused by ")